/*
  orrery Rev 2.7
  Ken Young (orrery.moko@gmail.com)
  First Version Aug. 3, 2007/usr/share/orrery/icons/fullMoonImage.xpm
  
  Copyright (C) (2007, 2008, 2009) Ken Young orrery.moko@gmail.com

  This program is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License 
  as published by the Free Software Foundation; either 
  version 2 of the License, or (at your option) any later 
  version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  This program displays the night (and day!) sky on an Openmoko
  device.
*/

#include <stdio.h>
#include <sys/stat.h>
#include <dirent.h>
#include <locale.h>
#include <fcntl.h>
#include <math.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <unistd.h>
#include <signal.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/time.h>
#include <stdlib.h>
#include <memory.h>
#include <gtk/gtk.h>
#include "orrery.h"
#include "phaseConstants.h"

#define ERROR_EXIT  (-1)
#define SUCCESS_EXIT (0)

#define GPSD_PORT (2947)

#define VISUAL_LIMIT (6.6)
#define CONSTELLATION_LIMIT (5.75)
#define ABSOLUTE_MAX_ZA (95.0 * DEGREES_TO_RADIANS)

#define AU (149.597870691) /* Million kilometers in an AU */

#define MAX_FILE_NAME_SIZE (256) /* Maximum size a file name is allowed to have */

#define DEGREES_TO_RADIANS (M_PI/180.0)
#define HOURS_TO_RADIANS   (M_PI/12.00)
#define HOURS_TO_DEGREES   (15.0)
#define M_2PI              (2.0*M_PI)

/* Star colors */
#define DARK_GREY_LIMIT   (6.25)
#define GREY_LIMIT        (5.75)
#define WHITE_LIMIT       (5.00)

#define FULL_CIRCLE (23040) /* Full circle in calls to draw_arc */

#define VGA_HEIGHT (640)
#define VGA_WIDTH  (480)
#define BUTTON_HEIGHT (70) /* Vertical size of a button */

#define STAR_NAME_V_OFFSET (14) /* Vertical offset of star name from star */

char *orreryVersion = "2.7";
char dataDir[100];
char dirName[80];

int starSize[12] = {11, 10, 9, 8, 7, 6, 5, 4, 4, 3, 3, 2};

int skyIsVisible = TRUE;
int debugMessagesOn = FALSE;
#define dprintf if (debugMessagesOn) printf /* Print IFF debugging          */

int locationChanged = TRUE;
int northernHemisphere = TRUE;
int gPSLocation = FALSE;
char satStatString[20];
double sinLatitude, cosLatitude, sinLatitudeCutoff;
double smallMooncalDateOffset = 0.0;

int showDeepSky = TRUE;  /* Show deep sky objects */
int showGreatCircles  = FALSE; /* Draw celestial equator, ecliptic and galactic plane */
int dayInc = 0;
float limitingMag   = VISUAL_LIMIT;  /* Faintest star to plot   */
float magScale = 1.0;
float magScale1 = 1.0;
float magScale2 = 1.0;
float faintestStar;

int showPlanets = TRUE; /* Plot planets in some way */
int showStars = TRUE; /* Plot fixed objects (stars, galaxies, etc) */
int displayConstellations = FALSE; /* Plot constellation stick figures */
int showNames = FALSE; /* Label bright/famous stars */
int showMeteors = FALSE; /* Show meteor show radiants */
int displayPlanetsAsSymbols = FALSE;

char *solarSystemNames[N_SOLAR_SYSTEM_OBJECTS] = {"Sun",     "Mercury",   "Venus",    "Earth",
						  "Moon",    "Mars",      "Jupiter",  "Saturn",
						  "Uranus",  "Neptune",   "Comet"};
float solarSystemScales[N_SOLAR_SYSTEM_OBJECTS] = {0.0, 20.0, 50.0, 50.0,
						   0.0, 100.0, 200.0, 500.0,
						   1000.0, 2000.0, 0.0};
						   
typedef struct orbitPlot {
  float *x, *y, maxX, omegaRad, peri;
} orbitPlot;
orbitPlot orbitPlots[N_SOLAR_SYSTEM_OBJECTS];

int shouldSwitchScreens = FALSE;
  
float cYear = 2009.0;
float cMonth = 1.0;
float cDay = 1.0;
float cHour = 0.0;
float cMinute = 0.0;
float cSecond = 0.0;

gdouble julianDate = 0.0;

/* Parameters which may be overriden by the config file */

float limitingMagnitude1   = 6.60;  /* Faintest star plotted on page 1           */
float limitingMagnitude2   = 4.50;  /* Faintest star plotted on page 2           */
float initialAzimuth       = 180.;  /* Azimuth initially centered on the display */
char locationName[100];             /* Name of default location                  */
double latitude            = 0.0;   /* Default latitude                          */
double longitude           = 0.0;   /* Default east longitude                    */
int showGreatCircles1      = FALSE; /* Plot the celestial equator, etc           */
int showGreatCircles2      = TRUE;
int showFaintStars         = FALSE;
int showStars1             = TRUE;  /* Plot stars                                */
int showStars2             = TRUE;
int showPlanets1           = TRUE;  /* Plot planets                              */
int showPlanets2           = TRUE;
int showDeepSky1           = TRUE;  /* Plot deep sky objects                     */
int showDeepSky2           = FALSE;
int showNames1             = FALSE; /* Plot the names of stars on page 1         */
int showNames2             = FALSE; /* Plot the names of stars on page 2         */
int showMeteors1           = FALSE; /* Plot meteor shower radiants on page 1     */
int showMeteors2           = FALSE; /* Plot meteor shower radiants on page 2     */
int useAsterisms           = FALSE; /* Plot asterisms instead of constellations  */
int useGPSD                = TRUE;  /* Get position info from gpsd               */

int useCurrentTime = TRUE;
int useCalendarTime = FALSE;
int useJulianTime = FALSE;
int haveFaintStars = FALSE;    /* True if optional faint catalog is available */
int showingFaintStars = FALSE; /* True if optional faint catalog is active    */
int savedTimeMode;
double savedTJD;

/* End of parameters which may be overriden by config file */

int inAzCompassMode = FALSE;

int useCustomLocation = FALSE;
int useMenuLocation = FALSE;
float oneHalf    = 0.5;

struct tm *gMT = NULL;
double tJD           = 0.0;
double myLST;
int needNewTime      = TRUE;

int labelMode = FALSE;
int shouldSayRedrawing = FALSE;

int displayWidth, displayHeight, plotWidth, plotHeight;
int xBorder = 5;
int yBorder = 5;
int xLeftLabelSkip = 10;
int yBottomLabelSkip = 18;
int yTopLabelSkip = 8;
int bigTick = 4;
int smallTick = 2;

float centerAz    = 270.0*DEGREES_TO_RADIANS;
/*
  A double preceision copy of the above variable is maintained in
  order to reduce the number of float->double conversions required
*/
float azSpan      =  90.0*DEGREES_TO_RADIANS;
float labelClipZA =   3.0;
double thetaOffset = 0.0;
double piMinusThetaOffset = M_PI;
double twoPiMinusThetaOffset = M_2PI;
double centerAzD;
double maximumZA  =  ABSOLUTE_MAX_ZA;
double minimumZA  = 0.0;
double maximumDisplayZA = 98.0*DEGREES_TO_RADIANS;
int zoomed = FALSE;

double latDD, latMM, latSS, longDD, longMM, longSS;

/* Variables used in transverse Mercator projection */

GdkPoint *darkGreyPoints;
int nDarkGreyPoints;
GdkPoint *greyPoints;
int nGreyPoints;
GdkPoint *whitePoints;
int nWhitePoints;

float mercatorScale;
float mercatorOffset;

int haveReadStarsFile = FALSE;
int inFlashlightMode = FALSE;
#define ABOUT_SCREEN                       (0)
#define SOLUNI_SCREEN                      (1)
#define BIG_MOONCAL_SCREEN                 (2)
#define SMALL_MOONCAL_SCREEN               (3)
#define PLANETCOMPASS_SCREEN               (4)
#define SOLAR_SYSTEM_SCHEMATIC_SCREEN      (5)
#define SOLAR_SYSTEM_SCALE_SCREEN          (6)
#define METEOR_SHOWERS_SCREEN              (7)

int aboutScreen = ABOUT_SCREEN;
int displayingAnOptsPage = FALSE;

int nSolarSystemDays = 1;

/* Constellation Types: */
#define MODERN    (0)
#define ZODIAC    (1)
#define PTOLEMAIC (2)
#define ASTERISM  (3)

typedef struct constellation {
  char *name;
  int type;
  double nameRA;
  double sinNameDec;
  double cosNameDec;
  int nPoints;
  short *starNeeded;
  double *rA;
  double *dec;
  double *sinDec;
  double *cosDec;
  struct constellation *next;
} constellation;

constellation *constellations = NULL;
int constellationsInitialized = FALSE;

int inFullscreenMode = FALSE;
/*
  fullscreenStateChanging is set when going into or out of
  fullscreen mode, to prevent the screen from being redrawn
  twice.
*/
int fullscreenStateChanging = FALSE;

GdkFont *smallFont;
GdkGC *blueGC, *whiteGC, *greyGC, *darkGreyGC, *greenGC, *darkGreenGC, *redGC, *planetGC;
GdkGC *yellowGC, *goldGC, *faintGoldGC, *creamGC, *blueGreenGC, *blackGC, *equatorGC;
GdkGC *starGC[5], *pinkGC, *faintPinkGC, *faintBlueGC, *lightYellowGC, *flashlightGC;
GdkGC *aboutGC, *darkCreamGC;
GtkWidget *window, *magSpin1, *magSpin2, *latSpinDD, *latSpinMM, *latSpinSS, *julSpin;
GtkWidget *cDateLabel, *cTimeLabel, *julSpinLabel;
GtkWidget *longSpinDD, *longSpinMM, *longSpinSS, *regionMenu;
GtkWidget *yearSpin, *monthSpin, *daySpin, *hourSpin, *minuteSpin, *secondSpin;
GtkWidget *mainBox, *drawingArea, *itemsTable, *locationTable, *optionsTable;
GtkWidget *controlButtonBox, *fullscreenButton, *itemsButton, *locationButton,
  *timeButton, *optionsButton;
GtkWidget *locationLatName, *locationLongName;
GdkPixmap *planetImages[N_SOLAR_SYSTEM_OBJECTS];      /* Used for Sky plots */
GdkPixmap *smallPlanetImages[N_SOLAR_SYSTEM_OBJECTS]; /* Used for Solar System plots */
GdkPixmap *firstPointImage, *meteorRadiantImage[3], *moonImage, *moonImageFlipped;
GdkPixmap *fullMoonImage, *fullMoonImageFlipped, *blueMoonImage, *blueMoonImageFlipped;

GdkPixmap *pixmap = NULL;

/*
  The following structure keeps track of regions on the
  drawingArea you can click the stylus on, to perform actions.
*/
typedef struct sensitiveArea {
  struct sensitiveArea *forwardPointer;
  int type; /* Type determines the action taken */
  float value;
  int bLCX, bLCY, tRCX, tRCY;
} sensitiveArea;

sensitiveArea *sensitiveAreaRoot = NULL;

/* Types of sensitiveAreas: */
#define SA_TOP_AREA            (1)
#define SA_FINGER_PAN_AREA     (2)
#define SA_MONTH_LEFT_ARROW    (3)
#define SA_MONTH_RIGHT_ARROW   (4)
#define SA_SOLAR_SYSTEM_BUTTON (5)

/* Types of deep sky object */
#define SUPERNOVA_REMNENT (0)
#define GLOBULAR_CLUSTER  (1)
#define OPEN_CLUSTER      (2)
#define DIFFUSE_NEBULA    (3)
#define PLANETARY_NEBULA  (4)
#define GALAXY            (5)

typedef struct deepSkyEntry {
  char *name;
  double rA;
  double dec;
  double sinDec;
  double cosDec;
  float mag;
  short type;
  struct deepSkyEntry *next;
} deepSkyEntry;

deepSkyEntry *deepSkyRoot = NULL;
int haveReadDeepSkyObjects = FALSE;

typedef struct objectListEntry {
  struct objectListEntry *forwardPointer;
  int hipNumber;
  double rAJ2000;
  double decJ2000;
  double cosDec;
  double sinDec;
  float mag;
  float color;
  char display;
} objectListEntry;

objectListEntry *objectListRoot = NULL;

typedef struct __attribute__((packed)) hipparcosEntry {
  int number;
  double rAJ2000;
  double decJ2000;
  float parallax;
  float muRA;
  float muDec;
  short mag;
  short color;
} hipparcosEntry;

/* Variable related to star names */
unsigned short nStarNames; /* The number of named stars */
typedef struct __attribute__((packed)) starName {
  unsigned short offset; /* Offset to star name within bigNameString */
  float rARad;
  float decRad;
} starName;
starName *starNames = NULL;
int bigNameStringLength;
int haveStarNames = FALSE;
char *bigNameString = NULL;

/* Menu-related definitions: */
typedef struct locationEntry {
  char *name;
  float latitude, longitude;
  struct locationEntry *next;
} locationEntry;

typedef struct locationClass {
  char *name;
  int key;
  int menuCreated;
  int nLocations;
  struct locationEntry *entry;
  GtkItemFactoryEntry *itemFactoryEntry;
  GtkItemFactory *itemFactory;
  GtkAccelGroup *accelGroup;
  GtkWidget *menu;
  struct locationClass *next;
} locationClass;

locationClass *locationClassRoot = NULL;

/*
  The variables below are used to hold some values used in the low level coordinate
  transformations.   Some of these variables have constant values, but constants
  are not used, in order to avoid unneeded double - > float type conversions,
  etc.
*/

/* These three constants are used in the mercator function: */
float mC5 = 1.0/24.0;
float mC3 = 1.0/6.0;
float mC1 = 1.0;

/* Constants used by thetaEta */
double maxTheta = 97.0*DEGREES_TO_RADIANS;

/* Put the color thresholds into float variables so that a double->float
   conversion is not needed.
*/
float colorThreshold1 = 2.00;
float colorThreshold2 = 0.01;
float colorThreshold3 = 0.40;
float colorThreshold4 = 0.80;
float colorThreshold5 = 1.30;

/* Variables used for touchpad Gestures */

double buttonPressTime, buttonReleaseTime;
struct timespec timeSpecNow;
GdkPoint *userPoly;
int notAPan = FALSE;
gint nUserPoints = 0;
int buttonPressed = FALSE;

int outermostPlanet = JUPITER; /* Furthest planet shown in to-scale SS view */

int gPSDSocketOpen = FALSE;
int gPSDSocket;

double monthLengths[12] = {31.0, 28.0, 31.0, 30.0, 31.0, 30.0, 31.0, 31.0, 30.0, 31.0, 30.0, 31.0};

/* Meteor Shower Stuff */

typedef struct radiantPosition { /*
				   Several of these entries per shower, to
				   track the radiant as it moves on sky
				 */
  short month, day;
  double rA, dec;
} radiantPosition;

typedef struct meteorShower { /* One of these entries per meteor shower */
  char *fullName;
  char threeLetterName[4];
  short startMonth, startDay;
  short endMonth, endDay;
  short maxMonth, maxDay;
  short eclipticLong;
  short radRA, radDec;
  short vInf;
  float r;
  short zHR;
  short nRadiantPositions;
  radiantPosition *radiantPositions;
  struct meteorShower *next;
} meteorShower;

meteorShower *meteorShowers = NULL;
int haveReadMeteorShowers = FALSE;

/* Definitions related to the Azimuth SLew Compass */
#define AZ_COMPASS_RADIUS      (190)     /* Radius of compass circle */
#define AZ_COMPASS_POINT_HEIGHT (20)     /* Height of N, E, S, W pointers */
#define AZ_COMPASS_TICK_LENGTH  (10)     /* Length of 5 degree hash marks */
int azCompassCenterX, azCompassCenterY;  /* Pixel coods of compass center */
float chosenAz;                          /* New center azimuth selected by user */

/*
char *defaultLocale;
*/

/*   E N D   O F   V A R I A B L E   D E C L A R A T I O N S   */

float roundf(float x);

void planetInfo(char *dataDir, int planetNumber, double tJD, double *rA, double *dec, float *F, float *mag);

void calculatePlanetPosition(int planetNumber, double tJD, double *eX, double *eY, double *eZ);
void getMoonPosition(double tJD, double sunELong, double *rA, double *dec, double *eLong, double *eLat, float *F);

void getCurrentOrbitalElements(int planet,         /* Planet we need elements for                  */
                               double tJD,         /* Time for which elements should be calculated */
                               double *a,          /* Semimajor axis                               */
                               double *e,          /* Eccentricity                                 */
                               double *I,          /* Inclination                                  */
                               double *L,          /* Mean longitude                               */
                               double *smallOmega, /* Longitude of perihelion                      */
                               double *bigOmega    /* Longitude of ascending node                  */
                               );

/*
  The following function adds a new sensitive area entry to
  the linked list.
*/
static void addSensitiveArea(int showArea, int type, int x1, int y1, int x2, int y2, float value)
{
  static sensitiveArea *last, *new;
  GdkPoint points[4];

  new = (sensitiveArea *)malloc(sizeof(sensitiveArea));
  if (new == NULL) {
    perror("malloc of new sensitiveArea");
    exit(ERROR_EXIT);
  }
  new->type = type;
  new->bLCX = x1; new->bLCY = y1; new->tRCX = x2; new->tRCY = y2;
  new->value = value;
  new->forwardPointer = NULL;
  if (sensitiveAreaRoot == NULL)
    sensitiveAreaRoot = new;
  else
    last->forwardPointer = new;
  last = new;
  if (showArea) {
    points[0].x = x1; points[0].y = y2;
    points[1].x = x2; points[1].y = y2;
    points[2].x = x2; points[2].y = y1;
    points[3].x = x1; points[3].y = y1;
    gdk_draw_polygon(pixmap, yellowGC, TRUE, points, 4);
  }
}

static void removeAllSensitiveAreas(void)
{
  sensitiveArea *sA = sensitiveAreaRoot;

  while (sA != NULL) {
    sensitiveArea *victim;

    victim = sA;
    sA = victim->forwardPointer;
    free(victim);
  }
  sensitiveAreaRoot = NULL;
}

double buildTJD(int year, int month, int day, int hour, int minute, int second)
{
  int y, m, a, b, c, dd;
  double tJD, d;

  y = 1900+year;
  m = 1+month;
  d = day + ((double)hour)/24.0 + ((double)minute)/1440.0 +
    ((double)second)/86400.0;
  if (m < 3) {
    y -= 1;
    m += 12;
  }
  a = (int)(y/100);
  b = 2 - a + a/4;
  c = (int)(365.25*(double)y);
  dd = (int)(30.6001*(double)(m+1));
  tJD = (double)(b+c+dd) + d + 1720994.5;
  return(tJD);

}

/*
  This function returns the Julian Date
*/

static double getTJD(void)
{
  double tJD;
  time_t t;

  if (useJulianTime) {
    return(julianDate);
  } else if (useCurrentTime) {
    t = time(NULL);
    gMT = gmtime(&t);
    cYear   = gMT->tm_year + 1900;
    cMonth  = gMT->tm_mon + 1;
    cDay    = gMT->tm_mday;
    cHour   = gMT->tm_hour;
    cMinute = gMT->tm_min;
    cSecond = gMT->tm_sec;
  } else {
    if (gMT == NULL) {
      gMT = (struct tm *)malloc(sizeof(struct tm));
      if (gMT == NULL) {
	perror("malloc of gMT");
	exit(ERROR_EXIT);
      }
    }
    gMT->tm_isdst = FALSE;
    gMT->tm_year  = cYear-1900;
    gMT->tm_mon   = cMonth-1;
    gMT->tm_mday  = cDay;
    gMT->tm_hour  = cHour;
    gMT->tm_min   = cMinute;
    gMT->tm_sec   = cSecond;
  }
  tJD = buildTJD(gMT->tm_year, gMT->tm_mon, gMT->tm_mday, gMT->tm_hour, gMT->tm_min, gMT->tm_sec);
  return(tJD);
}

#define N_GPS_SATELLITES_MAX (32)
#define MAX_GPSD_RESPONSE_LENGTH (28 + (N_GPS_SATELLITES_MAX*16))
/*
  Calculate some constants we'll need to update anytime the observer
  changes the position.
*/
static void newPosition(void)
{
  int nSatsUsed = 0;
  int nTuples = 0;
  int nRead, iDummy, used;
  double dummy1, dummy2;
  static char buffer[MAX_GPSD_RESPONSE_LENGTH];
  static struct sockaddr_in gPSDSockAddr;

  if (useGPSD) {
    if (gPSLocation)
      sprintf(satStatString, "(Stale)");
    else
      sprintf(satStatString, "(No GPS Fix)");
    if (!gPSDSocketOpen) {
      gPSDSocket = socket(AF_INET, SOCK_STREAM, 0);
      gPSDSockAddr.sin_family = AF_INET;
      gPSDSockAddr.sin_port = htons(GPSD_PORT);
      gPSDSockAddr.sin_addr.s_addr = inet_addr("127.0.0.1");
      bind(gPSDSocket, (struct sockaddr *)&gPSDSockAddr,
	   sizeof(gPSDSockAddr));
      if (connect(gPSDSocket, (struct sockaddr *)&gPSDSockAddr,
		  sizeof(gPSDSockAddr)) != 0) {
	perror("gpsd connect");
      } else
	gPSDSocketOpen = TRUE;
    }
    if (gPSDSocketOpen) {
      write(gPSDSocket, "y\n", 2);
      nRead = read(gPSDSocket, buffer, MAX_GPSD_RESPONSE_LENGTH);
      buffer[nRead] = (char)0;
      if ((strlen(buffer) > 0) && (strstr(buffer, "?") == NULL)) {
	nRead = sscanf(&buffer[10], "%lf %d:", &dummy1, &nTuples);
	if (nRead == 2) {
	  if (nTuples > 0) {
	    int ii, tuple;
	    char *token;

	    token = strtok(buffer, ":");
	    ii = tuple = 0;
	    while ((token != NULL) && (ii < nTuples)) {
	      token = strtok(NULL, ":");
	      if (token != NULL) {
		nRead= sscanf(token, "%d %d %d %d %d",
			      &iDummy, &iDummy, &iDummy, &iDummy, &used);
		if ((nRead == 5) && used)
		  nSatsUsed++;
	      }
	    }
	  }
	} else
	  nTuples = 0;
      }
      write(gPSDSocket, "p\n", 2);
      nRead = read(gPSDSocket, buffer, MAX_GPSD_RESPONSE_LENGTH);
      buffer[nRead] = (char)0;
      if (strstr(buffer, "?") == NULL) {
	nRead = sscanf(buffer, "GPSD,P=%lf %lf", &dummy1, &dummy2);
	if ((nRead == 2) && (!((dummy1 == 0.0) && (dummy2 == 0.0)))) {
	  latitude = dummy1 * DEGREES_TO_RADIANS;
	  longitude = dummy2 * DEGREES_TO_RADIANS;
	  dprintf("New Fix: %f, %f\n", latitude, longitude);
	  gPSLocation = TRUE;
	  locationChanged = TRUE;
	  if ((nTuples > 0) && (nTuples <= N_GPS_SATELLITES_MAX) &&
	      (nSatsUsed >= 0) && (nSatsUsed <= N_GPS_SATELLITES_MAX))
	    sprintf(satStatString, "(%d/%d)", nTuples, nSatsUsed);
	}
      }
    }
  }
  if (locationChanged) {
    if (latitude > 0.0) {
      sinLatitudeCutoff = sin((latitude - (5.0*DEGREES_TO_RADIANS)) - (M_PI*0.5));
      northernHemisphere = TRUE;
    } else {
      sinLatitudeCutoff = sin((latitude + (5.0*DEGREES_TO_RADIANS)) + (M_PI*0.5));
      northernHemisphere = FALSE;
    }
    if (fabs(latitude) <= (5.0*DEGREES_TO_RADIANS))
      sinLatitudeCutoff = atof("NaN");
    sinLatitude = sin(latitude);
    cosLatitude = cos(latitude);
    locationChanged = FALSE;
  }
}

double gMST(double tJD)
{
  int nDays;
  double tJD0h, s, t, t0;

  tJD0h = (double)((int)tJD) + 0.5;
  s = tJD0h - 2451545.0;
  t = s/36525.0;
  t0 = 6.697374558 +
    (2400.051336 * t) +
    (0.000025862*t*t);
  nDays = (int)(t0/24.0);
  t0 -= 24.0*(double)nDays;
  t0 += (tJD - tJD0h) * 24.0 * 1.002737909;
  while (t0 < 0)
    t0 += 24.0;
  while (t0 >= 24.0)
    t0 -= 24.0;
  return(t0);
}

double lSTAtTJD(double tJD)
{
  double theLST;

  theLST = gMST(tJD);
  theLST *= HOURS_TO_DEGREES;
  theLST += longitude/DEGREES_TO_RADIANS;
  while (theLST >= 360.0)
    theLST -= 360.0;
  while (theLST < 0.0)
    theLST += 360.0;
  theLST *= M_PI/180.0;
  return(theLST);
}

/*
  Return Local Sidereal Time
*/
static double lST(void)
{
  if (needNewTime) {
    newPosition();
    tJD = getTJD();
    myLST = lSTAtTJD(tJD);
    needNewTime = FALSE;
  }
  return(myLST);
}

/*
  refraction(el) returns the refraction correction (which should be added to
the fre-space elevation).   Both the argument and return value are in
radians.   The formula is derived from Saemundsson, Sky & Tel. Vol 72,
page 70 (1982).   It should be accurate to withing 4 arc seconds.
*/
float refraction(float elevation)
{
  float refrac, denom;

  denom = elevation + 8.9186e-2;
  if (denom > 0.0) {
    denom = tanf(elevation + (3.1376e-3/denom));
    if (denom != 0.0)
      refrac = 4.9450995e-6/denom;
    else
      refrac = 0.0;
  } else
    refrac = 0.0;
  /*
  printf("refraction(%f (degrees)) returns %f\"\n", elevation/DEGREES_TO_RADIANS,
	 refrac*3600.0/DEGREES_TO_RADIANS);
  */
  return(refrac);
}

/*
  Convert RA & Dec to altitude and azimuth.

  If abort = FALSE, the function will calculate Az and ZA even if it is clear that
  the object will not appear in the display window.   This is done for the Sun, because
  even if the Sun is not displayable, we most know it's Az and ZA if the Moon illumination
  is to be dosplayed properly.
*/
static int azZA(double ra, double sinDec, double cosDec, double *az, double *zA,
		int abort)
{
  static double localHourAngle, localZA, azimuthT, cosAzimuth, invSinZA, cosHA;

  if (abort) {
    if (northernHemisphere) {
      if (sinDec < sinLatitudeCutoff)
	return(FALSE);
    } else {
      if (sinDec > sinLatitudeCutoff)
	return(FALSE);
    }
  }
  localHourAngle = lST() - ra;
  cosHA = cos(localHourAngle);
  localZA = acos( sinLatitude*sinDec + cosLatitude*cosDec*cosHA );
  if ((localZA >= maximumDisplayZA) && abort)
    return(FALSE);
  else if (localZA == 0.0)
    *az = 0.0;
  else {
    invSinZA = 1.0/sin(localZA);
    azimuthT = -asin(cosDec*sin(localHourAngle)*invSinZA);
    cosAzimuth = (cosLatitude*sinDec - sinLatitude*cosDec*cosHA)*invSinZA;
    
    if (cosAzimuth < 0.0) {
      if (azimuthT < 0.0)
	*az = M_PI - azimuthT;
      else
	*az = acos(cosAzimuth);
    } else if (azimuthT < 0.0)
      *az = M_2PI + azimuthT;
    else
      *az = azimuthT;
  }
  *zA = localZA;
  return(TRUE);
}

/*
  Convert from Az-ZA coordinates to theta-eta coordinates, which are used for the
  Transverse Mercator projection.   Azimuth is first rotated, and then the
  sphere is rotated by 90 degrees, so that Az = 0 is on the equator in the theta-eta
  system.
*/
static int thetaEta(double az, double zA, double *theta, double *eta)
{
  double cosTheta, thetaT, cosEta, localEta, sinZA;

  az -= centerAzD;
  sinZA = sin(zA);
  localEta = asin(sinZA*cos(az));
  if (localEta == 0.0)
    *theta = 0.0;
  else {
    cosEta = cos(localEta);
    thetaT = -asin(sinZA*sin(az)/cosEta);
    cosTheta = cos(zA)/cosEta;
    if (cosTheta < 0.0) {
      if (thetaT < 0.0)
	*theta = piMinusThetaOffset - thetaT;
      else
	*theta = acos(cosTheta) - thetaOffset;
    } else if (thetaT < 0.0)
      *theta = twoPiMinusThetaOffset + thetaT;
    else
      *theta = thetaT - thetaOffset;
  }
  *eta = localEta;
  if (maxTheta < *theta)
    return(FALSE);
  else
    return(TRUE);
}

static int invThetaEta(float theta, float eta, float *az, float *zA)
{
  float cosAz, azT, cosEl, localEl;

  theta += thetaOffset;
  localEl = asinf(cosf(eta)*cosf(theta));
  if (localEl == 90.0)
    *az = 0.0;
  else {
    cosEl = cosf(localEl);
    azT = -asin(cosf(eta)*sin(theta)/cosEl);
    cosAz = sin(eta)/cosEl;
    if (cosAz < 0.0) {
      if (azT < 0.0)
	azT = M_PI - azT;
      else
	azT = acos(cosAz);
    } else if (azT < 0.0)
      azT = M_2PI + azT;
    azT += centerAz;
    if (azT < 0.0)
      azT += M_2PI;
    if (azT > M_2PI)
      azT -= M_2PI;
    *az = azT;
  }
  *zA = 0.5*M_PI - localEl;
  return(TRUE);
}

static void mercatorToPixels(float mX, float mY, int width, int height,
			     int *x, int *y)
{
  *x = (int)((mX+mercatorOffset)*mercatorScale);
  *y = ((int)(mY*mercatorScale))+yBorder+yTopLabelSkip;
}

static void pixelsToMercator(int x, int y, int width, int height,
			     float *mX, float *mY)
{
  *mX = (((float)x)/mercatorScale) - mercatorOffset;
  *mY = ((float)(y - yBorder - yTopLabelSkip))/mercatorScale;
}

/*
  This function performs the Mercator Projection
  The exact mercator projection would be

  *x = logf(tanf(pi*0.25 + phi*0.5));

  For computational efficiency, I'm approximating it by a 5th degree Taylor
  expansion, which should be accurate to better than one pixel on a 480x640
  display.
*/
static void mercator(float phi, float *x)
{
  float phisq;

  phisq = phi*phi;
  *x = ((mC5*phisq + mC3)*phisq + mC1)*phi;
}

static void invMercator(float x, float *phi)
{
  *phi = 2.0*(atanf(expf(x)) - M_PI*0.25);
}

/*
  Take the B-V color index, and return an index to one of the GCs used to
  plot stars in color.
*/
static int colorIndex(float bV, float mag)
{
  if (mag > colorThreshold1)
    return(1);
  else
    if (bV < colorThreshold2)
      return(0);
    else if (bV < colorThreshold3)
      return(1);
    else if (bV < colorThreshold4)
      return(2);
    else if (bV < colorThreshold5)
      return(3);
    else
      return(4);
}

static void inConstellation(objectListEntry *entry) {
  register int i;
  double rSquared;
  double xx, yy, rA, dec;
  constellation *thisConst;

  rA = entry->rAJ2000;
  dec = entry->decJ2000;
  thisConst = constellations;
  while (thisConst != NULL) {
    for (i = 0; i < thisConst->nPoints; i++) {
      if (thisConst->starNeeded[i]) {
	xx = rA - thisConst->rA[i];
	if (xx > M_PI)
	  xx -= M_2PI;
	else if (xx < -M_PI)
	  xx += M_2PI;
	xx *= xx;
	yy = dec - thisConst->dec[i];
	yy *= yy;
	rSquared = xx*xx + yy*yy;
	if (rSquared > 1.0)
	  break;
	else if (rSquared < 1.0e-9) {
	  thisConst->starNeeded[i] = FALSE;
	  entry->display = TRUE;
	  return;
	}
      }
    }
    thisConst = (constellation *)thisConst->next;
  }
  return;
}

void readMeteorShowers(void)
{
  int nRead;
  int readingRadiantLines;
  char fileName[MAX_FILE_NAME_SIZE];
  FILE *data;
  meteorShower *lastShower = NULL;

  sprintf(fileName, "%s/meteorShowers", dataDir);
  setlocale(LC_NUMERIC, "C");
  data = fopen(fileName, "r");
  if (data == NULL) {
    perror(fileName);
    exit(ERROR_EXIT);
  }
  readingRadiantLines = FALSE;
  while (!feof(data)) {
    char fullName[40];
    char threeLetterName[4];
    int startMonth, startDay, i;
    int endMonth, endDay;
    int maxMonth, maxDay;
    int eclipticLong;
    int radRA, radDec;
    int vInf, rMonth, rDay, rRA, rDec;
    float r;
    int zHR;
    int nRadiantPositions;
    meteorShower *thisShower;

    nRead = fscanf(data, "%s %s %d %d %d %d %d %d %d %d %d %d %f %d %d",
		   fullName, threeLetterName,
		   &startMonth, &startDay, &endMonth, &endDay, &maxMonth, &maxDay,
		   &eclipticLong, &radRA, &radDec, &vInf, &r, &zHR, &nRadiantPositions);
    if (nRead == 15) {
      thisShower = (meteorShower *)malloc(sizeof(meteorShower));
      if (thisShower == NULL) {
	perror("thisShower malloc");
	exit(ERROR_EXIT);
      }
      for (i = 0; i < strlen(fullName); i++)
	if (fullName[i] == '_')
	  fullName[i] = ' ';
      thisShower->fullName = malloc(strlen(fullName)+1);
      if (thisShower->fullName == NULL) {
	perror("malloc of meteor shower fullName");
	exit(ERROR_EXIT);
      }
      strcpy(thisShower->fullName, fullName);
      strcpy(thisShower->threeLetterName, threeLetterName);
      thisShower->startMonth = startMonth; thisShower->startDay = startDay;
      thisShower->endMonth   = endMonth;   thisShower->endDay   = endDay;
      thisShower->maxMonth   = maxMonth;   thisShower->maxDay   = maxDay;
      thisShower->eclipticLong = eclipticLong;
      thisShower->radRA = radRA; thisShower->radDec = radDec;
      thisShower->vInf = vInf; thisShower->r = r; thisShower->zHR = zHR;
      thisShower->nRadiantPositions = nRadiantPositions;
      thisShower->radiantPositions = (radiantPosition *)malloc(nRadiantPositions *
							       sizeof(radiantPosition));
      if (thisShower->radiantPositions == NULL) {
	perror("malloc of radiantPositions");
	exit(ERROR_EXIT);
      }
      for (i = 0; i < nRadiantPositions; i++) {
	nRead = fscanf(data, "%d %d %d %d", &rMonth, &rDay, &rRA, &rDec);
	if (nRead == 4) {
	  thisShower->radiantPositions[i].month = rMonth;
	  thisShower->radiantPositions[i].day   = rDay;
	  thisShower->radiantPositions[i].rA    = (double)rRA * DEGREES_TO_RADIANS;
	  thisShower->radiantPositions[i].dec   = (double)rDec * DEGREES_TO_RADIANS;
	} else {
	  fprintf(stderr, "Wrong number of parameters on meteor radiant line");
	  exit(ERROR_EXIT);
	}
      }
      thisShower->next = NULL;
      if (lastShower == NULL)
	meteorShowers = thisShower;
      else
	lastShower->next = thisShower;
      lastShower = thisShower;
    }
  }
  fclose(data);
}

/*
  The following function reads the individual entries in a star catalog,
  and puts them in a linked list.
*/

void readStarCatalog(int fD)
{
  static int firstCall = TRUE;
  int nRead;
  int nStarsRead = 0;
  hipparcosEntry entry;
  objectListEntry *lastEntry = NULL;
  objectListEntry *newEntry = NULL;

  if (!firstCall) {
    /*
      If this function has been called before, the first step is to free
      up the memory allocated during the previous call.
    */
    free(darkGreyPoints);
    free(greyPoints);
    free(whitePoints);
    newEntry = objectListRoot;
    while (newEntry != NULL) {
      lastEntry = newEntry;
      newEntry = newEntry->forwardPointer;
      free(lastEntry);
    }
    objectListRoot = NULL;
  }
  firstCall = nDarkGreyPoints = nGreyPoints = nWhitePoints = 0;
  faintestStar = -30.0;
  do {
    nRead = read(fD, &entry, sizeof(hipparcosEntry));
    if (nRead == sizeof(hipparcosEntry)) {
      /* Allocate a new entry to store this star */
      newEntry = (objectListEntry *)malloc(sizeof(objectListEntry));
      if (newEntry == NULL) {
	perror("allocating new object entry");
	exit(ERROR_EXIT);
      }
      /* Insert the new entry into the doubly linked list of objects */
      if (objectListRoot == NULL)
	/* First entry */
	objectListRoot = newEntry;
      else
	lastEntry->forwardPointer = newEntry;
      newEntry->forwardPointer = NULL;
      newEntry->hipNumber      = entry.number;
      newEntry->rAJ2000        = entry.rAJ2000;
      newEntry->decJ2000       = entry.decJ2000;
      newEntry->sinDec         = sin(entry.decJ2000);
      newEntry->cosDec         = cos(entry.decJ2000);
      newEntry->mag            = ((float)entry.mag)/100.0;
      if (newEntry->mag > faintestStar)
	faintestStar = newEntry->mag;
      newEntry->color          = ((float)entry.color)/1000.0;
      newEntry->display        = FALSE;
      lastEntry = newEntry;
      nStarsRead++;
    }
  } while (nRead == sizeof(hipparcosEntry));
  newEntry = objectListRoot;
  while (newEntry != NULL) {
    if (newEntry->mag*magScale > DARK_GREY_LIMIT)
      nDarkGreyPoints++;
    else if (newEntry->mag*magScale > GREY_LIMIT)
      nGreyPoints++;
    else if (newEntry->mag*magScale > WHITE_LIMIT)
      nWhitePoints++;
    newEntry = newEntry->forwardPointer;
  }
  /*
  printf("Read %d stars from the Hipparcos catalog\n", nStarsRead);
  */
  darkGreyPoints = (GdkPoint *)malloc(nDarkGreyPoints*sizeof(GdkPoint));
  if (darkGreyPoints == NULL) {
    perror("malloc of darkGreyPoints");
    exit(ERROR_EXIT);
  }
  greyPoints = (GdkPoint *)malloc(nGreyPoints*sizeof(GdkPoint));
  if (greyPoints == NULL) {
    perror("malloc of greyPoints");
    exit(ERROR_EXIT);
  }
  whitePoints = (GdkPoint *)malloc(nWhitePoints*sizeof(GdkPoint));
  if (whitePoints == NULL) {
    perror("malloc of whitePoints");
    exit(ERROR_EXIT);
  }
  close(fD);
}

/*
  This function plots the objects outside our solar system.
*/
static void plotFixedObjects(void)
{
  int mercXLow, mercXHigh, mercYLow, mercYHigh;
  int x, y, fD;
  float mag, color, darkGreyLimit, greyLimit, whiteLimit;
  float mX, mY;
  double az, zA, theta, eta;
  char fileName[MAX_FILE_NAME_SIZE];
  objectListEntry *currentEntry;
  
  if (haveFaintStars && showFaintStars && (!showingFaintStars)) {
    /*
      The faint star catalog is available, and should be show, but it is
      not currently being shown, so we should switch to the faint star
      catalog.
    */
    sprintf(fileName, "%s/faintStars", dataDir);
    fD = open(fileName, O_RDONLY);
    if (fD >= 0)
      readStarCatalog(fD);
    showingFaintStars = TRUE;
  } else if (showingFaintStars && (!showFaintStars)) {
    /*
      The faint star catalog is being used, but it should not be, so we
      must switch back to the default (naked-eye stars) catalog.
    */
    sprintf(fileName, "%s/hipparcos.dat", dataDir);
    fD = open(fileName, O_RDONLY);
    if (fD >= 0)
      readStarCatalog(fD);
    showingFaintStars = FALSE;
  }
  nDarkGreyPoints = nGreyPoints = nWhitePoints = 0;
  darkGreyLimit = DARK_GREY_LIMIT/magScale;
  greyLimit = GREY_LIMIT/magScale;
  whiteLimit = WHITE_LIMIT/magScale;
  mercXLow  = xBorder+xLeftLabelSkip;
  mercXHigh = xBorder+xLeftLabelSkip+plotWidth;
  mercYLow  = yBorder+yTopLabelSkip;
  mercYHigh = displayHeight;
  currentEntry = objectListRoot;
  while (currentEntry != NULL) {
    mag = currentEntry->mag;
    color = currentEntry->color;
    if ((mag < limitingMag) || (displayConstellations && (currentEntry->display))) {
      /* The object is bright enough to be displayed */
      if (azZA(currentEntry->rAJ2000, currentEntry->sinDec,
	       currentEntry->cosDec, &az, &zA, TRUE))
	if (thetaEta(az, zA, &theta, &eta)) {
	  mY = (float)theta;
	  mercator((float)eta, &mX);
	  mercatorToPixels(mX, mY, plotWidth, plotHeight, &x, &y);
	  if ((x > mercXLow) && (x < mercXHigh) && (y > mercYLow) && (y < mercYHigh)) {
	    if (mag > darkGreyLimit) {
	      darkGreyPoints[nDarkGreyPoints].x = x;
	      darkGreyPoints[nDarkGreyPoints++].y = y;
	    } else if (mag > greyLimit) {
	      greyPoints[nGreyPoints].x = x;
	      greyPoints[nGreyPoints++].y = y;
	    } else if (mag > whiteLimit) {
	      whitePoints[nWhitePoints].x = x;
	      whitePoints[nWhitePoints++].y = y;
	    } else {
	      int index;
	      
	      index = (int)(2.0*(mag*magScale + 1.0));
	      if (index < 0)
		index = 0;
	      if (index <= 11)
		gdk_draw_arc(pixmap, starGC[colorIndex(color, mag)], TRUE,
			     x-(starSize[index]/2), y-(starSize[index]/2),
			     starSize[index], starSize[index], 0, FULL_CIRCLE);
	      else {
		gdk_draw_point(pixmap, whiteGC, x, y);
		gdk_draw_point(pixmap, whiteGC, x+1, y);
	      }
	    }
	  }
	}
    }
    currentEntry = currentEntry->forwardPointer;
  }
  GdkPixbuf* starDrawingBuf = gdk_pixbuf_get_from_drawable( NULL,
                                                            pixmap,
                                                            gdk_colormap_get_system()
                                                            , 0, 0, 0, 0, displayWidth, displayHeight);
  g_assert (gdk_pixbuf_get_bits_per_sample (starDrawingBuf) == 8);
  guchar* p;
  int rowstride = gdk_pixbuf_get_rowstride (starDrawingBuf);
  guchar* pixels = gdk_pixbuf_get_pixels (starDrawingBuf);
  int n_channels = gdk_pixbuf_get_n_channels (starDrawingBuf);
  GdkGCValues starGCval;
  GdkColor starColor;
  int pointNum;
  gdk_gc_get_values(darkGreyGC, &starGCval);
  gdk_colormap_query_color( gdk_gc_get_colormap(darkGreyGC),starGCval.foreground.pixel, &starColor );
  for( pointNum = 0; pointNum < nDarkGreyPoints; pointNum++) {
    p = pixels + darkGreyPoints[pointNum].y * rowstride + darkGreyPoints[pointNum].x * n_channels;
    p[0] = starColor.red & 0xff;
    p[1] = starColor.green & 0xff;
    p[2] = starColor.blue & 0xff;
  }
  gdk_gc_get_values(greyGC, &starGCval);
  gdk_colormap_query_color( gdk_gc_get_colormap(greyGC), starGCval.foreground.pixel, &starColor );
  for( pointNum = 0; pointNum < nGreyPoints; pointNum++) {
    p = pixels + greyPoints[pointNum].y * rowstride + greyPoints[pointNum].x * n_channels;
    p[0] = starColor.red & 0xff;
    p[1] = starColor.green & 0xff;
    p[2] = starColor.blue & 0xff;
  }
  gdk_gc_get_values(whiteGC, &starGCval);
  gdk_colormap_query_color( gdk_gc_get_colormap(whiteGC), starGCval.foreground.pixel, &starColor );
  for( pointNum = 0; pointNum < nWhitePoints; pointNum++) {
    p = pixels + whitePoints[pointNum].y * rowstride + whitePoints[pointNum].x * n_channels;
    p[0] = starColor.red & 0xff;
    p[1] = starColor.green & 0xff;
    p[2] = starColor.blue & 0xff;
  }
  gdk_draw_pixbuf ( pixmap , NULL , starDrawingBuf,
            0, 0, 0, 0, displayWidth, displayHeight, GDK_RGB_DITHER_NORMAL, 0, 0 ) ;
}

void makeTimeString(char *string)
{
  int sHH, sMM, sSS;
  float sidereal;

  sidereal = lST()/HOURS_TO_RADIANS;
  sHH = (int)sidereal;
  sMM = (int)(60.0*(sidereal-(float)sHH));
  sSS = (int)(3600.0*(sidereal-(float)sHH - ((float)sMM)/60.0));
  if (useJulianTime)
    sprintf(string, "TJD %11.3f  LST %02d:%02d:%02d",
	    julianDate, sHH, sMM, sSS);
  else if (useCurrentTime)
    sprintf(string, "UT %02d:%02d:%02d  LST %02d:%02d:%02d",
	    gMT->tm_hour, gMT->tm_min, (int)gMT->tm_sec,
	    sHH, sMM, sSS);
  else {
    float lYear;

    if (cYear > 0)
      lYear = cYear;
    else
      lYear = cYear-1.0;
    sprintf(string, "%02.0f/%02.0f/%4.0f  UT %02d:%02d:%02d  LST %02d:%02d:%02d",
	    cMonth, cDay, lYear,
	    gMT->tm_hour, gMT->tm_min, (int)gMT->tm_sec,
	    sHH, sMM, sSS);
  }
}

/*
  Print the UT and LST at the top of the plot.
*/
static void showTime(void)
{
  char labelString[100];

  makeTimeString(labelString);
  if (useJulianTime)
    gdk_draw_string(pixmap, smallFont, redGC, 0,
		    yTopLabelSkip+3, labelString);
  else if (useCurrentTime)
    gdk_draw_string(pixmap, smallFont, whiteGC, 0,
		    yTopLabelSkip+3, labelString);
  else
    gdk_draw_string(pixmap, smallFont, redGC, 0,
		    yTopLabelSkip+3, labelString);
  if (gPSLocation && useGPSD) {
    int gLatDD, gLatMM, gLongDD, gLongMM;
    double gLatSS, gLongSS, gLatDeg, gLongDeg;

    gLatDeg = latitude/DEGREES_TO_RADIANS;
    gLatDD = (int)gLatDeg;
    gLatMM = (int)((gLatDeg-(double)gLatDD)*60.0);
    gLatSS = (gLatDeg - (double)gLatDD - ((double)gLatMM)/60.0)*3600.0;
    if (latitude < 0.0) {
      gLatMM *= -1;
      gLatSS *= -1.0;
    }
    gLongDeg = longitude/DEGREES_TO_RADIANS;
    gLongDD = (int)gLongDeg;
    gLongMM = (int)((gLongDeg-(double)gLongDD)*60.0);
    gLongSS = (gLongDeg - (double)gLongDD - ((double)gLongMM)/60.0)*3600.0;
    if (longitude < 0.0) {
      gLongMM *= -1;
      gLongSS *= -1.0;
    }
    sprintf(labelString, "  GPS Lat %02d:%02d:%02.0f Lon %02d:%02d:%02.0f %s",
	    gLatDD, gLatMM, gLatSS, gLongDD, gLongMM, gLongSS, satStatString);
  } else if (useGPSD)
    sprintf(labelString, "%s %s",
	    locationName, satStatString);
  else
    sprintf(labelString, "%s",
	    locationName);
  gdk_draw_string(pixmap, smallFont, whiteGC, displayWidth/2,
		  yTopLabelSkip+3, labelString);
}

/*
  Return the RA and Dec of a pair of ecliptic coordinates (in radians).
*/
static void ecliptic(float lambda, double *ra, double *dec, double *cDec)

#define ETA 23.4378527778*DEGREES_TO_RADIANS /* Earth axis inclination */
{
  static int   firstCall = TRUE;
  static float sinEta, cosEta, cosDec;
  float tRA, tDec, cosRA, sinRA;
  
  if (firstCall) {
    sinEta = sinf(ETA);
    cosEta = cosf(ETA);
    firstCall = FALSE;
  }
  tDec = asinf(sinf(lambda)*sinEta);
  cosDec = cosf(tDec);
  if (cosDec != 0.0) {
    cosRA = cosf(lambda)/cosDec;
    sinRA = cosEta*sinf(lambda)/cosDec;
  } else {
    cosRA = -1.0;
    sinRA = 0.0;
  }
  tRA = atan2f(sinRA, cosRA);
  *ra = (double)tRA;
  *dec = (double)tDec;
  *cDec = (double)cosDec;
}

/*
  Return the RA and Dec of a pair of galactic coordinates (in radians).
*/
static void galactic(float lambda, double *ra, double *dec, double *cDec)

#define ETAG 62.3*DEGREES_TO_RADIANS
{
  static int   firstCall = TRUE;
  static float sinEta, cosEta, cosDec, rAOffset, lambdaOffset;
  float tRA, tDec, cosRA, sinRA;
  
  if (firstCall) {
    sinEta = sinf(ETAG);
    cosEta = cosf(ETAG);
    rAOffset = 282.25*DEGREES_TO_RADIANS;
    lambdaOffset = -33.0*DEGREES_TO_RADIANS;
    firstCall = FALSE;
  }
  lambda += lambdaOffset;
  tDec = asinf(sinf(lambda)*sinEta);
  cosDec = cosf(tDec);
  if (cosDec != 0.0) {
    cosRA = cosf(lambda)/cosDec;
    sinRA = cosEta*sinf(lambda)/cosDec;
  } else {
    cosRA = -1.0;
    sinRA = 0.0;
  }
  tRA = atan2f(sinRA, cosRA);
  *ra = (double)tRA + rAOffset;
  *dec = (double)tDec;
  *cDec = cosDec;
}

/*
  Draw a red line depicting the ecliptic plane.
*/
static void plotEcliptic(void)
#define N_ECLIPTIC_STEPS 360
{
  static int firstCall = TRUE;
  static float eclipticStep;
  int i, targetX, targetY;
  int offset = 0;
  float mX, mY;
  float lambda = 0.0;
  double ra, dec, cDec, za, az;
  double eta, theta;
  GdkPoint eclipticPoints[N_ECLIPTIC_STEPS+1];
                                                                                                            
  if (firstCall) {
    eclipticStep = M_2PI/(double)N_ECLIPTIC_STEPS;
    firstCall = FALSE;
  }
  for (i = 0; i < N_ECLIPTIC_STEPS+1; i++) {
    ecliptic(lambda, &ra, &dec, &cDec);
    if (azZA(ra, sin(dec), cDec, &az, &za, TRUE))
      if (thetaEta(az, za, &theta, &eta)) {
	mY = (float)theta;
	mercator((float)eta, &mX);
	mercatorToPixels(mX, mY, plotWidth, plotHeight,
			 &targetX, &targetY);
	if ((targetX > xBorder+xLeftLabelSkip)
	    && (targetX < xBorder+xLeftLabelSkip+plotWidth)
	    && (targetY > yBorder+yTopLabelSkip)
	    && (targetY < displayHeight)) {
	  if ((i % 5) == 0)
	    gdk_draw_arc(pixmap, redGC, TRUE,
			 targetX-1, targetY-1,
			 3, 3, 0, FULL_CIRCLE);
	  eclipticPoints[offset].x = targetX;
	  eclipticPoints[offset].y = targetY;
	  offset++;
	}
      }
    lambda += eclipticStep;
  }
  if (offset > 1)
    gdk_draw_points(pixmap, redGC, eclipticPoints, offset);
}

/*
  Draw a green line depicting the galactic plane.
*/
static void plotGalactic(void)
#define N_GALACTIC_STEPS 360
{
  static int firstCall = TRUE;
  static float galacticStep;
  int i, targetX, targetY;
  int offset = 0;
  float mX, mY;
  float lambda = 0.0;
  double ra, dec, cDec, za, az;
  double eta, theta;
  GdkPoint galacticPoints[N_GALACTIC_STEPS+1];
                                                                                                            
  if (firstCall) {
    galacticStep = M_2PI/(double)N_GALACTIC_STEPS;
    firstCall = FALSE;
  }
  for (i = 0; i < N_GALACTIC_STEPS+1; i++) {
    galactic(lambda, &ra, &dec, &cDec);
    if (azZA(ra, sin(dec), cDec, &az, &za, TRUE))
      if (thetaEta(az, za, &theta, &eta)) {
	mY = (float)theta;
	mercator((float)eta, &mX);
	mercatorToPixels(mX, mY, plotWidth, plotHeight,
			 &targetX, &targetY);
	if ((targetX > xBorder+xLeftLabelSkip)
	    && (targetX < xBorder+xLeftLabelSkip+plotWidth)
	    && (targetY > yBorder+yTopLabelSkip)
	    && (targetY < displayHeight)) {
	  if ((i % 5) == 0)
	    gdk_draw_arc(pixmap, blueGreenGC, TRUE,
			 targetX-1, targetY-1,
			 3, 3, 0, FULL_CIRCLE);
	  galacticPoints[offset].x = targetX;
	  galacticPoints[offset].y = targetY;
	  offset++;
	}
      }
    lambda += galacticStep;
  }
  if (offset > 1)
    gdk_draw_points(pixmap, blueGreenGC, galacticPoints, offset);
}

/*
  Draw a grey line depicting the celestial equator.
*/
static void plotEquator(void)
#define N_EQUATOR_STEPS 360
{
  static int firstCall = TRUE;
  static double equatorStep;
  int i, targetX, targetY;
  int offset = 0;
  float mX, mY;
  double za, az;
  double eta, theta;
  double lambda = 0.0;
  GdkPoint equatorPoints[N_EQUATOR_STEPS+1];
                                                                                                            
  if (firstCall) {
    equatorStep = M_2PI/(double)N_EQUATOR_STEPS;
    firstCall = FALSE;
  }
  for (i = 0; i < N_EQUATOR_STEPS+1; i++) {
    if (azZA(lambda, 0.0, 1.0, &az, &za, TRUE))
      if (thetaEta(az, za, &theta, &eta)) {
	mY = (float)theta;
	mercator((float)eta, &mX);
	mercatorToPixels(mX, mY, plotWidth, plotHeight,
			 &targetX, &targetY);
	if ((targetX > xBorder+xLeftLabelSkip)
	    && (targetX < xBorder+xLeftLabelSkip+plotWidth)
	    && (targetY > yBorder+yTopLabelSkip)
	    && (targetY < displayHeight)) {
	  if ((i % 5) == 0)
	    gdk_draw_arc(pixmap, equatorGC, TRUE,
			 targetX-1, targetY-1,
			 3, 3, 0, FULL_CIRCLE);
	  equatorPoints[offset].x = targetX;
	  equatorPoints[offset].y = targetY;
	  offset++;
	}
      }
    lambda += equatorStep;
  }
  if (offset > 1)
    gdk_draw_points(pixmap, equatorGC, equatorPoints, offset);
}

/*
  Calculate the angle kappa though which the moon's terminator swings
  (actually 1/2 that angle is returned).
*/
static float terminatorRadius(float P)
{
  float appr;

  appr =
    1.5709                                +
   -1.57949   * P                         +
   -1.70266   * P * P                     +
   -5.049     * P * P * P                 +
   15.7448    * P * P * P * P             +
  -59.3318    * P * P * P * P * P         +
   96.033     * P * P * P * P * P * P     +
  -45.3849    * P * P * P * P * P * P * P;
  return(appr);
}

/*
  Calculate the direction of destination, when seen from source.   Used
  to rotate the moon drawing so that it properly faces the sun.
*/
static float greatCircleDirection(float sLong, float sLat, float dLong, float dLat)
{
  return(-atan2f(sinf(sLong-dLong),cosf(sLat)*tanf(dLat) - sinf(sLat)*cosf(sLong-dLong)));
}

/*
  Draw the moon, with the proper phase and orientation.
*/
static void drawMoon(float moonSize, float illum, float sunAngle, int x, int y,
		     int image, int full, int blue)
{
  int iMoonSize;
  int nPoints = 0;
  float angle = 90.0;
  float phaseSign; /* Controls whether a cresent or gibbous moon is drawn */
  float kappa, rad, offset;
  float angleStep = 5.0;
  GdkPoint points[104];

  dprintf("In draw moon, illum = %f, center (%d,%d), angle = %f\n",
	 illum, x, y, sunAngle/DEGREES_TO_RADIANS);
  iMoonSize = (int)(moonSize*2.0);
  if (image) {
    int w, h;

    gdk_drawable_get_size(moonImage, &w, &h);
    if (northernHemisphere) {
      if (full) {
	if (blue)
	  gdk_draw_drawable(pixmap, blueGC, blueMoonImage, 0, 0, x - w/2, y - h/2, w, h);
	else
	  gdk_draw_drawable(pixmap, blueGC, fullMoonImage, 0, 0, x - w/2, y - h/2, w, h);
      } else
	gdk_draw_drawable(pixmap, blueGC, moonImage, 0, 0, x - w/2, y - h/2, w, h);
    } else {
      if (full) {
	if (blue)
	  gdk_draw_drawable(pixmap, blueGC, blueMoonImageFlipped, 0, 0, x - w/2, y - h/2, w, h);
	else
	  gdk_draw_drawable(pixmap, blueGC, fullMoonImageFlipped, 0, 0, x - w/2, y - h/2, w, h);
      } else
	gdk_draw_drawable(pixmap, blueGC, moonImageFlipped, 0, 0, x - w/2, y - h/2, w, h);
    }
    illum = 1.0 - illum;
    sunAngle = M_PI + sunAngle;
    moonSize = w/2 + 1.075;
  }
  if ((illum > 0.98) && (!image)) {
    /* The moon is so close to full, just draw a circle) */
    if (blue && full)
      gdk_draw_arc(pixmap, blueGreenGC, TRUE,
		   x-(iMoonSize/2), y-(iMoonSize/2),
		   iMoonSize, iMoonSize, 0, FULL_CIRCLE);
    else
      gdk_draw_arc(pixmap, whiteGC, TRUE,
		   x-(iMoonSize/2), y-(iMoonSize/2),
		   iMoonSize, iMoonSize, 0, FULL_CIRCLE);
  } else {
    /* First draw a black circle of the right diameter to obscure backroung stars */
    if (!image) {
      if (illum > 0.10)
	gdk_draw_arc(pixmap, drawingArea->style->black_gc, TRUE,
		     x-(iMoonSize/2), y-(iMoonSize/2),
		     iMoonSize, iMoonSize, 0, FULL_CIRCLE);
      else
	/* Draw Earthshine */
	gdk_draw_arc(pixmap, darkGreyGC, TRUE,
		     x-(iMoonSize/2), y-(iMoonSize/2),
		     iMoonSize, iMoonSize, 0, FULL_CIRCLE);
    }
    if (illum > 0.02) {
      float dx, dy, sRot, cRot;

      sRot = sinf(sunAngle);
      cRot = cosf(sunAngle);
      angle = -90.0;
      while (angle <= 90.0) {
	dx = moonSize*cosf(angle*DEGREES_TO_RADIANS);
	dy = moonSize*sinf(angle*DEGREES_TO_RADIANS);
	points[nPoints].x = x + sRot*dy + cRot*dx;
	points[nPoints].y = y + cRot*dy - sRot*dx;
	nPoints++;
	angle += angleStep;
      }
      if ((illum < 0.47) || (illum > 0.53)) {
	/* If the illumination is very close to 50%, just draw a semicircle */
	if (illum < 0.5) {
	  phaseSign = 1.0;
	  kappa = terminatorRadius(illum);
	} else {
	  phaseSign = -1.0;
	  kappa = terminatorRadius(1.0-illum);
	}
	rad = 1.0 / sinf(kappa);
	angleStep = kappa/15.0;
	angle = kappa;
	offset = phaseSign*rad*cosf(kappa);
	dprintf("rad = %f, kappa = %f, offset = %f, step = %f\n", rad, kappa, offset, angleStep);
	while (angle > -kappa) {
	  dx = moonSize*(-offset + phaseSign*rad*cosf(angle));
	  dy = moonSize*(rad*sinf(angle));
	  points[nPoints].x = x + sRot*dy + cRot*dx;
	  points[nPoints].y = y + cRot*dy - sRot*dx;
	  nPoints++;
	  angle -= angleStep;
	}
      }
      if (image)
	gdk_draw_polygon(pixmap, blackGC, TRUE, points, nPoints);
      else if (blue && full)
	gdk_draw_polygon(pixmap, blueGreenGC, TRUE, points, nPoints);
      else
	gdk_draw_polygon(pixmap, whiteGC, TRUE, points, nPoints);
    }
  }
}

static void plotStarNames(void)
{
  int i, up, x, y, stringWidth;
  float mX, mY;
  double az, zA, eta, theta;

  for (i = 0; i < nStarNames; i++) {
    up = azZA((double)(starNames[i].rARad),
	      sin((double)(starNames[i].decRad)),
	      cos((double)(starNames[i].decRad)), &az, &zA, TRUE);
    if (up) {
      if (thetaEta(az, zA, &theta, &eta)) {
	mY = (float)theta;
	mercator((float)eta, &mX);
	mercatorToPixels(mX, mY, plotWidth, plotHeight,
			 &x, &y);
	y += STAR_NAME_V_OFFSET;
	stringWidth = gdk_string_width(smallFont, &bigNameString[starNames[i].offset])/2;
	if ((x >= 0) && (x <= displayWidth) &&
	    (y >= 0) && (y <= displayHeight))
	  gdk_draw_string(pixmap, smallFont, greenGC,
			  x-stringWidth, y, &bigNameString[starNames[i].offset]);
      }
    }
  }
}

static void plotPlanets(void)
{
  static int firstCall = TRUE;
  int i, stringWidth;
  int sunSize = 26;
  float illum, mag;
  double rA, dec;
  double sunRA, sunDec;
  char fileName[MAX_FILE_NAME_SIZE];

  planetInfo(dataDir, EARTH, tJD, &sunRA, &sunDec, &illum, &mag);
  if (firstCall) {
    sprintf(fileName, "%s/icons/%s.xpm", dataDir, "firstPointInAries");
    firstPointImage = gdk_pixmap_create_from_xpm(pixmap,
						 NULL, NULL, fileName);
    sprintf(fileName, "%s/icons/%s.xpm", dataDir, "MeteorRadiant");
    meteorRadiantImage[0] = gdk_pixmap_create_from_xpm(pixmap,
						       NULL, NULL, fileName);
    sprintf(fileName, "%s/icons/%s.xpm", dataDir, "WeakMeteorRadiant");
    meteorRadiantImage[1] = gdk_pixmap_create_from_xpm(pixmap,
						       NULL, NULL, fileName);
    sprintf(fileName, "%s/icons/%s.xpm", dataDir, "WeakestMeteorRadiant");
    meteorRadiantImage[2] = gdk_pixmap_create_from_xpm(pixmap,
						       NULL, NULL, fileName);
  }
  for (i = 0; i < N_SOLAR_SYSTEM_OBJECTS; i++) {
    if (firstCall) {
      sprintf(fileName, "%s/icons/%s.xpm", dataDir, solarSystemNames[i]);
      planetImages[i] = gdk_pixmap_create_from_xpm(pixmap,
						   NULL, NULL, fileName);
    }
    planetGC = NULL;
    switch (i) {
    case SUN:
      planetGC = lightYellowGC; break;
    case MERCURY:
    case VENUS:
    case MOON:
      planetGC = whiteGC;       break;
    case MARS:
      planetGC = starGC[4];     break;
    case JUPITER:
    case SATURN:
      planetGC = creamGC;       break;
    case URANUS:
    case NEPTUNE:
      planetGC = blueGreenGC;   break;
    }
    if (planetGC != NULL) {
      if (i == SUN) {
	rA = sunRA;
	dec = sunDec;
	illum = 1.0;
	mag = -26.8;
      } else
	planetInfo(dataDir, i, tJD, &rA, &dec, &illum, &mag);
      if (mag <= limitingMag) {
	int x, y, up;
	float mX, mY;
	static float sunAz, sunZA;
	double az, zA, eta, theta;
	
	if (i == SUN) {
	  up = azZA(rA, sin(dec), cos(dec), &az, &zA, FALSE);
	  sunAz = (float)az;
	  sunZA = (float)zA;
	} else
	  up = azZA(rA, sin(dec), cos(dec), &az, &zA, TRUE);
	if (up) {
	  if (thetaEta(az, zA, &theta, &eta)) {
	    mY = (float)theta;
	    mercator((float)eta, &mX);
	    mercatorToPixels(mX, mY, plotWidth, plotHeight,
			     &x, &y);
	    if ((x > xBorder+xLeftLabelSkip)
		&& (x < xBorder+xLeftLabelSkip+plotWidth)
		&& (y > yBorder+yTopLabelSkip)
		&& (y < displayHeight)) {
	      if (!displayPlanetsAsSymbols) {
		if (i == SUN) {
		  gdk_draw_arc(pixmap, planetGC, TRUE,
			       x-(sunSize/2), y-(sunSize/2),
			       sunSize, sunSize, 0, FULL_CIRCLE);
		} else if (i == MOON) {
		  float sunAngle;
		  
		  sunAngle = greatCircleDirection((float)az, (float)((M_PI*0.5)-zA),
						  sunAz, (M_PI*0.5)-sunZA);
		  drawMoon(13.0, illum, (M_PI*0.5)-sunAngle, x, y, FALSE, FALSE, FALSE);
		} else {
		  if (mag > DARK_GREY_LIMIT)
		    gdk_draw_point(pixmap, darkGreyGC, x, y);
		  else if (mag > GREY_LIMIT)
		    gdk_draw_point(pixmap, greyGC, x, y);
		  else if (mag > WHITE_LIMIT)
		    gdk_draw_point(pixmap, planetGC, x, y);
		  else {
		    int index;
		    
		    index = 2*(int)(mag + 1.5);
		    if (index < 0)
		      index = 0;
		    gdk_draw_arc(pixmap, planetGC, TRUE,
				 x-(starSize[index]/2), y-(starSize[index]/2),
				 starSize[index], starSize[index], 0,
				 FULL_CIRCLE);
		  }
		}
	      } else {
		int w, h;
		
		gdk_drawable_get_size(planetImages[i], &w, &h);
		gdk_draw_drawable(pixmap, blueGC, planetImages[i], 0, 0,
				  x-w/2, y-h/2, 20, 20);
	      }
	      if (showNames && (!displayPlanetsAsSymbols) && (i != SUN) && (i != MOON)) {
		int index;
		char *displayName;
		
		displayName = malloc(strlen(solarSystemNames[i])+1);
		strcpy(displayName, solarSystemNames[i]);
		y += STAR_NAME_V_OFFSET;
		index = 2*(int)(mag + 1.5);
		if (index < 0)
		  index = 0;
		y += starSize[index] / 4;
		stringWidth = gdk_string_width(smallFont, displayName)/2;
		if ((x >= 0) && (x <= displayWidth) &&
		    (y >= 0) && (y <= displayHeight))
		  gdk_draw_string(pixmap, smallFont, planetGC,
				  x-stringWidth, y, displayName);
		free(displayName);
	      }
	    }
	  }
	}
      }
    }
  }
  firstCall = FALSE;
}

/*
  Calculate parameters used by other routines when producing
  a transverse mercator projection.
 */
static void calculateTransverseMercatorParameters(void)
{
  int boxWidth, boxHeight;

  boxWidth   = displayWidth-(2*xBorder)-xLeftLabelSkip;
  boxHeight  = displayHeight-(2*yBorder)-yBottomLabelSkip;
  plotWidth  = boxWidth - 2;
  plotHeight = boxHeight - 2;
  mercatorScale = plotHeight / (maximumZA - minimumZA);
  mercatorOffset = plotWidth / (2.0*mercatorScale);
}

/*
  This function draws the bounding box, and optionally the horizon,
  in Transverse Mercator projection mode.
*/
static void  drawTransverseMercatorBorder(void)
{
  int x1, x2, y1, y2, stringWidth, iStartAz;
  float mX, mY;
  double az, zA, theta, eta, startAz, endAz;

  if (zoomed) {
    char *zoomed = "ZOOMED";

    gdk_draw_string(pixmap, smallFont, greyGC,
		    displayWidth-gdk_string_width(smallFont, zoomed)-10,
		    yTopLabelSkip+gdk_string_height(smallFont, zoomed)+10,
		    zoomed);
  }
  startAz = (centerAz - (M_PI+azSpan)*0.5)/DEGREES_TO_RADIANS;
  endAz = startAz+(azSpan/DEGREES_TO_RADIANS);
  if (endAz < 0.0) {
    startAz += 360.0;
    endAz += 360.0;
  }
  if ((startAz < 0.0) && (endAz > 180.0))
    startAz += 360.0;
  iStartAz = (int)(startAz/10.0);
  az = (float)(iStartAz*10);
  zA = M_PI*0.5;
  while (az < endAz) {
    char label[10];
    
    sprintf(label, "%2d", ((int)az) % 360);
    if (thetaEta(az*DEGREES_TO_RADIANS, zA, &theta, &eta)) {
      mY = (float)theta;
      mercator((float)eta, &mX);
      mercatorToPixels(mX, mY, plotWidth, plotHeight,
		       &x1, &y1);
      stringWidth = gdk_string_width(smallFont, label)/2;
      if ((x1 > xBorder+xLeftLabelSkip+stringWidth + 30)
	  && (x1 < xBorder+xLeftLabelSkip+plotWidth-stringWidth - 30)) {
	gdk_draw_string(pixmap, smallFont, greenGC,
			x1-stringWidth, displayHeight-2,
			label);
      }
    }
    if (azSpan/DEGREES_TO_RADIANS > 50.0)
      az += 10.0;
    else if (azSpan/DEGREES_TO_RADIANS > 20.0)
      az += 5.0;
    else
      az += 1.0;
  }

  if (inFullscreenMode)
    addSensitiveArea(FALSE, SA_TOP_AREA, 0, 0, displayWidth, displayHeight/5, 0.0);

  addSensitiveArea(FALSE, SA_FINGER_PAN_AREA, 0, 4*displayHeight/5, displayWidth,
		   displayHeight, 0.0);
  gdk_draw_line(pixmap, blueGC,
		0, yBorder+yTopLabelSkip, displayWidth, yBorder+yTopLabelSkip);
  /* Draw a line at the horizon */
  az = centerAz-(M_PI*0.25)-(M_PI*0.5); zA = M_PI*0.5;
  thetaEta(az, zA, &theta, &eta);
  mY = (float)theta;
  mercator((float)eta, &mX);
  mercatorToPixels(mX, mY, plotWidth, plotHeight,
		   &x1, &y1);
  az += M_PI*0.5;
  thetaEta(az, zA, &theta, &eta);
  mY = (float)theta;
  mercator((float)eta, &mX);
  mercatorToPixels(mX, mY, plotWidth, plotHeight,
		   &x2, &y2);
  gdk_draw_line(pixmap, blueGC,
		x1, y1, x2, y2);
  az = 0.0; zA = M_PI*0.4999;
  thetaEta(az, zA, &theta, &eta);
  mY = (float)theta;
  mercator((float)eta, &mX);
  mercatorToPixels(mX, mY, plotWidth, plotHeight,
		   &x1, &y1);
  stringWidth = gdk_string_width(smallFont, "North")/2;
  if ((x1 > xBorder+xLeftLabelSkip+stringWidth)
      && (x1 < xBorder+xLeftLabelSkip+plotWidth-stringWidth))
    gdk_draw_string(pixmap, smallFont, greenGC,
		    x1-stringWidth, y1+15,
		    "North");
  az = M_PI*0.5;
  thetaEta(az, zA, &theta, &eta);
  mY = (float)theta;
  mercator((float)eta, &mX);
  mercatorToPixels(mX, mY, plotWidth, plotHeight,
		   &x1, &y1);
  stringWidth = gdk_string_width(smallFont, "East")/2;
  if ((x1 > xBorder+xLeftLabelSkip+stringWidth)
      && (x1 < xBorder+xLeftLabelSkip+plotWidth-stringWidth))
    gdk_draw_string(pixmap, smallFont, greenGC,
		    x1-stringWidth, y1+15,
		    "East");
  az = M_PI;
  thetaEta(az, zA, &theta, &eta);
  mY = (float)theta;
  mercator((float)eta, &mX);
  mercatorToPixels(mX, mY, plotWidth, plotHeight,
		   &x1, &y1);
  stringWidth = gdk_string_width(smallFont, "South")/2;
  if ((x1 > xBorder+xLeftLabelSkip+stringWidth)
      && (x1 < xBorder+xLeftLabelSkip+plotWidth-stringWidth))
    gdk_draw_string(pixmap, smallFont, greenGC,
		    x1-stringWidth, y1+15,
		    "South");
  az = M_PI*1.5;
  thetaEta(az, zA, &theta, &eta);
  mY = (float)theta;
  mercator((float)eta, &mX);
  mercatorToPixels(mX, mY, plotWidth, plotHeight,
		   &x1, &y1);
  stringWidth = gdk_string_width(smallFont, "West")/2;
  if ((x1 > xBorder+xLeftLabelSkip+stringWidth)
      && (x1 < xBorder+xLeftLabelSkip+plotWidth-stringWidth))
    gdk_draw_string(pixmap, smallFont, greenGC,
		    x1-stringWidth, y1+15,
		    "West");
}

static void sayRedrawing(void)
{
  char *comment = "Redrawing Display";

  gdk_draw_rectangle(pixmap, blackGC, TRUE,
		     plotWidth/2 -
		     (gdk_string_width(smallFont, comment)/2) - 10,
		     plotHeight/2 - 20,
		     gdk_string_width(smallFont, comment) + 19,
		     32);
  gdk_draw_rectangle(pixmap, greenGC, FALSE,
		     plotWidth/2 -
		     (gdk_string_width(smallFont, comment)/2) - 10,
		     plotHeight/2 - 20,
		     gdk_string_width(smallFont, comment) + 19,
		     32);
  gdk_draw_string(pixmap, smallFont, greenGC,
		  plotWidth/2 -
		  (gdk_string_width(smallFont, comment)/2),
		  plotHeight/2,
		  comment);
  gdk_draw_drawable(drawingArea->window,
		    drawingArea->style->fg_gc[GTK_WIDGET_STATE (drawingArea)],
		    pixmap,
		    0,0,0,0,
		    displayWidth, displayHeight);
  gdk_flush();
  shouldSayRedrawing = FALSE;
}

/*
  Draw stick-figure constellations
*/
static void plotConstellations(void)
{
  static int firstCall = TRUE;
  int i;
  constellation *thisConst;

  if (firstCall) {
    while (!constellationsInitialized)
      usleep(100000);
    firstCall = FALSE;
  }
  thisConst = constellations;
  while (thisConst != NULL) {
    int x, y, stringWidth;
    int firstPoint;
    int lastX = 0;
    int lastY = 0;
    float mX, mY;
    double az, zA, theta, eta;

    firstPoint = TRUE;
    if ((useAsterisms && (thisConst->type == ASTERISM)) ||
	(!useAsterisms && (thisConst->type != ASTERISM))) {
      for (i = 0; i < thisConst->nPoints; i++) {
	if ((thisConst->rA[i] == 0.0) && (thisConst->dec[i] == 0.0)) {
	  x = y = -10000;
	} else {
	  if (azZA(thisConst->rA[i],
		   thisConst->sinDec[i],
		   thisConst->cosDec[i],
		   &az, &zA, TRUE)) {
	    thetaEta(az, zA, &theta, &eta);
	    mY = (float)theta;
	    mercator((float)eta, &mX);
	    mercatorToPixels(mX, mY, plotWidth, plotHeight,
			     &x, &y);
	  } else
	    x = y = -10000;
	}
	if (firstPoint)
	  firstPoint = FALSE;
	else if (
		 (y <= displayHeight) &&
		 (x != -10000) && (y != -10000) &&
		 (lastX != -10000) && (lastY != -10000) &&
		 (lastY <= displayHeight))
	  switch (thisConst->type) {
	  case ZODIAC:
	    gdk_draw_line(pixmap, faintPinkGC, x, y, lastX, lastY);
	    break;
	  case PTOLEMAIC:
	    gdk_draw_line(pixmap, faintGoldGC, x, y, lastX, lastY);
	    break;
	  case ASTERISM:
	    gdk_draw_line(pixmap, creamGC, x, y, lastX, lastY);
	    break;
	  default:
	    gdk_draw_line(pixmap, faintBlueGC, x, y, lastX, lastY);
	  }
	lastX = x;
	lastY = y;
      }
      if (azZA(thisConst->nameRA,
	       thisConst->sinNameDec,
	       thisConst->cosNameDec,
	       &az, &zA, TRUE)) {
	thetaEta(az, zA, &theta, &eta);
	mY = (float)theta;
	mercator((float)eta, &mX);
	mercatorToPixels(mX, mY, plotWidth, plotHeight,
			 &x, &y);
	stringWidth = gdk_string_width(smallFont, thisConst->name)/2;
	if ((x >= 0) && (x <= displayWidth) &&
	    (y >= 0) && (y <= displayHeight))
	  switch (thisConst->type) {
	case ZODIAC:
	  gdk_draw_string(pixmap, smallFont, pinkGC,
			  x-stringWidth, y+15, thisConst->name);
	  break;
	case PTOLEMAIC:
	  gdk_draw_string(pixmap, smallFont, yellowGC,
			  x-stringWidth, y+15, thisConst->name);
	  break;
	case ASTERISM:
	  gdk_draw_string(pixmap, smallFont, whiteGC,
			  x-stringWidth, y+15, thisConst->name);
	  break;
	  default:
	    gdk_draw_string(pixmap, smallFont, blueGC,
			    x-stringWidth, y+15, thisConst->name);
	  }
      }
    }
    thisConst = (constellation *)thisConst->next;
  }
}

/*
  Read a line of text from the config file, and strip comments (flagged by #).
*/
int getLine(int fD, char *buffer, int *eOF)
{
  char inChar = (char)0;
  int count = 0;
  int sawComment = FALSE;
  int foundSomething = FALSE;

  buffer[0] = (char)0;
  while ((!(*eOF)) && (inChar != '\n') && (count < 132)) {
    int nChar;

    nChar = read(fD, &inChar, 1);
    if (nChar > 0) {
      foundSomething = TRUE;
      if (inChar == '#')
        sawComment = TRUE;
      if (!sawComment)
        buffer[count++] = inChar;
    } else {
      *eOF = TRUE;
    }
  }
  if (foundSomething) {
    if (count > 0)
      buffer[count-1] = (char)0;
    return(TRUE);
  } else
    return(FALSE);
}

/*
  This function reads in all the files in the deepSky directory.
*/
static void readDeepSkyObjects(void)
{
  char dataDirName[1000], sourceName[100];
  int dataFD;
  deepSkyEntry *newEntry;
  deepSkyEntry *lastEntry = NULL;
  DIR *dirPtr;
  struct dirent *nextEnt;

  dprintf("Reading in Deep Sky Objects\n");
  sprintf(dataDirName, "%s/deepSky/", dataDir);
  dirPtr = opendir(dataDirName);
  while ((nextEnt = readdir(dirPtr)) != NULL)
    if (strstr(nextEnt->d_name, ".") == NULL) {
      dprintf("Processing Deep Sky file \"%s\"\n", nextEnt->d_name);
      sprintf(dataDirName, "%s/deepSky/%s", dataDir, nextEnt->d_name);
      dataFD = open(dataDirName, O_RDONLY);
      if (dataFD < 0) {
	perror("dataDirName");
      } else {
	int eof = FALSE;

	while (!eof) {
	  int nRead, rAHH, rAMM, decDD, decMM, type;
	  float rASS, decSS, mag;

	  getLine(dataFD, &dataDirName[0], &eof);
	  nRead = sscanf(dataDirName, "%s %d %d %f %d %d %f %d %f",
			 &sourceName[0], &rAHH, &rAMM, &rASS,
			 &decDD, &decMM, &decSS, &type, &mag);
	  if (nRead == 9) {
	    double rA, dec, decSign;
	    
	    rA = (double)rAHH + ((double)rAMM)/60.0 + ((double)rASS)/3600.0;
	    if (strstr(dataDirName, "-") == NULL)
	      decSign = 1.0;
	    else
	      decSign = -1.0;
	    dec = decSign*(fabs((double)decDD) + fabs((double)decMM)/60.0 + fabs((double)decSS)/3600.0);
	    newEntry = (deepSkyEntry *)malloc(sizeof(deepSkyEntry));
	    if (newEntry == NULL) {
	      perror("New deep sky entry");
	      exit(ERROR_EXIT);
	    }
	    newEntry->name = (char *)malloc(strlen(sourceName)+1);
	    if (newEntry->name == NULL) {
	      perror("newEntry->name");
	      exit(ERROR_EXIT);
	    }
	    strcpy(newEntry->name, sourceName);
	    newEntry->rA = rA * HOURS_TO_RADIANS;
	    newEntry->dec = dec * DEGREES_TO_RADIANS;
	    newEntry->sinDec = sin(newEntry->dec);
	    newEntry->cosDec = cos(newEntry->dec);
	    newEntry->type = type;
	    newEntry->mag = mag;
	    newEntry->next = NULL;
	    if (deepSkyRoot == NULL)
	      deepSkyRoot = newEntry;
	    else
	      lastEntry->next = newEntry;
	    lastEntry = newEntry;
	  }
	}
	close(dataFD);
      }
    }
  haveReadDeepSkyObjects = TRUE;
}

void plotDeepSkyObject(int kind, int x, int y)
{
  static int firstCall = TRUE;
  int i;
  static int galX[20], galY[20];
  GdkPoint points[20];

  if (firstCall) {
    /* Make template ellipse of galaxy figures */
    for (i = 0; i < 20; i++) {
      float angle;

      angle = M_2PI*0.05*(float)i;
      galX[i] = (int)roundf(2.0*cosf(angle) + 7.0*sinf(angle));
      galY[i] = -(int)roundf(4.0*cosf(angle) + 2.0*sinf(angle));
    }
    firstCall = FALSE;
  }
  switch (kind) {
  case SUPERNOVA_REMNENT:
    gdk_draw_arc(pixmap, yellowGC, FALSE,
		 x-6, y-6, 12, 12, 0, FULL_CIRCLE);
    break;
  case OPEN_CLUSTER:
    gdk_draw_arc(pixmap, yellowGC, TRUE,
		 x-6, y-6, 12, 12, 0, FULL_CIRCLE);
    gdk_draw_arc(pixmap, whiteGC, FALSE,
		 x-6, y-6, 12, 12, 0, FULL_CIRCLE);
    break;
  case GLOBULAR_CLUSTER:
    gdk_draw_arc(pixmap, yellowGC, TRUE,
		 x-6, y-6, 12, 12, 0, FULL_CIRCLE);
    gdk_draw_line(pixmap, blackGC, x, y-6, x, y+6);
    gdk_draw_line(pixmap, blackGC, x-6, y, x+6, y);
    break;
  case DIFFUSE_NEBULA:
    points[0].x = x-5; points[0].y = y-5;
    points[1].x = x+5; points[1].y = y-5;
    points[2].x = x+5; points[2].y = y+5;
    points[3].x = x-5; points[3].y = y+5;
    gdk_draw_polygon(pixmap, greenGC, TRUE, points, 4);
    break;
  case PLANETARY_NEBULA:
    gdk_draw_arc(pixmap, blueGreenGC, TRUE,
		 x-6, y-6, 12, 12, 0, FULL_CIRCLE);
    gdk_draw_arc(pixmap, blackGC, TRUE,
		 x-3, y-3, 6, 6, 0, FULL_CIRCLE);
    break;
  case GALAXY:
    for (i = 0; i < 20; i++) {
      points[i].x = x + galX[i]; points[i].y = y + galY[i];
    }
    gdk_draw_polygon(pixmap, redGC, TRUE, points, 20);
    break;
  }
}

static void plotDeepSkyObjects(void)
{
  int x, y, mercXLow, mercXHigh, mercYLow, mercYHigh, stringWidth;
  float mX, mY;
  double az, zA, theta, eta;
  deepSkyEntry *nextEntry;

  mercXLow  = xBorder+xLeftLabelSkip;
  mercXHigh = xBorder+xLeftLabelSkip+plotWidth;
  mercYLow  = yBorder+yTopLabelSkip;
  mercYHigh = displayHeight;
  nextEntry = deepSkyRoot;
  while (nextEntry != NULL) {
    if (azZA(nextEntry->rA, nextEntry->sinDec,
	     nextEntry->cosDec, &az, &zA, TRUE))
      if (thetaEta(az, zA, &theta, &eta)) {
	mY = (float)theta;
	mercator((float)eta, &mX);
	mercatorToPixels(mX, mY, plotWidth, plotHeight, &x, &y);
	if ((x > mercXLow) && (x < mercXHigh) && (y > mercYLow) && (y < mercYHigh)) {
	  plotDeepSkyObject(nextEntry->type, x, y);
	  stringWidth = gdk_string_width(smallFont, nextEntry->name);
	  gdk_draw_string(pixmap, smallFont, greenGC,
			  x-stringWidth/2, y+18, nextEntry->name);
	}
      }
    nextEntry = nextEntry->next;
  }
}

/*
  Return the day of the year, given the year, month and date.
*/
double makeDayNumber(int year, int month, int day)
{
  int i = 0;
  double dayOffset = 0.0;

  while (i < month - 1)
    dayOffset += monthLengths[i++];
  if ((year % 4 == 0) && (month > 2))
    dayOffset += 1.0;
  dayOffset += (double)day;
  return(dayOffset);
}

/*
  Convert the Julian day into Calendar Date as per
  "Astronomical Algorithms" (Meeus)
*/
void tJDToDate(double tJD, int *year, int *month, int *day)
{
  int Z, alpha, A, B, C, D, E;
  double F;

  Z = (int)(tJD+0.5);
  F = tJD + 0.5 - (double)Z;
  if (Z >= 2299161) {
    alpha = (int)(((double)Z - 1867216.25)/36524.25);
    A = Z + 1 + alpha - alpha/4;
  } else
    A = Z;
  B = A + 1524;
  C = (int)(((double)B - 122.1) / 365.25);
  D = (int)(365.25 * (double)C);
  E = (int)(((double)(B - D))/30.6001);
  *day = (int)((double)B - (double)D - (double)((int)(30.6001*(double)E)) + F);
  if (E < 14)
    *month = E - 1;
  else
    *month = E - 13;
  if (*month > 2)
    *year = C - 4716;
  else
    *year = C - 4715;
}

void plotMeteorShowerRadiants(void)
{
  int yearNow, monthNow, dayNow, occuringNow, flipped, lastDay;
  int mercXLow, mercXHigh, mercYLow, mercYHigh;
  double startDayNumber, endDayNumber;
  double dayOffset, jan0TJD, todayDayNumber, showerTJD;
  meteorShower *shower;
  GdkGC *radiantGC;
  GdkPixmap *radiantImage;
  
  if (!haveReadMeteorShowers) {
    readMeteorShowers();
    haveReadMeteorShowers = TRUE;
  }
  mercXLow  = xBorder+xLeftLabelSkip;
  mercXHigh = xBorder+xLeftLabelSkip+plotWidth;
  mercYLow  = yBorder+yTopLabelSkip;
  mercYHigh = displayHeight;
  tJDToDate(tJD, &yearNow, &monthNow, &dayNow);
  jan0TJD = buildTJD(yearNow-1900, 0, 0, 0, 0, 0);
  shower = meteorShowers;
  while (shower != NULL) {
    startDayNumber= makeDayNumber(yearNow, shower->startMonth, shower->startDay);
    endDayNumber= makeDayNumber(yearNow, shower->endMonth, shower->endDay);
    todayDayNumber= makeDayNumber(yearNow, monthNow, dayNow);
    dayOffset = makeDayNumber(yearNow, shower->maxMonth, shower->maxDay);
    showerTJD = jan0TJD + dayOffset;
    if (startDayNumber < endDayNumber) {
      flipped = FALSE;
      if ((todayDayNumber >= startDayNumber) && (todayDayNumber <= endDayNumber))
	occuringNow = TRUE;
      else
	occuringNow = FALSE;
    } else {
      /* Handle case when shower starts in December and ends in January */
      flipped = TRUE;
      if ((todayDayNumber < startDayNumber) && (todayDayNumber > endDayNumber))
	occuringNow = FALSE;
      else
	occuringNow = TRUE;
    }
    if (occuringNow) {
      int i, x, y, w, h, stringWidth;
      float mX, mY;
      double rA, dec, rA1, dec1, rA2, dec2, entryDayNumber, tJD1, tJD2, deltaDay;
      double az, zA, theta, eta;

      rA1  = rA2 = shower->radiantPositions[0].rA;
      dec1 = dec2 = shower->radiantPositions[0].dec;
      entryDayNumber = makeDayNumber(yearNow, shower->radiantPositions[0].month,
				     shower->radiantPositions[0].day);
      if (flipped && (entryDayNumber > 100.0) && (todayDayNumber < 100.0))
	entryDayNumber -= 365.0;
      else if (flipped && (entryDayNumber < 100.0) && (todayDayNumber > 100.0))
	entryDayNumber += 365.0;
      tJD1 = tJD2 = entryDayNumber;
      i = lastDay = 0;
      while ((i < (shower->nRadiantPositions-1)) && (!lastDay)) {
	i++;
	entryDayNumber = makeDayNumber(yearNow, shower->radiantPositions[i].month,
				       shower->radiantPositions[i].day);
	if (flipped && (entryDayNumber > 100.0) && (todayDayNumber < 100.0))
	  entryDayNumber -= 365.0;
	else if (flipped && (entryDayNumber < 100.0) && (todayDayNumber > 100.0))
	  entryDayNumber += 365.0;
	if (todayDayNumber > entryDayNumber)
	  lastDay = FALSE;
	else
	  lastDay = TRUE;
	rA1 = rA2; dec1 = dec2;
	tJD1 = tJD2;
	rA2 = shower->radiantPositions[i].rA;
	dec2 = shower->radiantPositions[i].dec;
	tJD2 = entryDayNumber;
      }
      if (!lastDay) {
	rA = rA2;
	dec = dec2;
      } else {
	deltaDay = tJD2 - tJD1;
	if (deltaDay <= 0) {
	  rA = rA1;
	  dec = dec1;
	} else {
	  rA  = ((deltaDay + tJD1 - todayDayNumber)*rA1  + (todayDayNumber - tJD1)*rA2)/deltaDay;
	  dec  = ((deltaDay + tJD1 - todayDayNumber)*dec1  + (todayDayNumber - tJD1)*dec2)/deltaDay;
	}
      }
      if (azZA(rA, sin(dec), cos(dec), &az, &zA, TRUE))
	if (thetaEta(az, zA, &theta, &eta)) {
	  mY = (float)theta;
	  mercator((float)eta, &mX);
	  mercatorToPixels(mX, mY, plotWidth, plotHeight, &x, &y);
	  if ((x > mercXLow) && (x < mercXHigh) && (y > mercYLow) && (y < mercYHigh)) {
	    if (shower->zHR >= 50) {
	      radiantImage = meteorRadiantImage[0];
	      radiantGC = whiteGC;
	    } else if (shower->zHR >= 20) {
	      radiantImage = meteorRadiantImage[1];
	      radiantGC = creamGC;
	    } else {
	      radiantImage = meteorRadiantImage[2];
	      radiantGC = greyGC;
	    }
	    gdk_drawable_get_size(radiantImage, &w, &h);
	    gdk_draw_drawable(pixmap, blueGC, radiantImage, 0, 0,
			      x-w/2, y-h/2, 20, 20);
	    stringWidth = gdk_string_width(smallFont, shower->threeLetterName);
	    gdk_draw_string(pixmap, smallFont, radiantGC,
			    x-stringWidth/2, y+21, shower->threeLetterName);
	  }
	}
    }
    shower = shower->next;
  }
}

/*
  This routing draws the sky for the Transverse Mercator projection.
*/
static void redrawScreenTransverseMercator(void)
{
  calculateTransverseMercatorParameters();
  needNewTime = TRUE;
  if (showGreatCircles) {
    plotEcliptic();
    plotEquator();
    plotGalactic();
  }
  if (displayConstellations)
    plotConstellations();
  if (showStars)
    plotFixedObjects();
  if (showPlanets)
    plotPlanets();
  if (showNames && haveStarNames)
    plotStarNames();
  if (showDeepSky) {
    if (!haveReadDeepSkyObjects)
      readDeepSkyObjects();
    plotDeepSkyObjects();
  }
  if (showMeteors)
    plotMeteorShowerRadiants();
  drawTransverseMercatorBorder();
  showTime();
}

/*
  Calculate the transit time of a Solar System object,   Return full TJD of
transit nearest to time passed in tJDS.
*/
double calcTransitTime(double tJDS, int obj, double *transitEl)
{
  int done = FALSE;
  int loopCount = 0;
  float illum, mag;
  double hA, rA, dec, myLST;

  while (!done && (loopCount++ < 10)) {
    myLST = lSTAtTJD(tJDS);
    planetInfo(dataDir, obj, tJDS, &rA, &dec, &illum, &mag);
    hA = myLST - rA;
    if (fabs(hA) < 3.0e-4)
      done = TRUE;
    else
      tJDS -= hA/M_2PI;
  }
  if (northernHemisphere)
    *transitEl = dec + M_PI*0.5 - latitude;
  else
    *transitEl = M_PI*0.5 + latitude - dec;
  return(tJDS);
}

/*
  Calculate the rising or setting time for a Solar System object.   h0 contains the
elevation the object has when the combination of refraction, the source's size,
and parallax makes some portion of the object first visible.   The full TJD
of the event time is returned.
*/
double  calcRiseOrSetTime(int rise, int obj, double tJDS, double lSTNow, double h0,
			  int *circumpolar, double *finalH, double *finalCosH, double *finalDec,
			  float *illum)
{
  double theTime = 0.0;
  double rA, dec, cosH;
  float mag;

  if (obj != EARTH)
    planetInfo(dataDir, obj, tJDS, &rA, &dec, illum, &mag);
  planetInfo(dataDir, obj, tJDS, &rA, &dec, illum, &mag);
  cosH = (sin(h0) - sin(latitude)*sin(dec))/(cos(latitude)*cos(dec));
  if (fabs(cosH) > 1.0) {
    if (cosH > 1.0)
      *circumpolar = 1;
    else
      *circumpolar = -1;
  } else {
    int iteration = 0;
    int closeEnough = FALSE;
    double h, hA;
    double lastHA = 1.0e6;

    *circumpolar = 0;
    while (!closeEnough && (iteration < 10)) {
      if (rise)
	h = -acos(cosH);
      else
	h = acos(cosH);
      hA = lSTNow - rA;
      if (hA > M_PI)
	hA -= M_2PI;
      else if (hA < -M_PI)
	hA += M_2PI;
      iteration++;
      if (fabs(hA - lastHA) < 7.27e-4)
	closeEnough = TRUE;
      else {
	lastHA = hA;
	planetInfo(dataDir, obj, tJDS+(h-hA)/M_2PI, &rA, &dec, illum, &mag);
	cosH = (sin(h0) - sin(latitude)*sin(dec))/(cos(latitude)*cos(dec));
      }
    }
    *finalCosH = cosH;
    *finalH = h;
    *finalDec = dec;
    theTime = tJDS+(h-hA)/M_2PI;
  }
  return(theTime);
}

/*
  Draw a radial line within the planet compass.
 */
void drawCompassLine(GdkGC *gC, float angle, int length, int xCen, int yCen, int radius)
{
  float x1, y1, x2, y2, oRadius, iRadius, sinA, cosA;

  oRadius = (float)radius;
  iRadius = oRadius - (float)length;
  sinA = sinf(angle);
  cosA = cosf(angle);
  x1 = x2 = (float)xCen;
  y1 = y2 = (float)yCen;
  x1 += oRadius*sinA;
  y1 -= oRadius*cosA;
  x2 += iRadius*sinA;
  y2 -= iRadius*cosA;
  gdk_draw_line(pixmap, gC, (int)(x1 + 0.5), (int)(y1 + 0.5), (int)(x2 + 0.5), (int)(y2 + 0.5));
}

double phaseTime(double k, int i)
{
  double T, E, F, M, MP, OM, testTJD, W;
  double A1, A2, A3, A4, A5, A6, A7, A8, A9, A10, A11, A12, A13, A14;
  
  T = k/1236.85;
  testTJD = PHASE_C0 + k*PHASE_C1 + T*T*PHASE_C2 + T*T*T*PHASE_C3 +
    T*T*T*T*PHASE_C4;
  E   = PHASE_E0 + PHASE_E1*T + PHASE_E2*T*T;
  M   = PHASE_M0 + PHASE_M1*k + PHASE_M2*T*T + PHASE_M3*T*T*T;
  M  *= DEGREES_TO_RADIANS;
  MP  = PHASE_MP0 + PHASE_MP1*k + PHASE_MP2*T*T + PHASE_MP3*T*T*T + PHASE_MP4*T*T*T*T;
  MP *= DEGREES_TO_RADIANS;
  F   = PHASE_F0 + PHASE_F1*k + PHASE_F2*T*T + PHASE_F3*T*T*T + PHASE_F4*T*T*T*T;
  F  *= DEGREES_TO_RADIANS;
  OM  = PHASE_OM0 + PHASE_OM1*k + PHASE_OM2*T*T + PHASE_OM3*T*T*T;
  OM *= DEGREES_TO_RADIANS;
  A1  = (PHASE_A1_0  + PHASE_A1_1 *k + PHASE_A1_2*T*T) * DEGREES_TO_RADIANS;
  A2  = (PHASE_A2_0  + PHASE_A2_1 *k) * DEGREES_TO_RADIANS;
  A3  = (PHASE_A3_0  + PHASE_A3_1 *k) * DEGREES_TO_RADIANS;
  A4  = (PHASE_A4_0  + PHASE_A4_1 *k) * DEGREES_TO_RADIANS;
  A5  = (PHASE_A5_0  + PHASE_A5_1 *k) * DEGREES_TO_RADIANS;
  A6  = (PHASE_A6_0  + PHASE_A6_1 *k) * DEGREES_TO_RADIANS;
  A7  = (PHASE_A7_0  + PHASE_A7_1 *k) * DEGREES_TO_RADIANS;
  A8  = (PHASE_A8_0  + PHASE_A8_1 *k) * DEGREES_TO_RADIANS;
  A9  = (PHASE_A9_0  + PHASE_A9_1 *k) * DEGREES_TO_RADIANS;
  A10 = (PHASE_A10_0 + PHASE_A10_1*k) * DEGREES_TO_RADIANS;
  A11 = (PHASE_A11_0 + PHASE_A11_1*k) * DEGREES_TO_RADIANS;
  A12 = (PHASE_A12_0 + PHASE_A12_1*k) * DEGREES_TO_RADIANS;
  A13 = (PHASE_A13_0 + PHASE_A13_1*k) * DEGREES_TO_RADIANS;
  A14 = (PHASE_A14_0 + PHASE_A14_1*k) * DEGREES_TO_RADIANS;
  
  switch(i) {
  case 0: /* New Moon */
    testTJD +=
      -0.40720         * sin(MP)
      +0.17241 * E     * sin(M)
      +0.01608         * sin(2.0*MP)
      +0.01039         * sin(2.0*F)
      +0.00739 * E     * sin(MP-M)
      -0.00514 * E     * sin(MP+M)
      +0.00208 * E * E * sin(2.0*M)
      -0.00111         * sin(MP-2.0*F)
      -0.00057         * sin(MP+2.0*F)
      +0.00056 * E     * sin(2.0*MP+M)
      -0.00042         * sin(3.0*MP)
      +0.00042 * E     * sin(M+2.0*F)
      +0.00038 * E     * sin(M-2.0*F)
      -0.00024 * E     * sin(2.0*MP-M)
      -0.00017         * sin(OM)
      -0.00007         * sin(MP+2.0*M)
      +0.00004         * sin(MP-2.0F)
      +0.00004         * sin(3.0*M)
      +0.00003         * sin(MP+M-2.0*F)
      +0.00003         * sin(2.0*(MP+F))
      -0.00003         * sin(MP+M+2.0*F)
      +0.00003         * sin(MP-M+2.0*F)
      -0.00002         * sin(MP-M-2.0*F)
      -0.00002         * sin(3.0*MP+M)
      +0.00002         * sin(4.0*MP);
    break;
  case 1: /* First and last quarters */
  case 3:
    testTJD +=
      -0.62801         * sin(MP)
      +0.17172 * E     * sin(M)
      -0.01183 * E     * sin(MP+M)
      +0.00862         * sin(2.0*MP)
      +0.00804         * sin(2.0*F)
      +0.00454 * E     * sin(MP-M)
      +0.00204 * E * E * sin(2.0*M)
      -0.00180         * sin(MP-2.0*F)
      -0.00070         * sin(MP+2.0*F)
      -0.00040         * sin(3.0*MP)
      -0.00034 * E     * sin(2.0*MP-M)
      +0.00032 * E     * sin(M+2.0*F)
      +0.00032 * E     * sin(M-2.0*F)
      -0.00028 * E * E * sin(MP+2.0*M)
      +0.00027 * E     * sin(2.0*MP+M)
      -0.00017         * sin(OM)
      -0.00005         * sin(MP-M-2.0*F)
      +0.00004         * sin(2.0*(MP+F))
      -0.00004         * sin(MP+M+2.0*F)
      +0.00004         * sin(MP-2.0*M)
      +0.00003         * sin(MP+M-2.0*F)
      +0.00003         * sin(3.0*M)
      +0.00002         * sin(2.0*(MP-F))
      +0.00002         * sin(MP-M+2.0*F)
      -0.00002         * sin(3.0*MP+M);
    W = 0.00306 - 0.00038*E*cos(M) + 0.00026*cos(MP)
      + 0.00002*(cos(MP+M) + cos(2.0*F) - cos(MP-M));
    if (i == 1)
      testTJD += W;
    else
      testTJD -= W;
    break;
  case 2: /* Full Moon */
    testTJD +=
      -0.40614         * sin(MP)
      +0.17302 * E     * sin(M)
      +0.01614         * sin(2.0*MP)
      +0.01043         * sin(2.0*F)
      +0.00734 * E     * sin(MP-M)
      -0.00515 * E     * sin(MP+M)
      +0.00209 * E * E * sin(2.0*M)
      -0.00111         * sin(MP-2.0*F)
      -0.00057         * sin(MP+2.0*F)
      +0.00056 * E     * sin(2.0*MP+M)
      -0.00042         * sin(3.0*MP)
      +0.00042 * E     * sin(M+2.0*F)
      +0.00038 * E     * sin(M-2.0*F)
      -0.00024 * E     * sin(2.0*MP-M)
      -0.00017         * sin(OM)
      -0.00007         * sin(MP+2.0*M)
      +0.00004         * sin(MP-2.0*F)
      +0.00004         * sin(3.0*M)
      +0.00003         * sin(MP+M-2.0F)
      +0.00003         * sin(2.0*(MP+F))
      -0.00003         * sin(MP+M+2.0*F)
      +0.00003         * sin(MP-M+2.0*F)
      -0.00002         * sin(MP-M-2.0*F)
      -0.00002         * sin(3.0*MP+M)
      +0.00002         * sin(4.0*MP);
    break;
  }
  
  testTJD +=
    +0.000325 * sin(A1 ) + 0.000056 * sin(A8 )
    +0.000165 * sin(A2 ) + 0.000047 * sin(A9 )
    +0.000164 * sin(A3 ) + 0.000042 * sin(A10)
    +0.000126 * sin(A4 ) + 0.000040 * sin(A11)
    +0.000110 * sin(A5 ) + 0.000037 * sin(A12)
    +0.000062 * sin(A6 ) + 0.000035 * sin(A13)
    +0.000060 * sin(A7 ) + 0.000023 * sin(A14);
  return(testTJD);
}

#define ONE_MINUTE_IN_DAYS (1.0/1440.0)
/*
  Calculate and return the tJDs for the next new moon, first quarter etc.
  See Meeus Astronomical Algorithms, Chapt 49.
*/
void nextPhases(double tJD, int nRows, double *nextNew,
                double *nextFirstQuarter, double *nextFull,
                double *nextLastQuarter)
{
  int i, done;
  double testTJD, delta, k, dYear;

  dYear = (tJD - 2451545.5)/365.25;
  tJD -= PHASE_OFFSET;
  k = (dYear * 12.3685);
  k = (double)((int)k);
  for (i = 0; i < 4; i++) {
    done = 0;
    do {
      testTJD = phaseTime(k, i);
      delta = testTJD - tJD;
      if (fabs(delta) < ONE_MINUTE_IN_DAYS)
        k += 1.0;
      else if (delta > PHASE_C1)
        k -= 1.0;
      else
        done = 1;
    } while (!done);
    switch (i) {
      double kk;
      int j;

    case 0:
      nextNew[nRows/2] = testTJD + PHASE_OFFSET;
      kk = k - (double)(nRows/2);
      for (j = 0; j < nRows; j++) {
	if (j != nRows/2)
	  nextNew[j] = phaseTime(kk, i) + PHASE_OFFSET;
	kk += 1.0;
      }
      k += 0.25;
      break;
    case 1:
      nextFirstQuarter[nRows/2] = testTJD + PHASE_OFFSET;
      kk = k - (double)(nRows/2);
      for (j = 0; j < nRows; j++) {
	if (j != nRows/2)
	  nextFirstQuarter[j] = phaseTime(kk, i) + PHASE_OFFSET;
	kk += 1.0;
      }
      k += 0.25;
      break;
    case 2:
      nextFull[nRows/2] = testTJD + PHASE_OFFSET;
      kk = k - (double)(nRows/2);
      for (j = 0; j < nRows; j++) {
	if (j != nRows/2)
	  nextFull[j] = phaseTime(kk, i) + PHASE_OFFSET;
	kk += 1.0;
      }
      k += 0.25;
      break;
    case 3:
      nextLastQuarter[nRows/2] = testTJD + PHASE_OFFSET;
      kk = k - (double)(nRows/2);
      for (j = 0; j < nRows; j++) {
	if (j != nRows/2)
	  nextLastQuarter[j] = phaseTime(kk, i) + PHASE_OFFSET;
	kk += 1.0;
      }
      break;
    }
  }
}

unsigned char normalYear[12] = {0, 3, 3, 6, 1, 4, 6, 2, 5, 0, 3, 5};
unsigned char leapYear[2] = {6, 2};

/*
  Return the day of the week (0->6) for year, month, day
*/
int dayOfWeek(int year, int month, int day)
{
  int century, c, weekDay, yy, y, m, leap;

  century = year / 100;
  yy = year - century*100;
  c = 2*(3 - (century % 4));
  y = yy + yy/4;
  if ((year % 400) == 0)
    leap = TRUE;
  else if ((year % 100) == 0)
    leap = FALSE;
  else if ((year % 4) == 0)
    leap = TRUE;
  else
    leap = FALSE;
  if (leap && (month < 3))
    m = (int)leapYear[month-1];
  else
    m = (int)normalYear[month-1];
  weekDay = (c + y + m + day) % 7;
  return(weekDay);
}

#define CAL_TOP_OFFSET (40)
#define CAL_LEFT_OFFSET (3)
#define CAL_DAY_WIDTH  (67)
#define CAL_DAY_HEIGHT (85)

/*
  Plot a little box for the day, with the moon shown with the proper phase, for
  the one month moon calendar.
*/

void plotDayBox(double tJDS, int week, int weekDay, int day, int today,
		int phase, int blue)
{
  int isFull = FALSE;
  double dummy;
  float illum2, illum, mag, sunAngle;
  char dayString[3];
  GdkPoint points[4];
  GdkGC *dayGC;

  planetInfo(dataDir, EARTH, tJDS, &dummy, &dummy, &illum2, &mag);
  planetInfo(dataDir, MOON, tJDS+0.04, &dummy, &dummy, &illum2, &mag);
  planetInfo(dataDir, MOON, tJDS, &dummy, &dummy, &illum, &mag);
  if (northernHemisphere) {
    if (illum < illum2)
      sunAngle = 0.0;
    else
      sunAngle = M_PI;
  } else {
    if (illum > illum2)
      sunAngle = 0.0;
    else
      sunAngle = M_PI;
  }
  points[0].x = CAL_LEFT_OFFSET + weekDay*CAL_DAY_WIDTH; points[0].y = CAL_TOP_OFFSET + week*CAL_DAY_HEIGHT;
  points[1].x = points[0].x;                             points[1].y = points[0].y + CAL_DAY_HEIGHT;
  points[2].x = points[1].x + CAL_DAY_WIDTH;             points[2].y = points[1].y;
  points[3].x = points[2].x;                             points[3].y = points[2].y - CAL_DAY_HEIGHT;
  if (phase == 2)
    isFull = TRUE;
  drawMoon(CAL_DAY_WIDTH/2 - 1, illum, sunAngle, points[0].x+CAL_DAY_WIDTH/2 + 1,
	   points[0].y+CAL_DAY_HEIGHT/2 + 7, TRUE, isFull, blue);
  gdk_draw_polygon(pixmap, blueGC, FALSE, points, 4);
  sprintf(dayString, "%2d", day);
  if (today)
    dayGC = creamGC;
  else
    dayGC = blueGC;
  gdk_draw_string(pixmap, smallFont, dayGC, points[0].x+gdk_string_width(smallFont, dayString)/2 - 1,
		  points[0].y+gdk_string_width(smallFont, dayString), dayString);
  if (phase == 0)
    gdk_draw_string(pixmap, smallFont, greenGC,
		    points[0].x + CAL_DAY_WIDTH/2 - gdk_string_width(smallFont, "New")/2 + 3,
		    points[0].y+gdk_string_width(smallFont, dayString), "New");
  else if (phase == 2)
    gdk_draw_string(pixmap, smallFont, greenGC,
		    points[0].x + CAL_DAY_WIDTH/2 - gdk_string_width(smallFont, "Full")/2 + 4,
		    points[0].y+gdk_string_width(smallFont, dayString), "Full");
}

#define SUN_DARK_EL_LIMIT  (-12.0*DEGREES_TO_RADIANS)
#define MOON_DARK_EL_LIMIT ( -6.0*DEGREES_TO_RADIANS)
/*
  `The darkNow() function returns TRUE if the Sun and Moon are far
enough below the horizon for the sky to be very dark.   It returns
FALSE otherwise.   It takes the phase of the moon into account when
deciding how far below the horizon it must be before the sky is
dark.
*/
int darkNow(double tJD)
{
  float dummy;
  double az, zA, el;
  static double savedSunTJD = 0.0;
  static double savedMoonTJD = 0.0;
  static double savedSunRA, savedSunDec, savedMoonRA, savedMoonDec;
  static double sinSavedSunDec, cosSavedSunDec, sinSavedMoonDec, cosSavedMoonDec;
  static float savedPhase, threshold;

  if (fabs(tJD-savedSunTJD) >= 0.5) {
    planetInfo(dataDir, EARTH, tJD, &savedSunRA, &savedSunDec, &dummy, &dummy);
    sinSavedSunDec = sin(savedSunDec);
    cosSavedSunDec = cos(savedSunDec);
    savedSunTJD = tJD;
  }
  azZA(savedSunRA, sinSavedSunDec, cosSavedSunDec, &az, &zA, FALSE);
  el = M_PI*0.5 - zA;
  /*
  printf("in darkNow(%f), sun el = %f\n", tJD, el/DEGREES_TO_RADIANS);
  */
  if (el > SUN_DARK_EL_LIMIT)
    return(FALSE);
  if (fabs(tJD-savedMoonTJD) >= 1.0/24.0) {
    planetInfo(dataDir, MOON, tJD, &savedMoonRA, &savedMoonDec, &savedPhase, &dummy);
    threshold = savedPhase*MOON_DARK_EL_LIMIT;
    sinSavedMoonDec = sin(savedMoonDec);
    cosSavedMoonDec = cos(savedMoonDec);
    savedMoonTJD = tJD;
  }
  azZA(savedMoonRA, sinSavedMoonDec, cosSavedMoonDec, &az, &zA, FALSE);
  el = M_PI*0.5 - zA;
  /*
  printf("in darkNow(%f), moon el = %f, phase = %f, thresh = %f\n",
	 tJD, el/DEGREES_TO_RADIANS, phase, threshold/DEGREES_TO_RADIANS);
  */
  if (el > threshold)
    return(FALSE);
  return(TRUE);
}

#define FIFTEEN_MINUTES_IN_DAYS (15.0/1440.0)

/*
  darkTime() returns the number of dark hours between startTJD and endTJD.
 */
float darkTime(double startTJD, double stopTJD, double *begin, double *end)
{
  int savedUseJulianTime;
  float darkHours;
  double tJD;

  savedUseJulianTime = useJulianTime;
  useJulianTime = TRUE;
  darkHours = *begin = *end = 0.0;
  tJD = startTJD;
  while (tJD <= stopTJD) {
    julianDate = tJD;
    myLST = lSTAtTJD(tJD);
    if (darkNow(tJD)) {
      darkHours += 0.25;
      if (*begin == 0)
	*begin = tJD;
    } else if ((*begin != 0.0) && (*end == 0.0))
      *end = tJD;
    tJD += FIFTEEN_MINUTES_IN_DAYS;
  }
  useJulianTime = savedUseJulianTime;
  return(darkHours);
}

#define ABOUT_ROW_STEP    (12)
#define BIG_FONT_OFFSET   (10)
#define STAR_SYMBOL_OFFSET (5)
#define REFRACTION_AT_HORIZON (9.89e-3) /* 34 arc minutes - typical */
#define N_RISE_SET_DAYS (5)
#define SUNLINE_X_OFFSET (42)
#define MOONLINE_X_OFFSET (5)
#define MOON_LABEL_SHIFT (112)
#define N_PHASE_DATES (17)
#define MAX_PHASE_DATES (24)

/*
  Draw various pages on selected from the "opts" menu
 */
static void drawOptsScreens(void)
{
  int stringWidth, rAHH, rAMM, decDD, decMM;
  int row = 1;
  int schematic = FALSE;
  int magIndex = 0;
  int r, c, w, h;
  float n = 1.0;
  float illum, mag, rAHours, decDegrees, el, sunAngle, sunAz, sunZA, illumR, illumS;
  double lSTNow, hA, az, zA, h0, fracTJD;
  double sunRA, sunDec, cosH;
  double nextNew[MAX_PHASE_DATES], nextFirstQ[MAX_PHASE_DATES], nextFull[MAX_PHASE_DATES], nextLastQ[MAX_PHASE_DATES];
  double tJDS[N_RISE_SET_DAYS], theAz, theTime, hD, tempDay;
  int day, year, month, dayNum, circumpolar;
  int nextRiseDone, nextSetDone;
  char scratchString[100];
  GdkPoint points[3];
  GdkGC *labelGC, *textGC;

  switch (aboutScreen) {
  case ABOUT_SCREEN:
    sprintf(scratchString, "Orrery Version %s", orreryVersion);
    stringWidth = gdk_string_width(smallFont, scratchString);
    gdk_draw_string(pixmap, smallFont, whiteGC,
		    (displayWidth/2)-stringWidth/2,
		    (row++)*ABOUT_ROW_STEP + BIG_FONT_OFFSET, scratchString);
    row++;
    gdk_draw_line(pixmap, blueGC,
		  xBorder, row*ABOUT_ROW_STEP, displayWidth-xBorder, row*ABOUT_ROW_STEP);
    row++;
    
    sprintf(scratchString, "Stellar Magnitudes");
    stringWidth = gdk_string_width(smallFont, scratchString);
    gdk_draw_string(pixmap, smallFont, whiteGC,
		    (displayWidth/2)-stringWidth/2,
		    row*ABOUT_ROW_STEP+BIG_FONT_OFFSET, scratchString);
    row += 3;
    for (c = 0; c < 3; c++)
      for (r = 0; r < 5; r++) {
	if ((r ==0) && (c == 0))
	  sprintf(scratchString, "      < %4.1f", -0.5/magScale);
	else if ((r == 4) && (c == 2))
	  sprintf(scratchString, "      > %4.1f", (0.5*n - 1.5)/magScale);
	else
	  sprintf(scratchString, "%4.1f -> %4.1f", (0.5*n - 1.5)/magScale,
		  (0.5*(n+1.0) - 1.5)/magScale);
	n += 1.0;
	scratchString[0] = toupper(scratchString[0]);
	gdk_draw_string(pixmap, smallFont, creamGC,
			(displayWidth/3)*c + 15,
			(row + 2*r)*ABOUT_ROW_STEP, scratchString);
	if (magIndex <= 11)
	  gdk_draw_arc(pixmap, whiteGC, TRUE,
		       (displayWidth/3)*c + 120-(starSize[magIndex]/2),
		       (row + 2*r)*ABOUT_ROW_STEP-(starSize[magIndex]/2) - STAR_SYMBOL_OFFSET,
		       starSize[magIndex], starSize[magIndex], 0, FULL_CIRCLE);
	else if (magIndex == 12) {
	  gdk_draw_point(pixmap, whiteGC, (displayWidth/3)*c + 120, (row + 2*r)*ABOUT_ROW_STEP - STAR_SYMBOL_OFFSET);
	  gdk_draw_point(pixmap, whiteGC, (displayWidth/3)*c + 120+1, (row + 2*r)*ABOUT_ROW_STEP - STAR_SYMBOL_OFFSET);
	} else if (magIndex == 13) 
	  gdk_draw_point(pixmap, greyGC, (displayWidth/3)*c + 120, (row + 2*r)*ABOUT_ROW_STEP - STAR_SYMBOL_OFFSET);
	else
	  gdk_draw_point(pixmap, darkGreyGC, (displayWidth/3)*c + 120, (row + 2*r)*ABOUT_ROW_STEP - STAR_SYMBOL_OFFSET);
	magIndex += 1;
      }
    row += 10;
    gdk_draw_line(pixmap, blueGC,
		  xBorder, row*ABOUT_ROW_STEP, displayWidth-xBorder, row*ABOUT_ROW_STEP);
    
    row++;
    sprintf(scratchString, "Solar System Symbols");
    stringWidth = gdk_string_width(smallFont, scratchString);
    gdk_draw_string(pixmap, smallFont, whiteGC,
		    (displayWidth/2)-stringWidth/2,
		    (row++)*ABOUT_ROW_STEP + BIG_FONT_OFFSET, scratchString);
    row += 2;
    for (r = 0; r < 3; r++)
      for (c = 0; c < 4; c++) {
	if ((r == 2) && (c == 3)) {
	  strcpy(scratchString, "Vernal");
	  gdk_draw_string(pixmap, smallFont, creamGC,
			  (displayWidth/4)*c + 15,
			  (row + 2*r)*ABOUT_ROW_STEP, scratchString);
	  strcpy(scratchString, "Equinox");
	  gdk_draw_string(pixmap, smallFont, creamGC,
			  (displayWidth/4)*c + 15,
			  (row + 2*r)*ABOUT_ROW_STEP + 16, scratchString);
	  gdk_drawable_get_size(firstPointImage, &w, &h);
	  gdk_draw_drawable(pixmap, blueGC, firstPointImage, 0, 0,
			    (displayWidth/4)*c + 75 - w/2,
			    (row + 2*r)*ABOUT_ROW_STEP-h/2 + 2, w, h);
	} else if ((r == 1) && (c == 3)) {
	  strcpy(scratchString, "Meteors");
	  gdk_draw_string(pixmap, smallFont, creamGC,
			  (displayWidth/4)*c + 15,
			  (row + 2*r)*ABOUT_ROW_STEP, scratchString);
	  gdk_drawable_get_size(meteorRadiantImage[0], &w, &h);
	  gdk_draw_drawable(pixmap, blueGC, meteorRadiantImage[0], 0, 0,
			    (displayWidth/4)*c + 75-w/2, (row + 2*r)*ABOUT_ROW_STEP-h/2 - 4, 20, 20);
	} else {
	  strcpy(scratchString, solarSystemNames[c*3 + r]);
	  gdk_draw_string(pixmap, smallFont, creamGC,
			  (displayWidth/4)*c + 15,
			  (row + 2*r)*ABOUT_ROW_STEP, scratchString);
	  gdk_drawable_get_size(planetImages[c*3 + r], &w, &h);
	  gdk_draw_drawable(pixmap, blueGC, planetImages[c*3 + r], 0, 0,
			    (displayWidth/4)*c + 75-w/2, (row + 2*r)*ABOUT_ROW_STEP-h/2 - 4, 20, 20);
	}
      }
    row += 6;
    gdk_draw_line(pixmap, blueGC,
		  xBorder, row*ABOUT_ROW_STEP, displayWidth-xBorder, row*ABOUT_ROW_STEP);
    row++;
    sprintf(scratchString, "Deep Sky Objects");
    stringWidth = gdk_string_width(smallFont, scratchString);
    gdk_draw_string(pixmap, smallFont, whiteGC,
		    (displayWidth/2)-stringWidth/2,
		    (row++)*ABOUT_ROW_STEP + BIG_FONT_OFFSET, scratchString);
    row += 2;
    sprintf(scratchString, "Supernova Rem.");
    stringWidth = gdk_string_width(smallFont, scratchString);
    gdk_draw_string(pixmap, smallFont, creamGC,
		    15, row*ABOUT_ROW_STEP, scratchString);
    plotDeepSkyObject(SUPERNOVA_REMNENT, 125, row*ABOUT_ROW_STEP-5);
    sprintf(scratchString, "Open Cluster");
    stringWidth = gdk_string_width(smallFont, scratchString);
    gdk_draw_string(pixmap, smallFont, creamGC,
		    displayWidth/3 + 15, row*ABOUT_ROW_STEP, scratchString);
    plotDeepSkyObject(OPEN_CLUSTER, displayWidth/3 + 125, row*ABOUT_ROW_STEP-5);
    sprintf(scratchString, "Globular Cluster");
    stringWidth = gdk_string_width(smallFont, scratchString);
    gdk_draw_string(pixmap, smallFont, creamGC,
		    2*displayWidth/3 + 15, row*ABOUT_ROW_STEP, scratchString);
    plotDeepSkyObject(GLOBULAR_CLUSTER, 2*displayWidth/3 + 125, row*ABOUT_ROW_STEP-5);
    row += 2;
    sprintf(scratchString, "Diffuse Nebula");
    stringWidth = gdk_string_width(smallFont, scratchString);
    gdk_draw_string(pixmap, smallFont, creamGC,
		    15, row*ABOUT_ROW_STEP, scratchString);
    plotDeepSkyObject(DIFFUSE_NEBULA, 125, row*ABOUT_ROW_STEP-5);
    sprintf(scratchString, "Planetary Nebula");
    stringWidth = gdk_string_width(smallFont, scratchString);
    gdk_draw_string(pixmap, smallFont, creamGC,
		    displayWidth/3 + 15, row*ABOUT_ROW_STEP, scratchString);
    plotDeepSkyObject(PLANETARY_NEBULA, displayWidth/3 + 125, row*ABOUT_ROW_STEP-5);
    sprintf(scratchString, "Galaxy");
    stringWidth = gdk_string_width(smallFont, scratchString);
    gdk_draw_string(pixmap, smallFont, creamGC,
		    2*displayWidth/3 + 15, row*ABOUT_ROW_STEP, scratchString);
    plotDeepSkyObject(GALAXY, 2*displayWidth/3 + 125, row*ABOUT_ROW_STEP-5);
    row++;
    gdk_draw_line(pixmap, blueGC,
		  xBorder, row*ABOUT_ROW_STEP, displayWidth-xBorder, row*ABOUT_ROW_STEP);
    row++;
    
    sprintf(scratchString, "Constellation Colors");
    stringWidth = gdk_string_width(smallFont, scratchString);
    gdk_draw_string(pixmap, smallFont, whiteGC,
		    (displayWidth/2)-stringWidth/2,
		    (row++)*ABOUT_ROW_STEP + BIG_FONT_OFFSET, scratchString);
    row += 2;
    sprintf(scratchString, "Zodiac");
    stringWidth = gdk_string_width(smallFont, scratchString);
    gdk_draw_string(pixmap, smallFont, pinkGC,
		    15, row*ABOUT_ROW_STEP, scratchString);
    sprintf(scratchString, "Ptolemaic");
    stringWidth = gdk_string_width(smallFont, scratchString);
    gdk_draw_string(pixmap, smallFont, yellowGC,
		    displayWidth/4 + 15, row*ABOUT_ROW_STEP, scratchString);
    sprintf(scratchString, "Modern Additions");
    stringWidth = gdk_string_width(smallFont, scratchString);
    gdk_draw_string(pixmap, smallFont, blueGC,
		  displayWidth/2 + 15, row*ABOUT_ROW_STEP, scratchString);
    sprintf(scratchString, "Asterisms");
    stringWidth = gdk_string_width(smallFont, scratchString);
    gdk_draw_string(pixmap, smallFont, creamGC,
		  3*displayWidth/4 + 15, (row++)*ABOUT_ROW_STEP, scratchString);
    gdk_draw_line(pixmap, blueGC,
		  xBorder, row*ABOUT_ROW_STEP, displayWidth-xBorder, row*ABOUT_ROW_STEP);
    row++;
    sprintf(scratchString, "Great Circles");
    stringWidth = gdk_string_width(smallFont, scratchString);
    gdk_draw_string(pixmap, smallFont, whiteGC,
		    (displayWidth/2)-stringWidth/2,
		    (row++)*ABOUT_ROW_STEP + BIG_FONT_OFFSET, scratchString);
    row += 2;
    sprintf(scratchString, "Celestial Equator");
    stringWidth = gdk_string_width(smallFont, scratchString);
    gdk_draw_string(pixmap, smallFont, equatorGC,
		    15, row*ABOUT_ROW_STEP, scratchString);
    sprintf(scratchString, "Ecliptic");
    stringWidth = gdk_string_width(smallFont, scratchString);
    gdk_draw_string(pixmap, smallFont, redGC,
		    displayWidth/3 + 15, row*ABOUT_ROW_STEP, scratchString);
    sprintf(scratchString, "Galactic Plane");
    stringWidth = gdk_string_width(smallFont, scratchString);
    gdk_draw_string(pixmap, smallFont, blueGreenGC,
		    2*displayWidth/3 + 15, (row++)*ABOUT_ROW_STEP, scratchString);
    gdk_draw_line(pixmap, blueGC,
		  xBorder, row*ABOUT_ROW_STEP, displayWidth-xBorder, row*ABOUT_ROW_STEP);
    row += 2;
    break;
  case SOLUNI_SCREEN:
    textGC = whiteGC;
    needNewTime = TRUE;
    lSTNow = lST();
    strcpy(scratchString, "Sun and Moon Information at ");
    makeTimeString(&scratchString[strlen(scratchString)]);
    stringWidth = gdk_string_width(smallFont, scratchString);
    gdk_draw_string(pixmap, smallFont, whiteGC,
		    (displayWidth/2)-stringWidth/2,
		    (row++)*ABOUT_ROW_STEP + BIG_FONT_OFFSET, scratchString);
    row += 2;
    stringWidth = gdk_string_width(smallFont, "Sun");
    gdk_draw_string(pixmap, smallFont, whiteGC,
		    (displayWidth/2)-stringWidth/2,
		    (row++)*ABOUT_ROW_STEP + BIG_FONT_OFFSET, "Sun");
    row += 1;
    planetInfo(dataDir, EARTH, tJD, &sunRA, &sunDec, &illum, &mag);
    rAHours = sunRA/HOURS_TO_RADIANS; decDegrees = sunDec/DEGREES_TO_RADIANS;
    rAHH = (int)rAHours;
    rAMM = (int)((rAHours - (float)rAHH)*60.0);
    gdk_draw_string(pixmap, smallFont, creamGC,
		    0,
		    row*ABOUT_ROW_STEP + BIG_FONT_OFFSET, "RA");
    sprintf(scratchString, "%02d:%02d", rAHH, rAMM);
    gdk_draw_string(pixmap, smallFont, whiteGC,
		    gdk_string_width(smallFont, "RA "),
		    row*ABOUT_ROW_STEP + BIG_FONT_OFFSET, scratchString);
    decDD = (int)decDegrees;
    decMM = (int)((decDegrees - (float)decDD)*60.0);
    gdk_draw_string(pixmap, smallFont, creamGC,
		    displayWidth/5,
		    row*ABOUT_ROW_STEP + BIG_FONT_OFFSET, "Dec");
    if (decDegrees > 0.0)
      sprintf(scratchString, "%02d:%02d", decDD, decMM);
    else
      sprintf(scratchString, "-%02d:%02d", -decDD, -decMM);
    gdk_draw_string(pixmap, smallFont, whiteGC,
		    displayWidth/5 + gdk_string_width(smallFont, "Dec "),
		    row*ABOUT_ROW_STEP + BIG_FONT_OFFSET, scratchString);
    hA = lSTNow - sunRA;
    rAHours = hA/HOURS_TO_RADIANS;
    gdk_draw_string(pixmap, smallFont, creamGC,
		    2*displayWidth/5,
		    row*ABOUT_ROW_STEP + BIG_FONT_OFFSET, "HA");
    sprintf(scratchString, "%7.3f", rAHours);
    gdk_draw_string(pixmap, smallFont, whiteGC,
		    2*displayWidth/5 + gdk_string_width(smallFont, "HA "),
		    row*ABOUT_ROW_STEP + BIG_FONT_OFFSET, scratchString);
    azZA(sunRA, sin(sunDec), cos(sunDec), &az, &zA, FALSE);
    sunAz = az; sunZA = zA;
    az /= DEGREES_TO_RADIANS;
    if (az < 0.0)
      az += 360.0;
    if (az > 360.0)
      az -= 360.0;
    gdk_draw_string(pixmap, smallFont, creamGC,
		    3*displayWidth/5,
		    row*ABOUT_ROW_STEP + BIG_FONT_OFFSET, "Az");
    sprintf(scratchString, "%7.3f", az);
    gdk_draw_string(pixmap, smallFont, whiteGC,
		    3*displayWidth/5 + gdk_string_width(smallFont, "Az "),
		    row*ABOUT_ROW_STEP + BIG_FONT_OFFSET, scratchString);
    el = M_PI*0.5 - zA;
    el += refraction(el);
    gdk_draw_string(pixmap, smallFont, creamGC,
		    4*displayWidth/5,
		    row*ABOUT_ROW_STEP + BIG_FONT_OFFSET, "El");
    sprintf(scratchString, "%7.3f", el/DEGREES_TO_RADIANS);
    gdk_draw_string(pixmap, smallFont, whiteGC,
		    4*displayWidth/5 + gdk_string_width(smallFont, "El "),
		    row*ABOUT_ROW_STEP + BIG_FONT_OFFSET, scratchString);
    row += 1;
    /* Calculate time of sun rise, transit and setting */
    nextRiseDone = nextSetDone = FALSE;
    h0 = -(50.0/60.0)*DEGREES_TO_RADIANS;
    fracTJD = tJD - (double)((int)tJD);
    for (day = 0; day < N_RISE_SET_DAYS; day++) {
      if (fracTJD < 0.5)
	tJDS[day] = (double)((int)tJD) + 0.5 + longitude/M_2PI + (double)(day - N_RISE_SET_DAYS/2);
      else
	tJDS[day] = (double)((int)tJD) + 1.5 + longitude/M_2PI + (double)(day - N_RISE_SET_DAYS/2);
      tJDToDate(tJDS[day], &year, &month, &dayNum);
      sprintf(scratchString, "%02d/%02d/%04d", month, dayNum, year);
      if (day - N_RISE_SET_DAYS/2 == 0)
	gdk_draw_string(pixmap, smallFont, whiteGC,
			SUNLINE_X_OFFSET, row*ABOUT_ROW_STEP + BIG_FONT_OFFSET, scratchString);
      else
	gdk_draw_string(pixmap, smallFont, greyGC,
			SUNLINE_X_OFFSET, row*ABOUT_ROW_STEP + BIG_FONT_OFFSET, scratchString);
      lSTNow = lSTAtTJD(tJDS[day]);
      /* Sunrise */
      theTime = calcRiseOrSetTime(TRUE, EARTH, tJDS[day], lSTNow, h0, &circumpolar, &hD,
				  &cosH, &sunDec, &illum);
      if (circumpolar != 0) {
	gdk_draw_string(pixmap, smallFont, darkCreamGC,
			SUNLINE_X_OFFSET+gdk_string_width(smallFont, "88/88/8888  "),
			row*ABOUT_ROW_STEP + BIG_FONT_OFFSET, "Sunrise");
	gdk_draw_string(pixmap, smallFont, greyGC,
			SUNLINE_X_OFFSET+gdk_string_width(smallFont, "88/88/8888  Sunrise "),
			row*ABOUT_ROW_STEP + BIG_FONT_OFFSET, "(none)");
      } else {
	theAz = atan2(sin(hD), cosH*sinLatitude-tan(sunDec)*cosLatitude) + M_PI;
	if (theAz < 0.0)
	  theAz += M_2PI;
	if (theAz > M_2PI)
	  theAz -= M_2PI;
	if ((!nextRiseDone) && (theTime - tJD >= 0.0)) {
	  labelGC = creamGC;
	    textGC = whiteGC;
	    nextRiseDone = TRUE;
	} else {
	  labelGC = darkCreamGC;
	  textGC = greyGC;
	}
	theTime -= 0.5;
	theTime = theTime - (double)((int)theTime);
	rAHours = theTime*24.0;
	rAHH = (int)rAHours;
	rAMM = (int)((rAHours - (float)rAHH)*60.0 + 0.5);
	sprintf(scratchString, "%02d:%02d ", rAHH, rAMM);
	gdk_draw_string(pixmap, smallFont, labelGC,
			SUNLINE_X_OFFSET+gdk_string_width(smallFont, "88/88/8888  "),
			row*ABOUT_ROW_STEP + BIG_FONT_OFFSET, "Sunrise");
	gdk_draw_string(pixmap, smallFont, textGC,
			SUNLINE_X_OFFSET+gdk_string_width(smallFont, "88/88/8888  Sunrise "),
			row*ABOUT_ROW_STEP + BIG_FONT_OFFSET, scratchString);
	gdk_draw_string(pixmap, smallFont, labelGC,
			SUNLINE_X_OFFSET+gdk_string_width(smallFont, "88/88/8888  Sunrise 88:88 "),
			row*ABOUT_ROW_STEP + BIG_FONT_OFFSET, "UT Az =");
	sprintf(scratchString, "%3.0f", theAz/DEGREES_TO_RADIANS);
	gdk_draw_string(pixmap, smallFont, textGC,
			SUNLINE_X_OFFSET+gdk_string_width(smallFont, "88/88/8888  Sunrise 88:88 UT Az = "),
			row*ABOUT_ROW_STEP + BIG_FONT_OFFSET, scratchString);
      }
      /* Sunset */
      theTime = calcRiseOrSetTime(FALSE, EARTH, tJDS[day], lSTNow, h0, &circumpolar, &hD,
				  &cosH, &sunDec, &illum);
      if (circumpolar != 0) {
	gdk_draw_string(pixmap, smallFont, darkCreamGC,
			SUNLINE_X_OFFSET+gdk_string_width(smallFont, "88/88/8888  Sunrise 88:88 UT Az = 888.8  "),
			row*ABOUT_ROW_STEP + BIG_FONT_OFFSET, "Sunset");
	gdk_draw_string(pixmap, smallFont, greyGC,
			SUNLINE_X_OFFSET+gdk_string_width(smallFont, "88/88/8888  Sunrise 88:88 UT Az = 888.8  Sunset "),
			row*ABOUT_ROW_STEP + BIG_FONT_OFFSET, "(none)");
      } else {
	theAz = atan2(sin(hD), cosH*sinLatitude-tan(sunDec)*cosLatitude) + M_PI;
	if (theAz < 0.0)
	  theAz += M_2PI;
	if (theAz > M_2PI)
	  theAz -= M_2PI;
	if ((!nextSetDone) && (theTime - tJD >= 0.0)) {
	  labelGC = creamGC;
	  textGC = whiteGC;
	  nextSetDone = TRUE;
	} else {
	  labelGC = darkCreamGC;
	  textGC = greyGC;
	}
	theTime -= 0.5;
	theTime = theTime - (double)((int)theTime);
	rAHours = theTime*24.0;
	rAHH = (int)rAHours;
	rAMM = (int)((rAHours - (float)rAHH)*60.0 + 0.5);
	sprintf(scratchString, "%02d:%02d ", rAHH, rAMM);
	gdk_draw_string(pixmap, smallFont, labelGC,
			SUNLINE_X_OFFSET+gdk_string_width(smallFont, "88/88/8888  Sunrise 88:88 UT Az = 888.8  "),
			row*ABOUT_ROW_STEP + BIG_FONT_OFFSET, "Sunset");
	gdk_draw_string(pixmap, smallFont, textGC,
			SUNLINE_X_OFFSET+gdk_string_width(smallFont, "88/88/8888  Sunrise 88:88 UT Az = 888.8  Sunset "),
			row*ABOUT_ROW_STEP + BIG_FONT_OFFSET, scratchString);
	gdk_draw_string(pixmap, smallFont, labelGC,
			SUNLINE_X_OFFSET+gdk_string_width(smallFont, "88/88/8888  Sunrise 88:88 UT Az = 888.8  Sunset 88:88 "),
			row*ABOUT_ROW_STEP + BIG_FONT_OFFSET, "UT Az =");
	sprintf(scratchString, "%3.0f", theAz/DEGREES_TO_RADIANS);
	  gdk_draw_string(pixmap, smallFont, textGC,
			  SUNLINE_X_OFFSET+gdk_string_width(smallFont, "88/88/8888  Sunrise 88:88 UT Az = 888.8  Sunset 88:88 UT Az = "),
			  row*ABOUT_ROW_STEP + BIG_FONT_OFFSET, scratchString);
      }
      row++;
    }
    /* Moon Stuff */
    row += 2;
    stringWidth = gdk_string_width(smallFont, "Moon");
    gdk_draw_string(pixmap, smallFont, whiteGC,
		    (displayWidth/2)-stringWidth/2 - MOON_LABEL_SHIFT,
		    row*ABOUT_ROW_STEP + BIG_FONT_OFFSET, "Moon");
    planetInfo(dataDir, EARTH, tJD, &sunRA, &sunDec, &illum, &mag);
    planetInfo(dataDir, MOON, tJD, &sunRA, &sunDec, &illum, &mag);
    /*
    azZA(sunRA, sin(sunDec), cos(sunDec), &az, &zA, FALSE);
    sunAngle = greatCircleDirection((float)az, (float)((M_PI*0.5)-zA),
				    sunAz, (M_PI*0.5)-sunZA);
    */
    sprintf(scratchString, "%2.0f%% Illuminated", illum*100.0);
    stringWidth = gdk_string_width(smallFont, scratchString);
    gdk_draw_string(pixmap, smallFont, whiteGC,
		    (displayWidth/2) + 55 - MOON_LABEL_SHIFT,
		    row*ABOUT_ROW_STEP + BIG_FONT_OFFSET, scratchString);
    sunAngle = 0.0;
    if (fabs(illum) < 0.02)
      sprintf(scratchString, "New Moon");
    else if (fabs(illum) > 0.98)
      sprintf(scratchString, "Full Moon");
    else {
      double dummy;
      float illum2;

      planetInfo(dataDir, EARTH, tJD, &dummy, &dummy, &illum2, &mag);
      planetInfo(dataDir, MOON, tJD+0.04, &dummy, &dummy, &illum2, &mag);
      if (northernHemisphere) {
	if (illum2 < illum)
	  sunAngle = M_PI;
      } else {
	if (illum2 > illum)
	  sunAngle = M_PI;
      }
      if (fabs(illum - 0.5) < 0.03) {
	if (illum2 > illum)
	  sprintf(scratchString, "First Quarter");
	else
	  sprintf(scratchString, "Last Quarter");
      } else if (illum > 0.5) {
	if (illum2 > illum)
	  sprintf(scratchString, "Waxing Gibbous");
	else
	  sprintf(scratchString, "Waining Gibbous");
      } else {
	if (illum2 > illum)
	  sprintf(scratchString, "Waxing Crescent");
	else
	  sprintf(scratchString, "Waining Crescent");
      }
    }
    gdk_draw_string(pixmap, smallFont, whiteGC,
		    (displayWidth/2) + 155 - MOON_LABEL_SHIFT,
		    row*ABOUT_ROW_STEP + BIG_FONT_OFFSET, scratchString);
    drawMoon(13.0, illum, sunAngle, (displayWidth/2) + 34 - MOON_LABEL_SHIFT,
	     row*ABOUT_ROW_STEP + BIG_FONT_OFFSET - 4, FALSE, FALSE, FALSE);
    row += 2;
    rAHours = sunRA/HOURS_TO_RADIANS; decDegrees = sunDec/DEGREES_TO_RADIANS;
    rAHH = (int)rAHours;
    rAMM = (int)((rAHours - (float)rAHH)*60.0);
    gdk_draw_string(pixmap, smallFont, creamGC,
		    0,
		    row*ABOUT_ROW_STEP + BIG_FONT_OFFSET, "RA");
    sprintf(scratchString, "%02d:%02d", rAHH, rAMM);
    gdk_draw_string(pixmap, smallFont, whiteGC,
		    gdk_string_width(smallFont, "RA "),
		    row*ABOUT_ROW_STEP + BIG_FONT_OFFSET, scratchString);
    decDD = (int)decDegrees;
    decMM = (int)((decDegrees - (float)decDD)*60.0);
    gdk_draw_string(pixmap, smallFont, creamGC,
		    displayWidth/5,
		    row*ABOUT_ROW_STEP + BIG_FONT_OFFSET, "Dec");
    if (decDegrees > 0.0)
      sprintf(scratchString, "%02d:%02d", decDD, decMM);
    else
      sprintf(scratchString, "-%02d:%02d", -decDD, -decMM);
    gdk_draw_string(pixmap, smallFont, whiteGC,
		    displayWidth/5 + gdk_string_width(smallFont, "Dec "),
		    row*ABOUT_ROW_STEP + BIG_FONT_OFFSET, scratchString);
    hA = lSTNow - sunRA;
    rAHours = hA/HOURS_TO_RADIANS;
    gdk_draw_string(pixmap, smallFont, creamGC,
		    2*displayWidth/5,
		    row*ABOUT_ROW_STEP + BIG_FONT_OFFSET, "HA");
    sprintf(scratchString, "%7.3f", rAHours);
    gdk_draw_string(pixmap, smallFont, whiteGC,
		    2*displayWidth/5 + gdk_string_width(smallFont, "HA "),
		    row*ABOUT_ROW_STEP + BIG_FONT_OFFSET, scratchString);
    azZA(sunRA, sin(sunDec), cos(sunDec), &az, &zA, FALSE);
    az /= DEGREES_TO_RADIANS;
    if (az < 0.0)
      az += 360.0;
    if (az > 360.0)
      az -= 360.0;
    gdk_draw_string(pixmap, smallFont, creamGC,
		    3*displayWidth/5,
		    row*ABOUT_ROW_STEP + BIG_FONT_OFFSET, "Az");
    sprintf(scratchString, "%7.3f", az);
    gdk_draw_string(pixmap, smallFont, whiteGC,
		    3*displayWidth/5 + gdk_string_width(smallFont, "Az "),
		    row*ABOUT_ROW_STEP + BIG_FONT_OFFSET, scratchString);
    el = M_PI*0.5 - zA;
    el += refraction(el);
    gdk_draw_string(pixmap, smallFont, creamGC,
		    4*displayWidth/5,
		    row*ABOUT_ROW_STEP + BIG_FONT_OFFSET, "El");
    sprintf(scratchString, "%7.3f", el/DEGREES_TO_RADIANS);
    gdk_draw_string(pixmap, smallFont, whiteGC,
		    4*displayWidth/5 + gdk_string_width(smallFont, "El "),
		    row*ABOUT_ROW_STEP + BIG_FONT_OFFSET, scratchString);
    row += 1;
    /* Calculate time of moon rise, transit and setting */
    nextRiseDone = nextSetDone = FALSE;
    h0 = 0.125*DEGREES_TO_RADIANS;
    for (day = 0; day < N_RISE_SET_DAYS; day++) {
      tJDToDate(tJDS[day], &year, &month, &dayNum);
      sprintf(scratchString, "%02d/%02d/%04d", month, dayNum, year);
      if (day - N_RISE_SET_DAYS/2 == 0)
	gdk_draw_string(pixmap, smallFont, whiteGC,
			MOONLINE_X_OFFSET, row*ABOUT_ROW_STEP + BIG_FONT_OFFSET, scratchString);
      else
	gdk_draw_string(pixmap, smallFont, greyGC,
			MOONLINE_X_OFFSET, row*ABOUT_ROW_STEP + BIG_FONT_OFFSET, scratchString);
      lSTNow = lSTAtTJD(tJDS[day]);
      /* Moonrise */
      theTime = calcRiseOrSetTime(TRUE, MOON, tJDS[day], lSTNow, h0, &circumpolar, &hD,
				  &cosH, &sunDec, &illumR);
      if (circumpolar != 0) {
	gdk_draw_string(pixmap, smallFont, darkCreamGC,
			MOONLINE_X_OFFSET+gdk_string_width(smallFont, "88/88/8888  "),
			row*ABOUT_ROW_STEP + BIG_FONT_OFFSET, "Moonrise");
	gdk_draw_string(pixmap, smallFont, greyGC,
			MOONLINE_X_OFFSET+gdk_string_width(smallFont, "88/88/8888  Moonrise "),
			row*ABOUT_ROW_STEP + BIG_FONT_OFFSET, "(none)");
      } else {
	theAz = atan2(sin(hD), cosH*sinLatitude-tan(sunDec)*cosLatitude) + M_PI;
	if (theAz < 0.0)
	  theAz += M_2PI;
	if (theAz > M_2PI)
	  theAz -= M_2PI;
	if ((!nextRiseDone) && (theTime - tJD >= 0.0)) {
	  labelGC = creamGC;
	  textGC = whiteGC;
	  nextRiseDone = TRUE;
	} else {
	  labelGC = darkCreamGC;
	  textGC = greyGC;
	}
	theTime -= 0.5;
	theTime = theTime - (double)((int)theTime);
	rAHours = theTime*24.0;
	rAHH = (int)rAHours;
	rAMM = (int)((rAHours - (float)rAHH)*60.0 + 0.5);
	sprintf(scratchString, "%02d:%02d ", rAHH, rAMM);
	gdk_draw_string(pixmap, smallFont, labelGC,
			MOONLINE_X_OFFSET+gdk_string_width(smallFont, "88/88/8888  "),
			row*ABOUT_ROW_STEP + BIG_FONT_OFFSET, "Moonrise");
	gdk_draw_string(pixmap, smallFont, textGC,
			MOONLINE_X_OFFSET+gdk_string_width(smallFont, "88/88/8888  Moonrise "),
			row*ABOUT_ROW_STEP + BIG_FONT_OFFSET, scratchString);
	gdk_draw_string(pixmap, smallFont, labelGC,
			MOONLINE_X_OFFSET+gdk_string_width(smallFont, "88/88/8888  Moonrise 88:88 "),
			row*ABOUT_ROW_STEP + BIG_FONT_OFFSET, "UT Az =");
	sprintf(scratchString, "%3.0f", theAz/DEGREES_TO_RADIANS);
	gdk_draw_string(pixmap, smallFont, textGC,
			MOONLINE_X_OFFSET+gdk_string_width(smallFont, "88/88/8888  Moonrise 88:88 UT Az = "),
			row*ABOUT_ROW_STEP + BIG_FONT_OFFSET, scratchString);
      }
      /* Moonset */
      theTime = calcRiseOrSetTime(FALSE, MOON, tJDS[day], lSTNow, h0, &circumpolar, &hD,
				  &cosH, &sunDec, &illumS);
      if (circumpolar != 0) {
	gdk_draw_string(pixmap, smallFont, darkCreamGC,
			MOONLINE_X_OFFSET+gdk_string_width(smallFont, "88/88/8888  Moonrise 88:88 UT Az = 888  Illum = 88%  "),
			row*ABOUT_ROW_STEP + BIG_FONT_OFFSET, "Moonset");
	gdk_draw_string(pixmap, smallFont, greyGC,
			MOONLINE_X_OFFSET+gdk_string_width(smallFont, "88/88/8888  Moonrise 88:88 UT Az = 888  Illum = 88%  Moonset "),
			row*ABOUT_ROW_STEP + BIG_FONT_OFFSET, "(none)");
      } else {
	theAz = atan2(sin(hD), cosH*sinLatitude-tan(sunDec)*cosLatitude) + M_PI;
	if (theAz < 0.0)
	  theAz += M_2PI;
	if (theAz > M_2PI)
	  theAz -= M_2PI;
	if ((!nextSetDone) && (theTime - tJD >= 0.0)) {
	  labelGC = creamGC;
	  textGC = whiteGC;
	  nextSetDone = TRUE;
	} else {
	  labelGC = darkCreamGC;
	  textGC = greyGC;
	}
	theTime -= 0.5;
	theTime = theTime - (double)((int)theTime);
	rAHours = theTime*24.0;
	rAHH = (int)rAHours;
	rAMM = (int)((rAHours - (float)rAHH)*60.0 + 0.5);
	sprintf(scratchString, "%02d:%02d ", rAHH, rAMM);
	gdk_draw_string(pixmap, smallFont, labelGC,
			MOONLINE_X_OFFSET+gdk_string_width(smallFont, "88/88/8888  Moonrise 88:88 UT Az = 888  Illum = 88%  "),
			row*ABOUT_ROW_STEP + BIG_FONT_OFFSET, "Moonset");
	gdk_draw_string(pixmap, smallFont, textGC,
			MOONLINE_X_OFFSET+gdk_string_width(smallFont, "88/88/8888  Moonrise 88:88 UT Az = 888  Illum = 88%  Moonset "),
			row*ABOUT_ROW_STEP + BIG_FONT_OFFSET, scratchString);
	gdk_draw_string(pixmap, smallFont, labelGC,
			MOONLINE_X_OFFSET+gdk_string_width(smallFont, "88/88/8888  Moonrise 88:88 UT Az = 888  Illum = 88%  Moonset 88:88 "),
			row*ABOUT_ROW_STEP + BIG_FONT_OFFSET, "UT Az =");
	sprintf(scratchString, "%3.0f", theAz/DEGREES_TO_RADIANS);
	gdk_draw_string(pixmap, smallFont, textGC,
			MOONLINE_X_OFFSET+gdk_string_width(smallFont, "88/88/8888  Moonrise 88:88 UT Az = 888  Illum = 88%  Moonset 88:88 UT Az = "),
			row*ABOUT_ROW_STEP + BIG_FONT_OFFSET, scratchString);
      }
      sprintf(scratchString, "Illum = %2.0f%%", (illumR+illumS)*50.0);
      gdk_draw_string(pixmap, smallFont, textGC,
		      MOONLINE_X_OFFSET+gdk_string_width(smallFont, "88/88/8888  Moonrise 88:88 UT Az = 888  "),
		      row*ABOUT_ROW_STEP + BIG_FONT_OFFSET, scratchString);
      row++;
    }
    row += 2;
    {
      /* Calculate time and date of the next four main moon phases. */
      int year, month, day, iDay, hH, mM, i;
      GdkGC *timeGC;

      sprintf(scratchString, "UT Dates and Times of Moon Phases");
      gdk_draw_string(pixmap, smallFont, creamGC,
		      displayWidth/2 - gdk_string_width(smallFont, scratchString)/2,
		      row*ABOUT_ROW_STEP + BIG_FONT_OFFSET, scratchString);
      row += 2;
      sprintf(scratchString, "New Moon");
      gdk_draw_string(pixmap, smallFont, creamGC,
		      displayWidth/8 - gdk_string_width(smallFont, scratchString)/2,
		      row*ABOUT_ROW_STEP + BIG_FONT_OFFSET, scratchString);
      sprintf(scratchString, "First Quarter");
      gdk_draw_string(pixmap, smallFont, creamGC,
		      3*displayWidth/8 - gdk_string_width(smallFont, scratchString)/2,
		      row*ABOUT_ROW_STEP + BIG_FONT_OFFSET, scratchString);
      sprintf(scratchString, "Full Moon");
      gdk_draw_string(pixmap, smallFont, creamGC,
		      5*displayWidth/8 - gdk_string_width(smallFont, scratchString)/2,
		      row*ABOUT_ROW_STEP + BIG_FONT_OFFSET, scratchString);
      sprintf(scratchString, "Last Quarter");
      gdk_draw_string(pixmap, smallFont, creamGC,
		      7*displayWidth/8 - gdk_string_width(smallFont, scratchString)/2,
		      row*ABOUT_ROW_STEP + BIG_FONT_OFFSET, scratchString);

      tJDToDate(tJD, &year, &month, &day);
      nextPhases(tJD, N_PHASE_DATES, &nextNew[0], &nextFirstQ[0], &nextFull[0], &nextLastQ[0]);

      row++;
      for (i = 0; i < N_PHASE_DATES; i++) {
	if (i == N_PHASE_DATES/2)
	  timeGC = whiteGC;
	else
	  timeGC = greyGC;
	iDay = (int)nextNew[i];
	tempDay = nextNew[i] - 0.5 - (double)iDay;
	if (tempDay < 0.0)
	  tempDay += 1.0;
	hH = (int)(tempDay * 24.0);
	mM = (int)((tempDay * 24.0 - (double)hH) * 60.0);
	tJDToDate(nextNew[i], &year, &month, &day);
	sprintf(scratchString, "%02d/%02d/%4d %02d:%02d",
		month, day, year, hH, mM);
	gdk_draw_string(pixmap, smallFont, timeGC,
			displayWidth/8 - gdk_string_width(smallFont, scratchString)/2,
			row*ABOUT_ROW_STEP + BIG_FONT_OFFSET, scratchString);
	iDay = (int)nextFirstQ[i];
	tempDay = nextFirstQ[i] - 0.5 - (double)iDay;
	if (tempDay < 0.0)
	  tempDay += 1.0;
	hH = (int)(tempDay * 24.0);
	mM = (int)((tempDay * 24.0 - (double)hH) * 60.0);
	tJDToDate(nextFirstQ[i], &year, &month, &day);
	sprintf(scratchString, "%02d/%02d/%4d %02d:%02d",
		month, day, year, hH, mM);
	gdk_draw_string(pixmap, smallFont, timeGC,
			3*displayWidth/8 - gdk_string_width(smallFont, scratchString)/2,
			row*ABOUT_ROW_STEP + BIG_FONT_OFFSET, scratchString);
	iDay = (int)nextFull[i];
	tempDay = nextFull[i] - 0.5 - (double)iDay;
	if (tempDay < 0.0)
	  tempDay += 1.0;
	hH = (int)(tempDay * 24.0);
	mM = (int)((tempDay * 24.0 - (double)hH) * 60.0);
	tJDToDate(nextFull[i], &year, &month, &day);
	sprintf(scratchString, "%02d/%02d/%4d %02d:%02d",
		month, day, year, hH, mM);
	gdk_draw_string(pixmap, smallFont, timeGC,
			5*displayWidth/8 - gdk_string_width(smallFont, scratchString)/2,
			row*ABOUT_ROW_STEP + BIG_FONT_OFFSET, scratchString);
	iDay = (int)nextLastQ[i];
	tempDay = nextLastQ[i] - 0.5 - (double)iDay;
	if (tempDay < 0.0)
	  tempDay += 1.0;
	hH = (int)(tempDay * 24.0);
	mM = (int)((tempDay * 24.0 - (double)hH) * 60.0);
	tJDToDate(nextLastQ[i], &year, &month, &day);
	sprintf(scratchString, "%02d/%02d/%4d %02d:%02d",
		month, day, year, hH, mM);
	gdk_draw_string(pixmap, smallFont, timeGC,
			7*displayWidth/8 - gdk_string_width(smallFont, scratchString)/2,
			row*ABOUT_ROW_STEP + BIG_FONT_OFFSET, scratchString);
	row++;
      }
    }
    break;
  case SMALL_MOONCAL_SCREEN:
    {
      static int firstCall = TRUE;
      int thisMonth, thisYear, iDummy, day, testMonth, phase, i, nextMonth, lastMonth;
      int fullSeen = FALSE;
      int blue = FALSE;
      int week = 0;
      int weekDay = 0;
      int xArrowOffset = 0;
      int lastWeekDay = -1;
      double tJDS, tJDM, tJDO, plusOneMonth, minusOneMonth;
      char *monthName[12] = {"January",   "February", "March",    "April",
			     "May",       "June",     "July",     "August",
			     "September", "October",  "November", "December"};
      char *dayName[7] = {"Sunday", "Monday", "Tuesday", "Wednesday",
			  "Thursday", "Friday", "Saturday"};

      if (firstCall) {
	char fileName[MAX_FILE_NAME_SIZE];

	sprintf(fileName, "%s/icons/%s.xpm", dataDir, "moonImage");
	moonImage = gdk_pixmap_create_from_xpm(pixmap, NULL, NULL, fileName);
	sprintf(fileName, "%s/icons/%s.xpm", dataDir, "moonImageFlipped");
	moonImageFlipped = gdk_pixmap_create_from_xpm(pixmap, NULL, NULL, fileName);
	sprintf(fileName, "%s/icons/%s.xpm", dataDir, "blueMoonImage");
	blueMoonImage = gdk_pixmap_create_from_xpm(pixmap, NULL, NULL, fileName);
	sprintf(fileName, "%s/icons/%s.xpm", dataDir, "blueMoonImageFlipped");
	blueMoonImageFlipped = gdk_pixmap_create_from_xpm(pixmap, NULL, NULL, fileName);
	sprintf(fileName, "%s/icons/%s.xpm", dataDir, "fullMoonImage");
	fullMoonImage = gdk_pixmap_create_from_xpm(pixmap, NULL, NULL, fileName);
	sprintf(fileName, "%s/icons/%s.xpm", dataDir, "fullMoonImageFlipped");
	fullMoonImageFlipped = gdk_pixmap_create_from_xpm(pixmap, NULL, NULL, fileName);
	firstCall = FALSE;
      }
      tJDO = tJD + smallMooncalDateOffset;
      nextPhases(tJDO, N_PHASE_DATES, &nextNew[0], &nextFirstQ[0], &nextFull[0], &nextLastQ[0]);
      tJDToDate(tJDO, &thisYear, &thisMonth, &iDummy);
      if ((thisYear % 4) == 0)
	monthLengths[1] += 1.0;
      sprintf(scratchString, "%s %d", monthName[thisMonth-1], thisYear);
      gdk_draw_string(pixmap, smallFont, creamGC,
		      displayWidth/2 - gdk_string_width(smallFont, scratchString)/2,
		      12, scratchString);
      for (day = 0; day < 7; day++) {
	sprintf(scratchString, "%s", dayName[day]);
	gdk_draw_string(pixmap, smallFont, blueGC,
			CAL_DAY_WIDTH/2 + day*CAL_DAY_WIDTH + CAL_LEFT_OFFSET -
			gdk_string_width(smallFont, scratchString)/2,
			30, scratchString);	
      }
      /* Generate days of the week */
      tJDM = (double)((int)(tJDO - 0.5)) + 0.5;
      for (tJDS = tJDM - 32.0; tJDS < tJDM + 32.0; tJDS += 1.0) {
	tJDToDate(tJDS, &thisYear, &testMonth, &day);
	if (testMonth == thisMonth) {
	  int today;

	  weekDay = dayOfWeek(thisYear, thisMonth, day);
	  if (weekDay < lastWeekDay)
	    week++;
	  lastWeekDay = weekDay;
	  if ((fabs(tJDS - tJDM) < 0.1) && (smallMooncalDateOffset == 0.0))
	    today = TRUE;
	  else
	    today = FALSE;
	  phase = -1;
	  for (i = 0; i < N_PHASE_DATES; i++)
	    if (fabs(nextNew[i] - tJDS - 0.5) < 0.5)
	      phase = 0;
	    else if (fabs(nextFirstQ[i] - tJDS - 0.5) < 0.5)
	      phase = 1;
	    else if (fabs(nextFull[i] - tJDS - 0.5) < 0.5) {
	      if (fullSeen)
		blue = TRUE;
	      else
		fullSeen = TRUE;
	      phase = 2;
	    } else if (fabs(nextLastQ[i] - tJDS - 0.5) < 0.5)
	      phase = 3;
	  plotDayBox(tJDS, week, weekDay, day, today, phase, blue);
	}
      }
      lastMonth = thisMonth-1;
      if (lastMonth < 1)
	lastMonth = 12;
      nextMonth = thisMonth + 1;
      if (nextMonth > 12)
	nextMonth = 1;
      plusOneMonth = monthLengths[thisMonth-1] - iDummy + 1;
      minusOneMonth = -iDummy;
      removeAllSensitiveAreas();
      if (week > 4)
	xArrowOffset = 50 + (weekDay+1)*CAL_DAY_WIDTH;
      points[0].x = xArrowOffset;                  points[0].y = 15*displayHeight/16 - 15;
      points[1].x = displayWidth/6 + xArrowOffset; points[1].y = points[0].y - 25;
      points[2].x = points[1].x;                   points[2].y = points[0].y + 25;
      gdk_draw_polygon(pixmap, creamGC, TRUE, points, 3);
      sprintf(scratchString, "%s", monthName[lastMonth-1]);
      gdk_draw_string(pixmap, smallFont, blackGC,
		      (points[0].x + points[1].x)/2 - gdk_string_width(smallFont, scratchString)/2 + 10,
		      points[0].y + 4, scratchString);
      dprintf("Adding sensitive area for left arrow %d, %d, %d, %d\n",
	      points[0].x-5, points[1].y+5, points[1].x+5, points[2].y-5);
      addSensitiveArea(FALSE, SA_MONTH_LEFT_ARROW,
		       points[0].x-5, points[1].y+5, points[1].x+5, points[2].y-5, minusOneMonth);
      points[0].x = displayWidth;                 points[0].y = 15*displayHeight/16 - 15;
      points[1].x = points[0].x - displayWidth/6; points[1].y = points[0].y - 25;
      points[2].x = points[1].x;                  points[2].y = points[0].y + 25;
      gdk_draw_polygon(pixmap, creamGC, TRUE, points, 3);
      sprintf(scratchString, "%s", monthName[nextMonth-1]);
      gdk_draw_string(pixmap, smallFont, blackGC,
		      (points[0].x + points[1].x)/2 - gdk_string_width(smallFont, scratchString)/2 - 10,
		      points[0].y + 4, scratchString);
      dprintf("Adding sensitive area for right arrow %d, %d, %d, %d\n",
	      points[1].x-5, points[1].y+5, points[0].x+5, points[2].y-5);
      addSensitiveArea(FALSE, SA_MONTH_RIGHT_ARROW,
		       points[1].x-5, points[1].y+5, points[0].x+5, points[2].y-5, plusOneMonth);
      if (xArrowOffset == 0) {
	sprintf(scratchString, "Tap arrows to change month");
	stringWidth = gdk_string_width(smallFont, scratchString);
	gdk_draw_string(pixmap, smallFont, blueGC,
			(displayWidth/2)-stringWidth/2,
			displayHeight-5*ABOUT_ROW_STEP, scratchString);
	
	sprintf(scratchString, "Tap elsewhere to exit this display");
	stringWidth = gdk_string_width(smallFont, scratchString);
	gdk_draw_string(pixmap, smallFont, blueGC,
			(displayWidth/2)-stringWidth/2,
			displayHeight-3*ABOUT_ROW_STEP, scratchString);
      }
    }
    break;
  case BIG_MOONCAL_SCREEN:
    /*
      Draw a moon calendar.   I will make a calendar with 21 months of moon
      phases, centered on the current month.
    */
    {
      int centerMonth, centerYear, startMonth, endMonth, iDummy, i, col, lastMonth, centerDay;
      int blue = FALSE;
      int fullSeen = FALSE;
      double tempTJD, startTJD, endTJD;
      char *monthLetters[12] = {"J","F","M","A","M","J","J","A","S","O","N","D"};
      GdkGC *labelGC;

      nextPhases(tJD, MAX_PHASE_DATES, &nextNew[0], &nextFirstQ[0], &nextFull[0], &nextLastQ[0]);
      needNewTime = TRUE;
      lSTNow = lST();
      tJDToDate(tJD, &centerYear, &centerMonth, &centerDay);
      startMonth = centerMonth - 10;
      if (startMonth < 1)
	startMonth += 12;
      endMonth = centerMonth + 11;
      if (endMonth > 12)
	endMonth -= 12;
      startTJD = tJD - 341;
      i = 0;
      while (i != startMonth) {
	startTJD += 1.0;
	tJDToDate(startTJD, &iDummy, &i, &iDummy);
      }
      endTJD = tJD + 250;
      i = 0;
      while (i != endMonth) {
	endTJD += 1.0;
	tJDToDate(endTJD, &iDummy, &i, &iDummy);
      }
      inFullscreenMode = TRUE;
      gtk_window_fullscreen(GTK_WINDOW(window));

      for (i = startMonth-1; i < startMonth+20; i++) {
	if ((i % 12) == centerMonth-1)
	  labelGC = creamGC;
	else
	  labelGC = blueGC;
	gdk_draw_string(pixmap, smallFont, labelGC, 28+(i - startMonth + 1)*21, 28, monthLetters[i % 12]);
	gdk_draw_string(pixmap, smallFont, labelGC, 28+(i - startMonth + 1)*21, displayHeight-8, monthLetters[i % 12]);
	if ((i % 12) == 11) {
	  gdk_draw_line(pixmap, greenGC, 39+(i - startMonth + 1)*21, 0, 39+(i - startMonth + 1)*21, displayHeight-1);
	  gdk_draw_line(pixmap, greenGC, 41+(i - startMonth + 1)*21, 0, 41+(i - startMonth + 1)*21, displayHeight-1);
	}
      }
      for (i = 1; i < 32; i++) {
	sprintf(scratchString, "%2d", i);
	if (i == centerDay)
	  labelGC = creamGC;
	else
	  labelGC = blueGC;
	gdk_draw_string(pixmap, smallFont, labelGC, 0, 26+i*19, scratchString);
	gdk_draw_string(pixmap, smallFont, labelGC, displayWidth-15, 26+i*19, scratchString);
      }
      /* Loop over all days in calendar */
      col = 0; lastMonth = startMonth;
      for (tempTJD = startTJD; tempTJD < endTJD; tempTJD += 1.0) {
	int month, row, full, ii;
	float illum, illum2, sunAngle;
	double dummy;

	tJDToDate(tempTJD, &iDummy, &month, &row);
	if (month != lastMonth) {
	  fullSeen = blue = FALSE;
	  col++;
	  lastMonth = month;
	}
	if ((month == 6) && (row == 1)) {
	  sprintf(scratchString, "%4d", iDummy);
	  gdk_draw_string(pixmap, smallFont, greenGC, 30+col*21, 10, scratchString);
	}
	planetInfo(dataDir, EARTH, tempTJD, &dummy, &dummy, &illum2, &mag);
	planetInfo(dataDir, MOON, tempTJD+0.04, &dummy, &dummy, &illum2, &mag);
	planetInfo(dataDir, MOON, tempTJD, &dummy, &dummy, &illum, &mag);
	if (northernHemisphere) {
	  if (illum < illum2)
	    sunAngle = 0.0;
	  else
	    sunAngle = M_PI;
	} else {
	  if (illum > illum2)
	    sunAngle = 0.0;
	  else
	    sunAngle = M_PI;
	}
	full = FALSE;
	for (ii = 0; ii < MAX_PHASE_DATES; ii++)
	  if (fabs(nextFull[ii] - tempTJD) <= 0.5) {
	    full = TRUE;
	    break;
	  }
	/*
	  Handle "blue moons" usung the definition that a blue moon is the second
	  full moon in a calendar month.
	*/
	if (full) {
	  if (fullSeen)
	    blue = TRUE;
	  else
	    fullSeen = TRUE;
	}
	drawMoon(9.0, illum, sunAngle, 30+col*21, 22+row*19, FALSE, full, blue);
	if (fabs(tempTJD - tJD) < 0.1) {
	  GdkPoint points[4];

	  points[0].x = 20+col*21;      points[0].y = 12+row*19;
	  points[1].x = points[0].x+20; points[1].y = points[0].y;
	  points[2].x = points[1].x;    points[2].y = points[1].y+20;
	  points[3].x = points[0].x;    points[3].y = points[2].y;
	  gdk_draw_polygon(pixmap, redGC, FALSE, points, 4);
	}
      }
    }
    break;
  case SOLAR_SYSTEM_SCHEMATIC_SCREEN:
    schematic = TRUE;
  case SOLAR_SYSTEM_SCALE_SCREEN:
#define SOLAR_SYSTEM_CENTER_X (263)
#define SOLAR_SYSTEM_CENTER_Y (280)
#define SCHEMATIC_ORBIT_INCREMENT (23)
#define ZOOM_SOLAR_SYSTEM_IN (-10000)
#define ZOOM_SOLAR_SYSTEM_OUT (10000)
    {
      int i, j, w, h, k, m, n, eD, year, month, day, tDayIncSign, plotPlanet;
      int ii, x1, x2, y1, y2;
      int plotStart, plotStop, plotInc;
      int tDayInc = 1;
      int solarSystemButtonHeight, solarSystemButtonWidth;
      int x = 0;
      int y = 0;
      static int firstCall = TRUE;
      static int havePlottedOrbits = FALSE;
      int firstIteration = TRUE;
      double eX, eY, eZ, eHyp, dDummy, eLong, dayIncSign;
      double a, e, I, L, smallOmega, bigOmega;
      float fDummy, eR;
      float orbitScale = 0.0;
      float eSin = 0.0;
      float eCos = 0.0;
      double sunELong = 0.0;
      char fileName[MAX_FILE_NAME_SIZE];
      GdkPoint points[4];

      removeAllSensitiveAreas();
      if (dayInc < 0)
	dayIncSign = -1.0;
      else
	dayIncSign = 1.0;
      nSolarSystemDays = abs(dayInc);
      j = 0;
      do {
	if (j == 0)
	  gdk_draw_rectangle(pixmap, drawingArea->style->black_gc,
			     TRUE, 0, 0,
			     drawingArea->allocation.width,
			     drawingArea->allocation.height);
	else
	  gdk_draw_rectangle(pixmap, drawingArea->style->black_gc,
			     TRUE, 0, BIG_FONT_OFFSET,
			     drawingArea->allocation.width,
			     drawingArea->allocation.height*6/7 - 14 - BIG_FONT_OFFSET);
	solarSystemButtonWidth = displayWidth/5 - 1;
	solarSystemButtonHeight = displayHeight/14 + 5;
	if (!schematic) {
	  /*
	    Draw the "in" and out" buttons which allow the user to zoom in and
	    out of the inner Solar System - this is required because with the
	    "to-scale" view makes the outer planet orbits so large, the inner
	    orbits cannot be displayed.
	  */
	  if (outermostPlanet > 2) {
	    points[0].x = 0;
	    points[0].y = displayHeight-2*solarSystemButtonHeight - 5;
	    points[1].x = points[0].x;
	    points[1].y = points[0].y-solarSystemButtonHeight;
	    points[2].x = points[1].x+4*solarSystemButtonWidth/5;
	    points[2].y = points[1].y;
	    points[3].x = points[2].x;
	    points[3].y = points[2].y+solarSystemButtonHeight;
	    strcpy(scratchString, "Zoom In");
	    n = gdk_string_width(smallFont, scratchString);
	    gdk_draw_polygon(pixmap, darkCreamGC, FALSE, points, 4);
	    gdk_draw_string(pixmap, smallFont, whiteGC,
			    (points[0].x + points[2].x - n)/2,
			    (points[0].y + points[1].y)/2 + 4, scratchString);
	    tDayInc = ZOOM_SOLAR_SYSTEM_IN;
	    points[2].x = points[1].x+solarSystemButtonWidth;
	    points[2].y += 10;
	    addSensitiveArea(FALSE, SA_SOLAR_SYSTEM_BUTTON,
			     points[0].x, points[2].y, points[2].x,
			     points[0].y, (float)tDayInc);
	  }
	  if (outermostPlanet < N_SOLAR_SYSTEM_OBJECTS-1) {
	    points[0].x = (displayWidth-1) - 4*solarSystemButtonWidth/5 - 4;
	    points[0].y = displayHeight-2*solarSystemButtonHeight - 5;
	    points[1].x = points[0].x;
	    points[1].y = points[0].y-solarSystemButtonHeight;
	    points[2].x = displayWidth-5;
	    points[2].y = points[1].y;
	    points[3].x = points[2].x;
	    points[3].y = points[2].y+solarSystemButtonHeight;
	    strcpy(scratchString, "Zoom Out");
	    n = gdk_string_width(smallFont, scratchString);
	    gdk_draw_polygon(pixmap, darkCreamGC, FALSE, points, 4);
	    gdk_draw_string(pixmap, smallFont, whiteGC,
			    (points[0].x + points[2].x - n)/2,
			    (points[0].y + points[1].y)/2 + 4, scratchString);
	    tDayInc = ZOOM_SOLAR_SYSTEM_OUT;
	    points[0].x = (displayWidth-1) - solarSystemButtonWidth - 4;
	    points[2].y += 10;
	    addSensitiveArea(FALSE, SA_SOLAR_SYSTEM_BUTTON,
			     points[0].x, points[2].y, points[2].x,
			     points[0].y, (float)tDayInc);
	  }
	}
	if ((j == 0) || (j == nSolarSystemDays-1)) {
	  for (k = 0; k < 2; k++)
	    for (m = 0; m < 5; m++) {
	      points[0].x = m*solarSystemButtonWidth;
	      points[0].y = displayHeight-k*solarSystemButtonHeight - 5;
	      points[1].x = points[0].x;
	      points[1].y = points[0].y-solarSystemButtonHeight;
	      points[2].x = points[1].x+solarSystemButtonWidth;
	      points[2].y = points[1].y;
	      points[3].x = points[2].x;
	      points[3].y = points[2].y+solarSystemButtonHeight;
	      if (k == 0) {
		tDayIncSign = -1;
		planetGC = redGC;
	      } else {
		tDayIncSign = 1;
		planetGC = whiteGC;
	      }
	      switch (m) {
	      case 0:
		tDayInc = tDayIncSign*2;
		if (k == 0)
		  strcpy(scratchString, "-1 day");
		else
		  strcpy(scratchString, "+1 day");
		break;
	      case 1:
		tDayInc = tDayIncSign*11;
		if (k == 0)
		  strcpy(scratchString, "-10 days");
		else
		  strcpy(scratchString, "+10 days");
		break;
	      case 2:
		tDayInc = tDayIncSign*101;
		if (k == 0)
		  strcpy(scratchString, "-100 days");
		else
		  strcpy(scratchString, "+100 days");
		break;
	      case 3:
		tDayInc = tDayIncSign*366;
		if (k == 0)
		  strcpy(scratchString, "-1 year");
		else
		  strcpy(scratchString, "+1 year");
		break;
	      case 4:
		tDayInc = tDayIncSign*3653;
		if (k == 0)
		  strcpy(scratchString, "-10 years");
		else
		  strcpy(scratchString, "+10 years");
		break;
	      }
	      n = gdk_string_width(smallFont, scratchString);
	      if ((j == 0) && (nSolarSystemDays > 1) &&
		  (nSolarSystemDays == abs(tDayInc)) &&
		  ((int)dayIncSign == tDayIncSign)) {
		gdk_draw_polygon(pixmap, darkCreamGC, TRUE, points, 4);
		gdk_draw_string(pixmap, smallFont, planetGC,
				(points[0].x + points[2].x - n)/2,
				(points[0].y + points[1].y)/2 + 4, scratchString);
		gdk_draw_drawable(drawingArea->window,
				  drawingArea->style->fg_gc[GTK_WIDGET_STATE (drawingArea)],
				  pixmap, points[0].x, points[1].y, points[0].x, points[1].y,
				  points[2].x-points[0].x, points[0].y-points[1].y);
	      } else if ((j == nSolarSystemDays - 1) && (nSolarSystemDays > 1) &&
			 (nSolarSystemDays == abs(tDayInc)) &&
			 ((int)dayIncSign == tDayIncSign)) {
		gdk_draw_rectangle(pixmap, drawingArea->style->black_gc,
				   TRUE, points[0].x, points[1].y,
				   points[2].x-points[0].x, points[0].y-points[1].y);
		gdk_draw_polygon(pixmap, greyGC, FALSE, points, 4);
	      } else if (j == 0)
		gdk_draw_polygon(pixmap, greyGC, FALSE, points, 4);
	      addSensitiveArea(FALSE, SA_SOLAR_SYSTEM_BUTTON,
			       points[0].x, points[2].y, points[2].x,
			       points[0].y, (float)tDayInc);
	      gdk_draw_string(pixmap, smallFont, planetGC,
			      (points[0].x + points[2].x - n)/2,
			      (points[0].y + points[1].y)/2 + 4, scratchString);
	    }
	}
	tJDToDate(tJD, &year, &month, &day);
	if (schematic)
	  sprintf(fileName, "Solar System Schematic View on %02d/%02d/%04d", month, day, year);
	else
	  sprintf(fileName, "Solar System on %02d/%02d/%04d", month, day, year);
	k = gdk_string_width(smallFont, fileName);
	gdk_draw_string(pixmap, smallFont, whiteGC,
			(displayWidth/2)-k/2, 2*BIG_FONT_OFFSET, fileName);
	if (!schematic && !havePlottedOrbits) {
	  for (i = 0; i < N_SOLAR_SYSTEM_OBJECTS; i++) {
	    if ((i <= NEPTUNE) && (i != MOON) && (i != SUN)) {
	      int ii;
	      float r, theta, aOneMinusEE;

	      /*
		Build a structure to hold a plot of the planet's orbit;
	      */
	      getCurrentOrbitalElements(i, tJD, &a, &e, &I, &L, &smallOmega, &bigOmega);
	      orbitPlots[i].omegaRad = smallOmega * DEGREES_TO_RADIANS;
	      orbitPlots[i].peri = a - e*a;
	      aOneMinusEE = a*(1.0 - e*e);
	      orbitPlots[i].x = (float *)malloc(360 * sizeof(float));
	      if (orbitPlots[i].x == NULL) {
		perror("malloc of orbit plot xs");
		exit(ERROR_EXIT);
	      }
	      orbitPlots[i].y = (float *)malloc(360 * sizeof(float));
	      if (orbitPlots[i].y == NULL) {
		perror("malloc of orbit plot ys");
		exit(ERROR_EXIT);
	      }
	      orbitPlots[i].maxX = 0.0;
	      for (ii = 0; ii < 360; ii++) {
		theta = ((float)ii) * DEGREES_TO_RADIANS;
		r = aOneMinusEE/(1.0 + e*cosf(theta - orbitPlots[i].omegaRad));
		orbitPlots[i].x[ii] = -r*cosf(theta);
		orbitPlots[i].y[ii] = r*sinf(theta);
		if (r > orbitPlots[i].maxX)
		  orbitPlots[i].maxX = r;
	      }
	    }
	  }
	}
	if (!schematic) {
	  char scaleLabel[30];

	  if (outermostPlanet-1 == MOON)
	    ii = EARTH;
	  else
	    ii = outermostPlanet-1;
	  orbitScale = ((float)(displayWidth - SCHEMATIC_ORBIT_INCREMENT))
	    / (2.0 * orbitPlots[ii].maxX);
	  x1 = x2 = 7*displayWidth/8 - roundf(0.5*solarSystemScales[ii]*orbitScale/AU);
	  y1 = displayHeight/12 - 3;
	  y2 = y1 + 6;
	  gdk_draw_line(pixmap, blueGC, x1, y1, x2, y2);
	  x1 = x2 = 7*displayWidth/8 + roundf(0.5*solarSystemScales[ii]*orbitScale/AU);
	  gdk_draw_line(pixmap, blueGC, x1, y1, x2, y2);
	  x1 = 7*displayWidth/8 - roundf(0.5*solarSystemScales[ii]*orbitScale/AU);
	  y1 = y2 = displayHeight/12;
	  gdk_draw_line(pixmap, blueGC, x1, y1, x2, y2);
	  if (solarSystemScales[ii] < 1000.0)
	    sprintf(scaleLabel, "%1.0f million km", solarSystemScales[ii]);
	  else
	    sprintf(scaleLabel, "%1.0f billion km", solarSystemScales[ii]*0.001);
	  k = gdk_string_width(smallFont, scaleLabel);
	  gdk_draw_string(pixmap, smallFont, whiteGC,
			  7*displayWidth/8 - k/2, y1 - 10, scaleLabel);
	}
	gdk_drawable_get_size(firstPointImage, &w, &h);
	gdk_draw_drawable(pixmap, blueGC, firstPointImage, 0, 0,
			  displayWidth/8, displayHeight/12 - 11, w, h);
	x1 = 10; x2 = x1 + 40; y1 = y2 = displayHeight/12 - 1; 
	gdk_draw_line(pixmap, blueGC, x1, y1, x2, y2);
	x2 = x1 + 10; y2 = y1 + 5;
	gdk_draw_line(pixmap, blueGC, x1, y1, x2, y2);
	y2 = y1 - 5;
	gdk_draw_line(pixmap, blueGC, x1, y1, x2, y2);
	if (schematic) {
	  plotStart = 0;
	  plotStop = N_SOLAR_SYSTEM_OBJECTS - 1;
	  plotInc = 1;
	} else {
	  plotStop = -1;
	  plotStart = N_SOLAR_SYSTEM_OBJECTS - 1;
	  plotInc = -1;
	}
	for (i = plotStart; i != plotStop; i += plotInc) {
	  plotPlanet = TRUE;
	  if (firstCall) {
	    sprintf(fileName, "%s/icons/%sSmall.xpm", dataDir, solarSystemNames[i]);
	    smallPlanetImages[i] = gdk_pixmap_create_from_xpm(pixmap,
							      NULL, NULL, fileName);
	  }
	  if (((schematic) || ( i < outermostPlanet)) &&
	      !(!schematic && (i == MOON))) {
	    gdk_drawable_get_size(smallPlanetImages[i], &w, &h);
	    /* Choose colors for the planets */
	    switch (i) {
	    case SUN:
	      planetGC = lightYellowGC;
	      break;
	    case MERCURY:
	    case MOON:
	      planetGC = greyGC;
	      break;
	    case VENUS:
	      planetGC = whiteGC;
	      break;
	    case EARTH:
	      planetGC = blueGreenGC;
	      break;
	    case MARS:
	      planetGC = redGC;
	      break;
	    case JUPITER:
	    case SATURN:
	      planetGC = creamGC;
	      break;
	    case URANUS:
	    case NEPTUNE:
	      planetGC = blueGC;
	    }
	    calculatePlanetPosition(i, tJD, &eX, &eY, &eZ);
	    /* Draw the planet */
	    if (i != SUN) {
	      eHyp = sqrt(eX*eX + eY*eY);
	      eD  = roundf(atan2f(eY, eX) / DEGREES_TO_RADIANS);
	      while (eD < 0)
		eD += 360;
	      while (eD >= 360)
		eD -= 360;
	      eSin = (float)(eY/eHyp);
	      eCos = -(float)(eX/eHyp);
	    } else
	      sunELong = atan2(eY, eX);
	    switch (i) {
	    case SUN:
	      x = SOLAR_SYSTEM_CENTER_X;
	      y = SOLAR_SYSTEM_CENTER_Y;
	      gdk_draw_arc(pixmap, planetGC, TRUE,
			   x-33, y-33, 20, 20, 0, FULL_CIRCLE);
	      gdk_draw_drawable(pixmap, blueGC, smallPlanetImages[i], 0, 0,
				x-w/2-23, y-h/2-23, 20, 20);
	      
	      break;
	    case MERCURY:
	    case VENUS:
	      if (schematic) {
		x = SOLAR_SYSTEM_CENTER_X - SCHEMATIC_ORBIT_INCREMENT*(i+1);
		y = SOLAR_SYSTEM_CENTER_Y - SCHEMATIC_ORBIT_INCREMENT*(i+1);
		gdk_draw_arc(pixmap, blueGC, FALSE,
			     x, y, 2*i*SCHEMATIC_ORBIT_INCREMENT, 2*i*SCHEMATIC_ORBIT_INCREMENT, 0, FULL_CIRCLE);
		x = SOLAR_SYSTEM_CENTER_X + roundf(eCos*(float)(i*SCHEMATIC_ORBIT_INCREMENT));
		y = SOLAR_SYSTEM_CENTER_Y + roundf(eSin*(float)(i*SCHEMATIC_ORBIT_INCREMENT));
	      } else {
		int ii, px1, px2, py1, py2;
		GdkPoint orbit[360];

		if (orbitPlots[i].maxX*orbitScale > roundf(1.5*(float)SCHEMATIC_ORBIT_INCREMENT)) {
		  for (ii = 0; ii < 360; ii++) {
		    orbit[ii].x = SOLAR_SYSTEM_CENTER_X + orbitPlots[i].x[ii]*orbitScale - 23;
		    orbit[ii].y = SOLAR_SYSTEM_CENTER_Y + orbitPlots[i].y[ii]*orbitScale - 23;
		  }
		  gdk_draw_polygon(pixmap, blueGC, FALSE, orbit, 360);
		  px1 = SOLAR_SYSTEM_CENTER_X -
		    roundf(((float)(SCHEMATIC_ORBIT_INCREMENT/2))*cos(orbitPlots[i].omegaRad)) - 23;
		  py1 = SOLAR_SYSTEM_CENTER_Y +
		    roundf(((float)(SCHEMATIC_ORBIT_INCREMENT/2))*sin(orbitPlots[i].omegaRad)) - 23;
		  px2 = SOLAR_SYSTEM_CENTER_X -
		    roundf((orbitPlots[i].peri*orbitScale) * cos(orbitPlots[i].omegaRad)) - 23;
		  py2 = SOLAR_SYSTEM_CENTER_Y +
		    roundf((orbitPlots[i].peri*orbitScale) * sin(orbitPlots[i].omegaRad)) - 23;
		  eR = sqrt(orbitPlots[i].x[eD]*orbitPlots[i].x[eD] +
			    orbitPlots[i].y[eD]*orbitPlots[i].y[eD]);
		  gdk_draw_line(pixmap, darkGreyGC, px1, py1, px2, py2);
		  orbit[0].x = px2-1; orbit[0].y = py2-1;
		  orbit[1].x = px2+1; orbit[1].y = py2-1;
		  orbit[2].x = px2+1; orbit[2].y = py2+1;
		  orbit[3].x = px2-1; orbit[3].y = py2+1;
		  gdk_draw_polygon(pixmap, whiteGC, TRUE, orbit, 4);
		  x = SOLAR_SYSTEM_CENTER_X + eR*eCos*orbitScale;
		  y = SOLAR_SYSTEM_CENTER_Y + eR*eSin*orbitScale;
		} else
		  plotPlanet = FALSE;
	      }
	      if (plotPlanet) {
		gdk_draw_arc(pixmap, planetGC, TRUE,
			     x-33, y-33, 20, 20, 0, FULL_CIRCLE);
		gdk_draw_drawable(pixmap, blueGC, smallPlanetImages[i], 0, 0,
				  x-w/2-23, y-h/2-23, 20, 20);
	      }
	      break;
	    case EARTH:
	    case MARS:
	    case JUPITER:
	    case SATURN:
	    case URANUS:
	    case NEPTUNE:
	      if (schematic) {
		x = SOLAR_SYSTEM_CENTER_X - SCHEMATIC_ORBIT_INCREMENT*(i+2);
		y = SOLAR_SYSTEM_CENTER_Y - SCHEMATIC_ORBIT_INCREMENT*(i+2);
		gdk_draw_arc(pixmap, blueGC, FALSE,
			     x, y, 2*(i+1)*SCHEMATIC_ORBIT_INCREMENT, 2*(i+1)*SCHEMATIC_ORBIT_INCREMENT, 0, FULL_CIRCLE);
		x = SOLAR_SYSTEM_CENTER_X + roundf(eCos*(float)((i+1)*SCHEMATIC_ORBIT_INCREMENT));
		y = SOLAR_SYSTEM_CENTER_Y + roundf(eSin*(float)((i+1)*SCHEMATIC_ORBIT_INCREMENT));
	      } else {
		int ii, px1, px2, py1, py2;
		GdkPoint orbit[360];
		
		if (orbitPlots[i].maxX*orbitScale > roundf(1.5*(float)SCHEMATIC_ORBIT_INCREMENT)) {
		  for (ii = 0; ii < 360; ii++) {
		    orbit[ii].x = SOLAR_SYSTEM_CENTER_X + orbitPlots[i].x[ii]*orbitScale - 23;
		    orbit[ii].y = SOLAR_SYSTEM_CENTER_Y + orbitPlots[i].y[ii]*orbitScale - 23;
		  }
		  gdk_draw_polygon(pixmap, blueGC, FALSE, orbit, 360);
		  px1 = SOLAR_SYSTEM_CENTER_X -
		    roundf(((float)(SCHEMATIC_ORBIT_INCREMENT/2))*cos(orbitPlots[i].omegaRad)) - 23;
		  py1 = SOLAR_SYSTEM_CENTER_Y +
		    roundf(((float)(SCHEMATIC_ORBIT_INCREMENT/2))*sin(orbitPlots[i].omegaRad)) - 23;
		  px2 = SOLAR_SYSTEM_CENTER_X -
		    roundf((orbitPlots[i].peri*orbitScale) * cos(orbitPlots[i].omegaRad)) - 23;
		  py2 = SOLAR_SYSTEM_CENTER_Y +
		    roundf((orbitPlots[i].peri*orbitScale) * sin(orbitPlots[i].omegaRad)) - 23;
		  eR = sqrt(orbitPlots[i].x[eD]*orbitPlots[i].x[eD] +
			    orbitPlots[i].y[eD]*orbitPlots[i].y[eD]);
		  gdk_draw_line(pixmap, darkGreyGC, px1, py1, px2, py2);
		  orbit[0].x = px2-1; orbit[0].y = py2-1;
		  orbit[1].x = px2+1; orbit[1].y = py2-1;
		  orbit[2].x = px2+1; orbit[2].y = py2+1;
		  orbit[3].x = px2-1; orbit[3].y = py2+1;
		  gdk_draw_polygon(pixmap, whiteGC, TRUE, orbit, 4);
		  eR = sqrt(orbitPlots[i].x[eD]*orbitPlots[i].x[eD] +
			    orbitPlots[i].y[eD]*orbitPlots[i].y[eD]);
		  x = SOLAR_SYSTEM_CENTER_X + eR*eCos*orbitScale;
		  y = SOLAR_SYSTEM_CENTER_Y + eR*eSin*orbitScale;
		} else
		  plotPlanet = FALSE;
	      }
	      if (plotPlanet) {
		gdk_draw_arc(pixmap, planetGC, TRUE,
			     x-33, y-33, 20, 20, 0, FULL_CIRCLE);
		gdk_draw_drawable(pixmap, blueGC, smallPlanetImages[i], 0, 0,
				  x-w/2-23, y-h/2-23, 20, 20);
	      }
	      break;
	    case MOON:
	      if (schematic) {
		calculatePlanetPosition(EARTH, tJD, &eX, &eY, &eZ);
		sunELong = atan2(eY, eX);
		x -= 2*SCHEMATIC_ORBIT_INCREMENT;
		y -= 2*SCHEMATIC_ORBIT_INCREMENT;
		gdk_draw_arc(pixmap, blueGC, FALSE,
			     x, y, 2*SCHEMATIC_ORBIT_INCREMENT, 2*SCHEMATIC_ORBIT_INCREMENT, 0, FULL_CIRCLE);
		getMoonPosition(tJD, sunELong, &dDummy, &dDummy, &eLong, &dDummy, &fDummy);
		x += 2*SCHEMATIC_ORBIT_INCREMENT - roundf(cos(eLong)*SCHEMATIC_ORBIT_INCREMENT);
		y += 2*SCHEMATIC_ORBIT_INCREMENT + roundf(sin(eLong)*SCHEMATIC_ORBIT_INCREMENT);
		gdk_draw_arc(pixmap, planetGC, TRUE,
			     x-33, y-33, 20, 20, 0, FULL_CIRCLE);
		gdk_draw_drawable(pixmap, blueGC, smallPlanetImages[i], 0, 0,
				  x-w/2-23, y-h/2-23, 20, 20);
	      }
	      break;
	    }
	  }
	}
	if (j > 0) {
	  if (j < nSolarSystemDays -1)
	    gdk_draw_drawable(drawingArea->window,
			      drawingArea->style->fg_gc[GTK_WIDGET_STATE (drawingArea)],
			      pixmap, 0, BIG_FONT_OFFSET, 0, BIG_FONT_OFFSET,
			      displayWidth, displayHeight*6/7 - 14 - BIG_FONT_OFFSET);
	  else
	    gdk_draw_drawable(drawingArea->window,
			      drawingArea->style->fg_gc[GTK_WIDGET_STATE (drawingArea)],
			      pixmap, 0, 0, 0, 0,
			      displayWidth, displayHeight - 14);
	  gdk_flush();
	}
	firstCall = firstIteration = FALSE;
	j++;
	if (j < nSolarSystemDays)
	  tJD += dayIncSign;
      } while (j < nSolarSystemDays);
      addSensitiveArea(FALSE, SA_SOLAR_SYSTEM_BUTTON,
		       0, displayHeight*3/4, displayWidth,
		       displayHeight, 0.0);
      if (!schematic)
	havePlottedOrbits = TRUE;
    }
    break;
  case METEOR_SHOWERS_SCREEN:
    {
      int yearNow, monthNow, dayNow, occuringNow, circumpolar, darkStartHH, darkEndHH;
      int darkStartMM, darkEndMM;
      float illum, illum2;
      double dayOffset, showerTJD, dummy, startDayNumber, endDayNumber, todayDayNumber;
      double jan0TJD, darkHours, midnight, darkHoursTonight, rA, dec, az, zA, h0, startNight;
      double darkStart, darkEnd;
      char rateString[4], phaseGrad[2];
      meteorShower *shower;
      GdkGC *lineGC;

      if (!haveReadMeteorShowers) {
	readMeteorShowers();
	haveReadMeteorShowers = TRUE;
      }
      needNewTime = TRUE;
      lSTNow = lST();
      planetInfo(dataDir, EARTH, tJD, &rA, &dec, &illum, &illum);
      azZA(rA, sin(dec), cos(dec), &az, &zA, FALSE);
      h0 = -(50.0/60.0)*DEGREES_TO_RADIANS;
      if (zA < M_PI*0.5) {
	/* The sun is up, calculate dark time from next sunset to following sunset */
	startNight = calcRiseOrSetTime(FALSE, EARTH, tJD, lSTAtTJD(tJD),
				       h0, &circumpolar, &dummy, &dummy, &dummy, &illum);
      } else {
	/* The sun is down, calculate dark time from last sunset to next sunset */
	startNight = calcRiseOrSetTime(FALSE, EARTH, tJD-1.0, lSTAtTJD(tJD-1.0),
				       h0, &circumpolar, &dummy, &dummy, &dummy, &illum);
	if ((circumpolar == 0) && (tJD - startNight > 1.0))
	  startNight = calcRiseOrSetTime(FALSE, EARTH, tJD, lSTAtTJD(tJD),
					 h0, &circumpolar, &dummy, &dummy, &dummy, &illum);
      }
      tJDToDate(startNight, &year, &month, &dayNum);
      if (circumpolar == 0)
	darkHoursTonight = darkTime(startNight, startNight + 1.0, &darkStart, &darkEnd);
      else
	darkHoursTonight = darkTime(tJD-0.5, tJD+0.5, &darkStart, &darkEnd);
      dummy = 24.0*(darkStart - (double)((int)darkStart)) + 12.0;
      if (dummy >= 24.0)
	dummy -= 24.0;
      darkStartHH = (int)dummy;
      darkStartMM = (int)((dummy - (double)darkStartHH)*60.0);
      dummy = 24.0*(darkEnd - (double)((int)darkEnd)) + 12.0;
      if (dummy >= 24.0)
	dummy -= 24.0;
      darkEndHH = (int)dummy;
      darkEndMM = (int)((dummy - (double)darkEndHH)*60.0);
      tJDToDate(tJD, &yearNow, &monthNow, &dayNow);
      jan0TJD = buildTJD(yearNow-1900, 0, 0, 0, 0, 0);
      sprintf(scratchString, "Meteor Shower Information for %04d", yearNow);
      stringWidth = gdk_string_width(smallFont, scratchString);
      gdk_draw_string(pixmap, smallFont, whiteGC,
		      (displayWidth/2)-stringWidth/2,
		      (row++)*ABOUT_ROW_STEP + BIG_FONT_OFFSET, scratchString);
      row++;
      gdk_draw_line(pixmap, blueGC,
		    xBorder, row*ABOUT_ROW_STEP, displayWidth-xBorder, row*ABOUT_ROW_STEP);
      row++;
      sprintf(scratchString, "     Shower Name       Abr  Rate      Dates       Max    V Moon%% Dark Hours");
      stringWidth = gdk_string_width(smallFont, scratchString);
      gdk_draw_string(pixmap, smallFont, blueGC, 0,
		      (row++)*ABOUT_ROW_STEP + BIG_FONT_OFFSET, scratchString);
      shower = meteorShowers;
      while (shower != NULL) {
	startDayNumber= makeDayNumber(yearNow, shower->startMonth, shower->startDay);
	endDayNumber= makeDayNumber(yearNow, shower->endMonth, shower->endDay);
	todayDayNumber= makeDayNumber(yearNow, monthNow, dayNow);
	dayOffset = makeDayNumber(yearNow, shower->maxMonth, shower->maxDay);
	showerTJD = jan0TJD + dayOffset;
	if (startDayNumber < endDayNumber) {
	  if ((todayDayNumber >= startDayNumber) && (todayDayNumber <= endDayNumber))
	    occuringNow = TRUE;
	  else
	    occuringNow = FALSE;
	} else {
	  /* Handle case when shower starts in December and ends in January */
	  if ((todayDayNumber < startDayNumber) && (todayDayNumber > endDayNumber))
	    occuringNow = FALSE;
	  else
	    occuringNow = TRUE;
	}
	/*
	printf("%02d/%02d <= %02d/%02d <= %02d/%02d : %d\n",
	       shower->startMonth, shower->startDay, monthNow, dayNow,
	       shower->endMonth, shower->endDay, occuringNow);
	*/
	planetInfo(dataDir, EARTH, showerTJD, &dummy, &dummy, &illum2, &mag);
	planetInfo(dataDir, MOON, showerTJD+0.04, &dummy, &dummy, &illum2, &mag);
	planetInfo(dataDir, MOON, showerTJD, &dummy, &dummy, &illum, &mag);
	midnight = (double)((int)(showerTJD + 0.5)) + 0.5;
	darkHours = darkTime(midnight - 0.5, midnight + 0.5, &dummy, &dummy);
	if (illum2 > illum)
	  strcpy(phaseGrad, "+");
	else
	  strcpy(phaseGrad, "-");
	if (occuringNow && (darkHoursTonight > 2.0))
	  lineGC = greenGC;
	else if (occuringNow)
	  lineGC = darkGreenGC;
	else if (darkHours > 2.0)
	  lineGC = creamGC;
	else
	  lineGC = greyGC;
	if (shower->zHR > 0)
	  sprintf(rateString, "%3d", shower->zHR);
	else
	  strcpy(rateString, "Var");
	sprintf(scratchString, "%21s  %3s  %3s  %02d/%02d -> %02d/%02d  %02d/%02d  %d  %3.0f%s    %4.1f",
		shower->fullName, shower->threeLetterName,
		rateString, shower->startMonth, shower->startDay,
		shower->endMonth, shower->endDay, shower->maxMonth, shower->maxDay,
		shower->vInf, illum*100, phaseGrad, darkHours);
	gdk_draw_string(pixmap, smallFont, lineGC, 0,
			(row++)*(ABOUT_ROW_STEP+2) + BIG_FONT_OFFSET, scratchString);
	shower = shower->next;
      }
      if (darkHoursTonight == 0.0) {
	sprintf(scratchString, "At no time tonight will the sky be very dark.");
	lineGC = redGC;
      } else if (darkHoursTonight > 2.0) {
	sprintf(scratchString,
		"The sky will be very dark for %3.1f hours (%02d:%02d -> %02d:%02d UT) tonight.",
		darkHoursTonight, darkStartHH, darkStartMM, darkEndHH, darkEndMM);
	lineGC = whiteGC;
      } else {
	sprintf(scratchString,
		"The sky will be very dark for only %3.1f hours (%02d:%02d -> %02d:%02d UT) tonight.",
		darkHoursTonight, darkStartHH, darkStartMM, darkEndHH, darkEndMM);
	lineGC = redGC;
      }
      stringWidth = gdk_string_width(smallFont, scratchString);
      gdk_draw_string(pixmap, smallFont, lineGC, (displayWidth - stringWidth)/2,
		      (row++)*(ABOUT_ROW_STEP+2) + BIG_FONT_OFFSET + 5, scratchString);
      
    }
    break;
  case PLANETCOMPASS_SCREEN:
    {
      int circleRadius, circleX, circleY, i,
	x[N_SOLAR_SYSTEM_OBJECTS], y[N_SOLAR_SYSTEM_OBJECTS],
	xx[N_SOLAR_SYSTEM_OBJECTS], yy[N_SOLAR_SYSTEM_OBJECTS];
      float dummy;
      double rA, dec, az, zA;
      GdkPoint triangle[3];
      GdkGC *riseGC, *transitGC, *setGC, *hAGC, *nameGC;

      needNewTime = TRUE;
      lSTNow = lST();
      sprintf(scratchString, "Planet positions at ");
      makeTimeString(&scratchString[strlen(scratchString)]);
      stringWidth = gdk_string_width(smallFont, scratchString);
      gdk_draw_string(pixmap, smallFont, whiteGC,
		      (displayWidth/2)-stringWidth/2,
		      (row++)*ABOUT_ROW_STEP + BIG_FONT_OFFSET, scratchString);
      row += 3;
      circleRadius = 20 + displayWidth/3;
      circleX = circleRadius-61;
      circleY = row*ABOUT_ROW_STEP;
      gdk_draw_arc(pixmap, creamGC, FALSE,
		   circleX, circleY, circleRadius*2,
		   circleRadius*2, 0, FULL_CIRCLE);
      /* Draw cardinal points */
      triangle[0].x = circleX+circleRadius - 4; triangle[0].y = circleY;
      triangle[1].x = triangle[0].x        + 8; triangle[1].y = triangle[0].y;
      triangle[2].x = triangle[0].x        + 4; triangle[2].y = triangle[0].y + 15;
      gdk_draw_polygon(pixmap, blueGC, TRUE, triangle, 3);
      gdk_draw_string(pixmap, smallFont, greenGC, 1+circleX+circleRadius-gdk_string_width(smallFont, "N")/2,
		      circleY+29, "N");
      triangle[0].x = circleX+circleRadius - 4; triangle[0].y = circleY + 2*circleRadius + 1;
      triangle[1].x = triangle[0].x        + 8; triangle[1].y = triangle[0].y;
      triangle[2].x = triangle[0].x        + 4; triangle[2].y = triangle[0].y - 15;
      gdk_draw_polygon(pixmap, blueGC, TRUE, triangle, 3);
      gdk_draw_string(pixmap, smallFont, greenGC, circleX+circleRadius-gdk_string_width(smallFont, "S")/2,
		      circleY+2*circleRadius-21, "S");
      triangle[0].x = circleX;                  triangle[0].y = circleY + circleRadius - 4;
      triangle[1].x = triangle[0].x;            triangle[1].y = triangle[0].y          + 8;
      triangle[2].x = triangle[0].x + 15;       triangle[2].y = triangle[0].y          + 4;
      gdk_draw_polygon(pixmap, blueGC, TRUE, triangle, 3);
      gdk_draw_string(pixmap, smallFont, greenGC, circleX+19, circleY+circleRadius+4, "W");
      triangle[0].x = circleX + 2*circleRadius; triangle[0].y = circleY + circleRadius - 4;
      triangle[1].x = triangle[0].x;            triangle[1].y = triangle[0].y          + 8;
      triangle[2].x = triangle[0].x - 15;       triangle[2].y = triangle[0].y          + 4;
      gdk_draw_polygon(pixmap, blueGC, TRUE, triangle, 3);
      gdk_draw_string(pixmap, smallFont, greenGC, circleX+2*circleRadius-19-gdk_string_width(smallFont, "E")/2,
		      circleY+circleRadius+4, "E");
      drawCompassLine(blueGC, M_PI*0.25, 8, circleX+circleRadius, circleY+circleRadius, circleRadius);
      drawCompassLine(blueGC, M_PI*0.75, 8, circleX+circleRadius, circleY+circleRadius, circleRadius);
      drawCompassLine(blueGC, M_PI*1.25, 8, circleX+circleRadius, circleY+circleRadius, circleRadius);
      drawCompassLine(blueGC, M_PI*1.75, 8, circleX+circleRadius, circleY+circleRadius, circleRadius);
      for (i = 10; i < 360; i += 10) {
	int x, y;
	float angle;

	if (i % 9 != 0) {
	  angle = (float)i * M_2PI/360.0;
	  x = (int)((float)(circleX+circleRadius) + (float)(circleRadius-5) * sin(angle) + 0.5);
	  y = (int)((float)(circleY+circleRadius) + (float)(circleRadius-5) * cos(angle) + 0.5);
	  gdk_draw_arc(pixmap, blueGC, TRUE, x-1, y-1, 3, 3, 0, FULL_CIRCLE);
	}
      }
      gdk_draw_arc(pixmap, creamGC, FALSE,
		   0, circleY, circleRadius*2,
		   circleRadius*2, FULL_CIRCLE/4, FULL_CIRCLE/2);
      triangle[0].x = 0;                        triangle[0].y = circleY + circleRadius - 4;
      triangle[1].x = triangle[0].x;            triangle[1].y = triangle[0].y          + 8;
      triangle[2].x = triangle[0].x + 15;       triangle[2].y = triangle[0].y          + 4;
      gdk_draw_polygon(pixmap, blueGC, TRUE, triangle, 3);
      triangle[0].x = circleRadius - 4;         triangle[0].y = circleY;
      triangle[1].x = triangle[0].x        + 8; triangle[1].y = triangle[0].y;
      triangle[2].x = triangle[0].x        + 4; triangle[2].y = triangle[0].y + 15;
      gdk_draw_polygon(pixmap, blueGC, TRUE, triangle, 3);
      triangle[0].x = circleRadius - 4;         triangle[0].y = circleY + 2*circleRadius + 1;
      triangle[1].x = triangle[0].x        + 8; triangle[1].y = triangle[0].y;
      triangle[2].x = triangle[0].x        + 4; triangle[2].y = triangle[0].y - 15;
      gdk_draw_polygon(pixmap, blueGC, TRUE, triangle, 3);
      drawCompassLine(blueGC, M_PI*1.25, 8, circleRadius, circleY+circleRadius, circleRadius);
      drawCompassLine(blueGC, M_PI*1.75, 8, circleRadius, circleY+circleRadius, circleRadius);
      gdk_draw_string(pixmap, smallFont, greenGC, circleRadius-gdk_string_width(smallFont, "Zenith")/2,
		      circleY-8, "Zenith");
      gdk_draw_string(pixmap, smallFont, greenGC, 20, circleY+circleRadius+4, "Horizon");
      gdk_draw_string(pixmap, smallFont, greenGC, circleRadius-gdk_string_width(smallFont, "Nadir")/2,
		      circleY+2*circleRadius+15, "Nadir");
      for (i = 190; i < 360; i += 10) {
	int x, y;
	float angle;

	if (i % 9 != 0) {
	  angle = (float)i * M_2PI/360.0;
	  x = (int)((float)(circleRadius) + (float)(circleRadius-5) * sin(angle) + 0.5);
	  y = (int)((float)(circleY+circleRadius) + (float)(circleRadius-5) * cos(angle) + 0.5);
	  gdk_draw_arc(pixmap, blueGC, TRUE, x-1, y-1, 3, 3, 0, FULL_CIRCLE);
	}
      }
      /* Loop down to 0, so that the Sun doesn't get covered if objects are close to it in sky */
      planetInfo(dataDir, EARTH, tJD, &rA, &dec, &dummy, &dummy);
      for (i = N_SOLAR_SYSTEM_OBJECTS-2; i >=  1; i--) {
	int w, h, boost, clear, j, dSq;

	planetInfo(dataDir, i, tJD, &rA, &dec, &dummy, &dummy);
	azZA(rA, sin(dec), cos(dec), &az, &zA, FALSE);
	boost = 0;
	do {
	  x[i] = (int)((float)(circleX+circleRadius) + (float)(circleRadius-33-boost*20) * sin(az) + 0.5);
	  y[i] = (int)((float)(circleY+circleRadius) - (float)(circleRadius-33-boost*20) * cos(az) + 0.5);
	  clear = TRUE;
	  if (i < N_SOLAR_SYSTEM_OBJECTS-2) {
	    for (j = N_SOLAR_SYSTEM_OBJECTS-2; j > i; j--) {
	      dSq = ((x[i]-x[j])*(x[i]-x[j])) + ((y[i]-y[j])*(y[i]-y[j]));
	      if (dSq < 200) {
		boost++;
		clear = FALSE;
		break;
	      }
	    }
	  }
	} while (!clear);
	if (zA <= M_PI*0.5)
	  drawCompassLine(whiteGC, az, 20+boost*20, circleX+circleRadius, circleY+circleRadius, circleRadius);
	else
	  drawCompassLine(redGC, az, 20+boost*20, circleX+circleRadius, circleY+circleRadius, circleRadius);
	gdk_drawable_get_size(planetImages[i], &w, &h);
	if (i != EARTH)
	  gdk_draw_drawable(pixmap, blueGC, planetImages[i], 0, 0,
			    x[i]-w/2, y[i]-h/2, 20, 20);
	else
	  gdk_draw_drawable(pixmap, blueGC, planetImages[SUN], 0, 0,
			    x[i]-w/2, y[i]-h/2, 20, 20);
	boost = 0;
	do {
	  xx[i] = (int)((float)(circleRadius) + (float)(circleRadius-30-boost*20) * sin(M_PI+zA) + 0.5);
	  yy[i] = (int)((float)(circleY+circleRadius) + (float)(circleRadius-30-boost*20) * cos(M_PI+zA) + 0.5);
	  clear = TRUE;
	  if (i < N_SOLAR_SYSTEM_OBJECTS-2) {
	    for (j = N_SOLAR_SYSTEM_OBJECTS-2; j > i; j--) {
	      dSq = ((xx[i]-xx[j])*(xx[i]-xx[j])) + ((yy[i]-yy[j])*(yy[i]-yy[j]));
	      if (dSq < 200) {
		boost++;
		clear = FALSE;
		break;
	      }
	    }
	  }
	} while (!clear);
	if (zA <= M_PI*0.5)
	  drawCompassLine(whiteGC, M_2PI-zA, 20+boost*20, circleRadius, circleY+circleRadius, circleRadius);
	else
	  drawCompassLine(redGC, M_2PI-zA, 20+boost*20, circleRadius, circleY+circleRadius, circleRadius);
	if (i != EARTH)
	  gdk_draw_drawable(pixmap, blueGC, planetImages[i], 0, 0,
			    xx[i]-w/2, yy[i]-h/2, 20, 20);
	else
	  gdk_draw_drawable(pixmap, blueGC, planetImages[SUN], 0, 0,
			    xx[i]-w/2, yy[i]-h/2, 20, 20);
      }
      row += 33;
      sprintf(scratchString, "                         Rising             Transit            Setting     ");
      stringWidth = gdk_string_width(smallFont, scratchString);
      gdk_draw_string(pixmap, smallFont, blueGC,
		      (displayWidth/2)-stringWidth/2,
		      (row++)*ABOUT_ROW_STEP, scratchString);
      sprintf(scratchString, "Planet      HA        Time      Az        Time      El       Time      Az  ");
      stringWidth = gdk_string_width(smallFont, scratchString);
      gdk_draw_string(pixmap, smallFont, blueGC,
		      (displayWidth/2)-stringWidth/2,
		      (row++)*ABOUT_ROW_STEP, scratchString);
      h0 = -(34.0/60.0)*DEGREES_TO_RADIANS;
      fracTJD = tJD - (double)((int)tJD);
      if (fracTJD < 0.5)
	tJDS[0] = (double)((int)tJD) + 0.5 + longitude/M_2PI;
      else
	tJDS[0] = (double)((int)tJD) + 1.5 + longitude/M_2PI;
      /*
	Display rise, transit and set info for planets.   Don't do
	it for Sun, Earth or Moon.
      */
      for (i = 1; i < N_SOLAR_SYSTEM_OBJECTS-1; i++)
	if ((i != EARTH) && (i != MOON)) {
	  int hASign;
	  float fDummy;
	  double hA, dec, rA, transitEl, theAz, az, zA;

	  sprintf(scratchString, "%s",solarSystemNames[i]);
	  planetInfo(dataDir, i, tJD, &rA, &dec, &fDummy, &fDummy);
	  azZA(rA, sin(dec), cos(dec), &az, &zA, FALSE);
	  lSTNow = lST();
	  hA = lSTNow - rA;
	  if (hA < -M_PI)
	    hA += M_2PI;
	  else if (hA > M_PI)
	    hA -= M_2PI;
	  riseGC = transitGC = setGC = hAGC = greyGC;
	  nameGC = creamGC;
	  if (zA > M_PI*0.5) {
	    riseGC = whiteGC;
	    nameGC = darkCreamGC;
	  } else if (hA < 0.0)
	    transitGC = hAGC = whiteGC;
	  else
	    setGC = hAGC = whiteGC;
	  gdk_draw_string(pixmap, smallFont, nameGC,
			  (displayWidth/2)-stringWidth/2,
			  row*ABOUT_ROW_STEP, scratchString);
	  hA /= HOURS_TO_RADIANS;
	  if (hA < 0.0) {
	    hASign = -1;
	    hA *= -1.0;
	  } else
	    hASign = 1;
	  rAHH = (int)hA;
	  rAMM = (int)((hA - (double)rAHH)*60.0 + 0.5);
	  if (hASign < 0)
	    sprintf(scratchString, "-%02d:%02d", rAHH, rAMM);
	  else
	    sprintf(scratchString, " %02d:%02d", rAHH, rAMM);
	  gdk_draw_string(pixmap, smallFont, hAGC,
			  (displayWidth/2)-stringWidth/2+gdk_string_width(smallFont, "Planet    "),
			  row*ABOUT_ROW_STEP, scratchString);
	  lSTNow = lSTAtTJD(tJDS[0]);
	  theTime = calcRiseOrSetTime(TRUE, i, tJDS[0], lSTNow, h0, &circumpolar,
				      &hD, &cosH, &sunDec, &illum);
	  if (circumpolar != 0) {
	    gdk_draw_string(pixmap, smallFont, greyGC,
			    (displayWidth/2)-stringWidth/2+gdk_string_width(smallFont, "Planet    HA        "),
			    row*ABOUT_ROW_STEP, "Not today");
	  } else {
	    theTime -= 0.5;
	    theTime = theTime - (double)((int)theTime);
	    rAHours = theTime*24.0;
	    rAHH = (int)rAHours;
	    rAMM = (int)((rAHours - (float)rAHH)*60.0 + 0.5);
	    theAz = atan2(sin(hD), cosH*sinLatitude-tan(sunDec)*cosLatitude) + M_PI;
	    if (theAz < 0.0)
	      theAz += M_2PI;
	    if (theAz > M_2PI)
	      theAz -= M_2PI;
	    sprintf(scratchString, "%02d:%02d UT   %5.1f", rAHH, rAMM, theAz/DEGREES_TO_RADIANS);
	    gdk_draw_string(pixmap, smallFont, riseGC,
			    (displayWidth/2)-stringWidth/2+gdk_string_width(smallFont, "Planet    HA        "),
			    row*ABOUT_ROW_STEP, scratchString);
	  }
	  theTime = calcTransitTime(tJDS[0], i, &transitEl);
	  theTime -= 0.5;
	  theTime = theTime - (double)((int)theTime);
	  rAHours = theTime*24.0;
	  rAHH = (int)rAHours;
	  rAMM = (int)((rAHours - (float)rAHH)*60.0 + 0.5);
	  sprintf(scratchString, "%02d:%02d UT  %5.1f", rAHH, rAMM, transitEl/DEGREES_TO_RADIANS);
	  gdk_draw_string(pixmap, smallFont, transitGC,
			  (displayWidth/2)-stringWidth/2+gdk_string_width(smallFont, "Planet             Time    Az           "),
			  row*ABOUT_ROW_STEP, scratchString);
	  theTime = calcRiseOrSetTime(FALSE, i, tJDS[0], lSTNow, h0, &circumpolar,
				      &hD, &cosH, &sunDec, &illum);
	  if (circumpolar != 0) {
	    gdk_draw_string(pixmap, smallFont, greyGC,
			    (displayWidth/2)-stringWidth/2+gdk_string_width(smallFont, "Planet      Rise Time  Az        Transit Time   El         "),
			    row*ABOUT_ROW_STEP, "Not today");
	  } else {
	    theTime -= 0.5;
	    theTime = theTime - (double)((int)theTime);
	    rAHours = theTime*24.0;
	    rAHH = (int)rAHours;
	    rAMM = (int)((rAHours - (float)rAHH)*60.0 + 0.5);
	    theAz = atan2(sin(hD), cosH*sinLatitude-tan(sunDec)*cosLatitude) + M_PI;
	    if (theAz < 0.0)
	      theAz += M_2PI;
	    if (theAz > M_2PI)
	      theAz -= M_2PI;
	    sprintf(scratchString, "%02d:%02d UT   %5.1f", rAHH, rAMM, theAz/DEGREES_TO_RADIANS);
	    gdk_draw_string(pixmap, smallFont, setGC,
			    (displayWidth/2)-stringWidth/2+gdk_string_width(smallFont, "Planet       Rise Time  Az      Transit Time   El         S"),
			    row*ABOUT_ROW_STEP, scratchString);
	  }
	  row++;
	}
      sprintf(scratchString, "Tap the screen to exit this page");
      stringWidth = gdk_string_width(smallFont, scratchString);
      gdk_draw_string(pixmap, smallFont, greenGC,
		      (displayWidth/2)-stringWidth/2,
		      displayHeight-ABOUT_ROW_STEP, scratchString);
    }
    break;
  }
  if ((aboutScreen != BIG_MOONCAL_SCREEN) && (aboutScreen != PLANETCOMPASS_SCREEN) &&
      (aboutScreen != SMALL_MOONCAL_SCREEN) && (aboutScreen != SOLAR_SYSTEM_SCHEMATIC_SCREEN)
      && (aboutScreen != SOLAR_SYSTEM_SCALE_SCREEN)) {
    sprintf(scratchString, "Tap the screen to exit this page");
    stringWidth = gdk_string_width(smallFont, scratchString);
    gdk_draw_string(pixmap, smallFont, greenGC,
		    (displayWidth/2)-stringWidth/2,
		    displayHeight-3*ABOUT_ROW_STEP, scratchString);
    
    sprintf(scratchString, "Please send comments, suggestions or questions to orrery.moko@gmail.com");
    stringWidth = gdk_string_width(smallFont, scratchString);
    gdk_draw_string(pixmap, smallFont, blueGC,
		    (displayWidth/2)-stringWidth/2,
		    displayHeight-ABOUT_ROW_STEP, scratchString);
  }
}

/*
  This function is the top functions for redrawing the plot in
  the drawing area.   It calls other functions to draw the
  different plot styles.
*/
static void redrawScreen(void)
{
  if (inFlashlightMode) {
    gdk_draw_rectangle(pixmap, flashlightGC,
		       TRUE, 0, 0,
		       drawingArea->allocation.width,
		       drawingArea->allocation.height);
  } else {
    gdk_draw_rectangle(pixmap, drawingArea->style->black_gc,
		       TRUE, 0, 0,
		       drawingArea->allocation.width,
		       drawingArea->allocation.height);
    if (displayingAnOptsPage)
      drawOptsScreens();
    else  {
      xBorder = 1;
      yBorder = 5;
      xLeftLabelSkip = 0;
      redrawScreenTransverseMercator();
    }
  }
}

/*
  This function is called when a portion of the drawing area
  is exposed.
*/
static gboolean exposeEvent(GtkWidget *widget, GdkEventExpose *event)
{
  gdk_draw_drawable (widget->window,
                     widget->style->fg_gc[GTK_WIDGET_STATE (widget)],
                     pixmap,
                     event->area.x, event->area.y,
                     event->area.x, event->area.y,
                     event->area.width, event->area.height);
  return(FALSE);
}

/*
  This function creates and initializes a backing pixmap
  of the appropriate size for the drawing area.
*/
static int configureEvent(GtkWidget *widget, GdkEventConfigure *event)
{
  if (pixmap)
    g_object_unref(pixmap);
  displayHeight = widget->allocation.height;
  displayWidth  = widget->allocation.width;
  pixmap = gdk_pixmap_new(widget->window,
			  widget->allocation.width,
			  widget->allocation.height, -1);
  dprintf("In configureEvent, w: %d, h: %d, f = %d\n",
	 displayWidth, displayHeight, fullscreenStateChanging);
  if (!fullscreenStateChanging || !inFullscreenMode || !inFlashlightMode ||
      (displayHeight == VGA_HEIGHT))
    redrawScreen();
  if (displayHeight == VGA_HEIGHT)
    fullscreenStateChanging = FALSE;
  return TRUE;
}

/*
  Make all the Graphic Contexts the program will need.
*/
static void makeGraphicContexts(GtkWidget *widget)
{
  int stat, i;
  short starRed[5] = {55000, 65535, 65535, 65535, 65535};
  short starBlue[5] = {65535, 65535, 40000, 40000, 45000};
  short starGreen[5] = {55000, 65535, 65535, 50000, 45000};
  GdkGCValues blueGCValues, whiteGCValues, greyGCValues, greenGCValues;
  GdkGCValues redGCValues, darkGreyGCValues, blackGCValues, equatorGCValues;
  GdkGCValues yellowGCValues, creamGCValues, blueGreenGCValues, starGCValues[5],
    goldGCValues, faintGoldGCValues, pinkGCValues, faintPinkGCValues, faintBlueGCValues,
    lightYellowGCValues, darkCreamGCValues, darkGreenGCValues;
  GdkColor blue, black, white, grey, darkGrey, red, green, yellow, cream, blueGreen, star[5],
    equator, gold, faintGold, pink, faintPink, faintBlue, lightYellow, darkCream, darkGreen;
  GdkGCValuesMask gCValuesMask;

  gCValuesMask = GDK_GC_FOREGROUND;

  blue.red   = 30000;
  blue.green = 30000;
  blue.blue  = 65535;
  stat = gdk_colormap_alloc_color(widget->style->colormap,
				  &blue,
				  FALSE,
				  TRUE);
  if (stat != TRUE) {
    fprintf(stderr, "Error allocating light blue color\n");
    exit(ERROR_EXIT);
  }
  blueGCValues.foreground = blue;
  blueGC = gtk_gc_get(widget->style->depth,
		      widget->style->colormap,
		      &blueGCValues, gCValuesMask
		      );
  if (blueGC == NULL) {
    fprintf(stderr, "gtk_gc_get failed for blueGC\n");
    exit(ERROR_EXIT);
  }

  white.red   = 65535;
  white.green = 65535;
  white.blue  = 65535;
  stat = gdk_colormap_alloc_color(widget->style->colormap,
				  &white,
				  FALSE,
				  TRUE);
  if (stat != TRUE) {
    fprintf(stderr, "Error allocating white color\n");
    exit(ERROR_EXIT);
  }
  whiteGCValues.foreground = white;
  whiteGC = gtk_gc_get(widget->style->depth,
		       widget->style->colormap,
		       &whiteGCValues, gCValuesMask
		       );
  if (whiteGC == NULL) {
    fprintf(stderr, "gtk_gc_get failed for whiteGC\n");
    exit(ERROR_EXIT);
  }
  grey.red   = 32000;
  grey.green = 32000;
  grey.blue  = 32000;
  stat = gdk_colormap_alloc_color(widget->style->colormap,
				  &grey,
				  FALSE,
				  TRUE);
  if (stat != TRUE) {
    fprintf(stderr, "Error allocating grey color\n");
    exit(ERROR_EXIT);
  }
  greyGCValues.foreground = grey;
  greyGC = gtk_gc_get(widget->style->depth,
		       widget->style->colormap,
		       &greyGCValues, gCValuesMask
		       );
  if (greyGC == NULL) {
    fprintf(stderr, "gtk_gc_get failed for greyGC\n");
    exit(ERROR_EXIT);
  }
  darkGrey.red   = 20000;
  darkGrey.green = 20000;
  darkGrey.blue  = 20000;
  stat = gdk_colormap_alloc_color(widget->style->colormap,
				  &darkGrey,
				  FALSE,
				  TRUE);
  if (stat != TRUE) {
    fprintf(stderr, "Error allocating darkGrey color\n");
    exit(ERROR_EXIT);
  }
  darkGreyGCValues.foreground = darkGrey;
  darkGreyGC = gtk_gc_get(widget->style->depth,
		       widget->style->colormap,
		       &darkGreyGCValues, gCValuesMask
		       );
  if (darkGreyGC == NULL) {
    fprintf(stderr, "gtk_gc_get failed for darkGreyGC\n");
    exit(ERROR_EXIT);
  }
  red.red   = 65535;
  red.green = 0;
  red.blue  = 0;
  stat = gdk_colormap_alloc_color(widget->style->colormap,
				  &red,
				  FALSE,
				  TRUE);
  if (stat != TRUE) {
    fprintf(stderr, "Error allocating red color\n");
    exit(ERROR_EXIT);
  }
  redGCValues.foreground = red;
  redGC = gtk_gc_get(widget->style->depth,
		       widget->style->colormap,
		       &redGCValues, gCValuesMask
		       );
  if (redGC == NULL) {
    fprintf(stderr, "gtk_gc_get failed for redGC\n");
    exit(ERROR_EXIT);
  }
  green.red   = 0;
  green.green = 65535;
  green.blue  = 0;
  stat = gdk_colormap_alloc_color(widget->style->colormap,
				  &green,
				  FALSE,
				  TRUE);
  if (stat != TRUE) {
    fprintf(stderr, "Error allocating green color\n");
    exit(ERROR_EXIT);
  }
  greenGCValues.foreground = green;
  greenGC = gtk_gc_get(widget->style->depth,
		       widget->style->colormap,
		       &greenGCValues, gCValuesMask
		       );
  if (greenGC == NULL) {
    fprintf(stderr, "gtk_gc_get failed for greenGC\n");
    exit(ERROR_EXIT);
  }
  black.red   = 0;
  black.green = 0;
  black.blue  = 0;
  stat = gdk_colormap_alloc_color(widget->style->colormap,
				  &black,
				  FALSE,
				  TRUE);
  if (stat != TRUE) {
    fprintf(stderr, "Error allocating black color\n");
    exit(ERROR_EXIT);
  }
  blackGCValues.foreground = black;
  blackGC = gtk_gc_get(widget->style->depth,
		       widget->style->colormap,
		       &blackGCValues, gCValuesMask
		       );
  if (blackGC == NULL) {
    fprintf(stderr, "gtk_gc_get failed for blackGC\n");
    exit(ERROR_EXIT);
  }
  lightYellow.red   = 65535;
  lightYellow.green = 65535;
  lightYellow.blue  = 32000;
  stat = gdk_colormap_alloc_color(widget->style->colormap,
				  &lightYellow,
				  FALSE,
				  TRUE);
  if (stat != TRUE) {
    fprintf(stderr, "Error allocating lightYellow color\n");
    exit(ERROR_EXIT);
  }
  lightYellowGCValues.foreground = lightYellow;
  lightYellowGC = gtk_gc_get(widget->style->depth,
		       widget->style->colormap,
		       &lightYellowGCValues, gCValuesMask
		       );
  if (lightYellowGC == NULL) {
    fprintf(stderr, "gtk_gc_get failed for lightYellowGC\n");
    exit(ERROR_EXIT);
  }
  yellow.red   = 65535;
  yellow.green = 65535;
  yellow.blue  = 0;
  stat = gdk_colormap_alloc_color(widget->style->colormap,
				  &yellow,
				  FALSE,
				  TRUE);
  if (stat != TRUE) {
    fprintf(stderr, "Error allocating yellow color\n");
    exit(ERROR_EXIT);
  }
  yellowGCValues.foreground = yellow;
  yellowGC = gtk_gc_get(widget->style->depth,
		       widget->style->colormap,
		       &yellowGCValues, gCValuesMask
		       );
  if (yellowGC == NULL) {
    fprintf(stderr, "gtk_gc_get failed for yellowGC\n");
    exit(ERROR_EXIT);
  }
  faintBlue.red   = (int)(30000.0 * 0.7);
  faintBlue.green = (int)(30000.0 * 0.7);
  faintBlue.blue  = (int)(65535.0 * 0.7);
  stat = gdk_colormap_alloc_color(widget->style->colormap,
				  &faintBlue,
				  FALSE,
				  TRUE);
  if (stat != TRUE) {
    fprintf(stderr, "Error allocating faintBlue color\n");
    exit(ERROR_EXIT);
  }
  faintBlueGCValues.foreground = faintBlue;
  faintBlueGC = gtk_gc_get(widget->style->depth,
		       widget->style->colormap,
		       &faintBlueGCValues, gCValuesMask
		       );
  if (faintBlueGC == NULL) {
    fprintf(stderr, "gtk_gc_get failed for faintBlueGC\n");
    exit(ERROR_EXIT);
  }
  pink.red   = 65535;
  pink.green = 14906;
  pink.blue  = 25186;
  pink.red   = 65535;
  pink.green = 25000;
  pink.blue  = 35000;
  stat = gdk_colormap_alloc_color(widget->style->colormap,
				  &pink,
				  FALSE,
				  TRUE);
  if (stat != TRUE) {
    fprintf(stderr, "Error allocating pink color\n");
    exit(ERROR_EXIT);
  }
  pinkGCValues.foreground = pink;
  pinkGC = gtk_gc_get(widget->style->depth,
		       widget->style->colormap,
		       &pinkGCValues, gCValuesMask
		       );
  if (pinkGC == NULL) {
    fprintf(stderr, "gtk_gc_get failed for pinkGC\n");
    exit(ERROR_EXIT);
  }
  faintPink.red   = (int)(65535.0 * 0.9);
  faintPink.green = (int)(14906.0 * 0.9);
  faintPink.blue  = (int)(25186.0 * 0.9);
  stat = gdk_colormap_alloc_color(widget->style->colormap,
				  &faintPink,
				  FALSE,
				  TRUE);
  if (stat != TRUE) {
    fprintf(stderr, "Error allocating faintPink color\n");
    exit(ERROR_EXIT);
  }
  faintPinkGCValues.foreground = faintPink;
  faintPinkGC = gtk_gc_get(widget->style->depth,
		       widget->style->colormap,
		       &faintPinkGCValues, gCValuesMask
		       );
  if (faintPinkGC == NULL) {
    fprintf(stderr, "gtk_gc_get failed for faintPinkGC\n");
    exit(ERROR_EXIT);
  }
  gold.red   = 65535;
  gold.green = 55255;
  gold.blue  = 00000;
  stat = gdk_colormap_alloc_color(widget->style->colormap,
				  &gold,
				  FALSE,
				  TRUE);
  if (stat != TRUE) {
    fprintf(stderr, "Error allocating gold color\n");
    exit(ERROR_EXIT);
  }
  goldGCValues.foreground = gold;
  goldGC = gtk_gc_get(widget->style->depth,
		       widget->style->colormap,
		       &goldGCValues, gCValuesMask
		       );
  if (goldGC == NULL) {
    fprintf(stderr, "gtk_gc_get failed for goldGC\n");
    exit(ERROR_EXIT);
  }
  faintGold.red   = (int)(65535.0 * 0.6);
  faintGold.green = (int)(55255.0 * 0.6);
  faintGold.blue  = 00000;
  stat = gdk_colormap_alloc_color(widget->style->colormap,
				  &faintGold,
				  FALSE,
				  TRUE);
  if (stat != TRUE) {
    fprintf(stderr, "Error allocating faintGold color\n");
    exit(ERROR_EXIT);
  }
  faintGoldGCValues.foreground = faintGold;
  faintGoldGC = gtk_gc_get(widget->style->depth,
		       widget->style->colormap,
		       &faintGoldGCValues, gCValuesMask
		       );
  if (faintGoldGC == NULL) {
    fprintf(stderr, "gtk_gc_get failed for faintGoldGC\n");
    exit(ERROR_EXIT);
  }
  darkGreen.red   = 0;
  darkGreen.green = 40000;
  darkGreen.blue  = 0;
  stat = gdk_colormap_alloc_color(widget->style->colormap,
				  &darkGreen,
				  FALSE,
				  TRUE);
  if (stat != TRUE) {
    fprintf(stderr, "Error allocating darkGreen color\n");
    exit(ERROR_EXIT);
  }
  darkGreenGCValues.foreground = darkGreen;
  darkGreenGC = gtk_gc_get(widget->style->depth,
		       widget->style->colormap,
		       &darkGreenGCValues, gCValuesMask
		       );
  if (darkGreenGC == NULL) {
    fprintf(stderr, "gtk_gc_get failed for darkGreenGC\n");
    exit(ERROR_EXIT);
  }
  cream.red   = 65535;
  cream.green = 65535;
  cream.blue  = 40000;
  stat = gdk_colormap_alloc_color(widget->style->colormap,
				  &cream,
				  FALSE,
				  TRUE);
  if (stat != TRUE) {
    fprintf(stderr, "Error allocating cream color\n");
    exit(ERROR_EXIT);
  }
  creamGCValues.foreground = cream;
  creamGC = gtk_gc_get(widget->style->depth,
		       widget->style->colormap,
		       &creamGCValues, gCValuesMask
		       );
  if (creamGC == NULL) {
    fprintf(stderr, "gtk_gc_get failed for creamGC\n");
    exit(ERROR_EXIT);
  }
  darkCream.red   = (int)(65535.0*0.6);
  darkCream.green = (int)(65535.0*0.6);
  darkCream.blue  = (int)(40000.0*0.6);
  stat = gdk_colormap_alloc_color(widget->style->colormap,
				  &darkCream,
				  FALSE,
				  TRUE);
  if (stat != TRUE) {
    fprintf(stderr, "Error allocating darkCream color\n");
    exit(ERROR_EXIT);
  }
  darkCreamGCValues.foreground = darkCream;
  darkCreamGC = gtk_gc_get(widget->style->depth,
		       widget->style->colormap,
		       &darkCreamGCValues, gCValuesMask
		       );
  if (darkCreamGC == NULL) {
    fprintf(stderr, "gtk_gc_get failed for darkCreamGC\n");
    exit(ERROR_EXIT);
  }
  blueGreen.red   = 00000;
  blueGreen.green = 65535;
  blueGreen.blue  = 65535;
  stat = gdk_colormap_alloc_color(widget->style->colormap,
				  &blueGreen,
				  FALSE,
				  TRUE);
  if (stat != TRUE) {
    fprintf(stderr, "Error allocating blueGreen color\n");
    exit(ERROR_EXIT);
  }
  blueGreenGCValues.foreground = blueGreen;
  blueGreenGC = gtk_gc_get(widget->style->depth,
		       widget->style->colormap,
		       &blueGreenGCValues, gCValuesMask
		       );
  if (blueGreenGC == NULL) {
    fprintf(stderr, "gtk_gc_get failed for blueGreenGC\n");
    exit(ERROR_EXIT);
  }
  equator.red   = 65535;
  equator.green = 65535;
  equator.blue  = 32767;
  stat = gdk_colormap_alloc_color(widget->style->colormap,
				  &equator,
				  FALSE,
				  TRUE);
  if (stat != TRUE) {
    fprintf(stderr, "Error allocating equator color\n");
    exit(ERROR_EXIT);
  }
  equatorGCValues.foreground = equator;
  equatorGC = gtk_gc_get(widget->style->depth,
			widget->style->colormap,
			&equatorGCValues, gCValuesMask
		       );
  if (equatorGC == NULL) {
    fprintf(stderr, "gtk_gc_get failed for equatorGC\n");
    exit(ERROR_EXIT);
  }
  for (i = 0; i < 5; i++) {
    star[i].red   = starRed[i];
    star[i].green = starGreen[i];
    star[i].blue  = starBlue[i];
    stat = gdk_colormap_alloc_color(widget->style->colormap,
				    &star[i],
				    FALSE,
				    TRUE);
    if (stat != TRUE) {
      fprintf(stderr, "Error allocating star[%d] color\n", i);
      exit(ERROR_EXIT);
    }
    starGCValues[i].foreground = star[i];
    starGC[i] = gtk_gc_get(widget->style->depth,
			   widget->style->colormap,
			   &starGCValues[i], gCValuesMask
			     );
    if (starGC[i] == NULL) {
      fprintf(stderr, "gtk_gc_get failed for starGC[%d]\n", i);
      exit(ERROR_EXIT);
    }
  }
}

/*
  Set up the fonts I need.
*/
static void makeFonts(GtkWidget *widget)
{
  smallFont =
    gdk_font_load("-misc-fixed-medium-r-semicondensed--13-100-100-100-c-60-iso8859-1");
  if (smallFont == NULL) {
    fprintf(stderr, "Error making smallFont\n");
    exit(ERROR_EXIT);
  }
}

/*
  This function reads in the file containing star names.   The file is organized
  as follows:
  First two bytes:                        The number of stars in this file (nStarNames)
  Next nStarNames*sizeof(starName) bytes: nStars starName structure elements
  Next four bytes:                        Size of the name string, in bytes (bignameStringlength)
  Last bigNameStringLength bytes:         A large caracter string holding all star names
*/
static void readStarNamesFile(void)
{
  int i, fD;
  char fileName[MAX_FILE_NAME_SIZE];

  sprintf(fileName, "%s/starNames.dat", dataDir);
  fD = open(fileName, O_RDONLY);
  if (fD < 0) {
    perror(fileName);
    return;
  }
  /* Read in the number of named stars */
  read(fD, &nStarNames, 2);

  /* Get storage for string offsets and coordinates */
  starNames = malloc(nStarNames*sizeof(starName));
  if (starNames == NULL) {
    perror("malloc of starNames");
    return;
  }

  /* Read in array of string offsets and coordinates */
  for (i = 0; i < nStarNames; i++)
    read(fD, &starNames[i], sizeof(starNames[i]));

  /* Read in size of the name string */
  read(fD, &bigNameStringLength, 4);

  /* Allocate storage for big name string */
  bigNameString = malloc(bigNameStringLength);
  if (bigNameString == NULL) {
    perror("malloc of bigNameString");
    return;
  }

  /* Read in the big name string */
  read(fD, bigNameString, bigNameStringLength);

  close(fD);
  haveStarNames = TRUE;
}

/*
  This function reads in the binary Hipparcos star catalog file, and
  allocates a linked list to store the data.
*/
static void readHipparcosCatalog(void)
{
  int fD;
  char fileName[MAX_FILE_NAME_SIZE];

  /*
    Check if the faint star catalog exists - if so, I'll make a button
    to select it later.
  */
  sprintf(fileName, "%s/faintStars", dataDir);
  fD = open(fileName, O_RDONLY);
  if (fD >= 0) {
    haveFaintStars = TRUE;
    close(fD);
  }
  /*
    Now read in the normal, naked-eye star catalog.
  */
  sprintf(fileName, "%s/hipparcos.dat", dataDir);
  fD = open(fileName, O_RDONLY);
  if (fD < 0) {
    perror("opening Hipparcos star catalog.");
    exit(ERROR_EXIT);
  }
  readStarCatalog(fD);
  showingFaintStars = FALSE;
}

/*
  Draw an arrow pointing to a particular aziuth on the azimuth
  slew compass
 */
void drawAzArrow(int eraseOld, float az)
{
  int i;
  float sinAz, cosAz;
  static GdkPoint oldWedge[3];
  GdkPoint newWedge[3], unRotated[3];
  char comment[40];
  
  if (eraseOld)
    gdk_draw_polygon(drawingArea->window, blackGC, TRUE, oldWedge, 3);
  az -= M_PI*0.5;
  if (az < 0.0)
    az += M_2PI;
  sinAz = sinf(az); cosAz = cosf(az);
  unRotated[0].x = - AZ_COMPASS_POINT_HEIGHT/2;
  unRotated[0].y = -(AZ_COMPASS_RADIUS - 2*AZ_COMPASS_TICK_LENGTH - AZ_COMPASS_POINT_HEIGHT);
  unRotated[1].y = unRotated[0].y;
  unRotated[1].x = unRotated[0].x + AZ_COMPASS_POINT_HEIGHT;
  unRotated[2].x = 0.0;
  unRotated[2].y = unRotated[0].y - (3*AZ_COMPASS_POINT_HEIGHT)/2;
  for (i = 0; i < 3; i++) {
    newWedge[i].x = azCompassCenterX + cosAz*unRotated[i].x - sinAz*unRotated[i].y;
    newWedge[i].y = azCompassCenterY + cosAz*unRotated[i].y + sinAz*unRotated[i].x;
    oldWedge[i].x = newWedge[i].x; oldWedge[i].y = newWedge[i].y;
  }
  sprintf(comment, "az = %3.0f", az/DEGREES_TO_RADIANS);
  if (eraseOld) {
    gdk_draw_polygon(drawingArea->window, whiteGC, TRUE, newWedge, 3);
    gdk_draw_rectangle(drawingArea->window, blackGC, TRUE,
		       plotWidth/2 -
		       (gdk_string_width(smallFont, comment)/2) - 10,
		       plotHeight/2 + 20,
		       gdk_string_width(smallFont, comment) + 19,
		       32);
    gdk_draw_rectangle(drawingArea->window, creamGC, FALSE,
		       plotWidth/2 -
		       (gdk_string_width(smallFont, comment)/2) - 10,
		       plotHeight/2 + 20,
		       gdk_string_width(smallFont, comment) + 19,
		       32);
    gdk_draw_string(drawingArea->window, smallFont, creamGC,
		    plotWidth/2 -
		    (gdk_string_width(smallFont, comment)/2),
		    plotHeight/2 + 40,
		    comment);
  } else {
    gdk_draw_polygon(pixmap, whiteGC, TRUE, newWedge, 3);
    gdk_draw_rectangle(pixmap, blackGC, TRUE,
		       plotWidth/2 -
		       (gdk_string_width(smallFont, comment)/2) - 10,
		       plotHeight/2 + 20,
		       gdk_string_width(smallFont, comment) + 19,
		       32);
    gdk_draw_rectangle(pixmap, creamGC, FALSE,
		       plotWidth/2 -
		       (gdk_string_width(smallFont, comment)/2) - 10,
		       plotHeight/2 + 20,
		       gdk_string_width(smallFont, comment) + 19,
		       32);
    gdk_draw_string(pixmap, smallFont, creamGC,
		    plotWidth/2 -
		    (gdk_string_width(smallFont, comment)/2),
		    plotHeight/2 + 40,
		    comment);
  }
}

/*
  Draw a large compass in the center of the screen which will allow the
  user to select a new central azimuth for the display.
*/
void drawAzCompass(void)
{
  int i;
  float az, xi, xo, yi, yo, fScreenCenterX, fScreenCenterY, fCompassRadius;
  float fInnerRadius, fInnerRadius2, cosAz, sinAz, sqr2Over2;
  GdkPoint point[3], smallArrow[3];

  inAzCompassMode = TRUE;
  removeAllSensitiveAreas();
  gdk_draw_rectangle(pixmap, drawingArea->style->black_gc,
		     TRUE, 0, 0,
		     drawingArea->allocation.width,
		     drawingArea->allocation.height);
  gdk_draw_string(pixmap, smallFont, creamGC,
		  plotWidth/2 -
		  (gdk_string_width(smallFont,
				    "Select a new center azimuth with finger or stylus")/2),
		  20, "Select a new center azimuth with finger or stylus");
  azCompassCenterX = displayWidth/2;
  azCompassCenterY = displayHeight/2;
  gdk_draw_arc(pixmap, blueGC, FALSE,
	       azCompassCenterX-AZ_COMPASS_RADIUS,
	       azCompassCenterY-AZ_COMPASS_RADIUS, AZ_COMPASS_RADIUS*2,
	       AZ_COMPASS_RADIUS*2, 0, FULL_CIRCLE);
  point[0].x = azCompassCenterX - AZ_COMPASS_POINT_HEIGHT/2;
  point[0].y = azCompassCenterY - AZ_COMPASS_RADIUS;
  point[1].y = point[0].y;
  point[1].x = point[0].x + AZ_COMPASS_POINT_HEIGHT;
  point[2].x = azCompassCenterX;
  point[2].y = point[0].y - (3*AZ_COMPASS_POINT_HEIGHT)/2;
  gdk_draw_polygon(pixmap, creamGC, TRUE, point, 3);
  gdk_draw_string(pixmap, smallFont, greenGC,
		  azCompassCenterX-gdk_string_width(smallFont, "N")/2,
		  point[2].y - 8, "N");
  point[0].x = azCompassCenterX - AZ_COMPASS_POINT_HEIGHT/2;
  point[0].y = azCompassCenterY + AZ_COMPASS_RADIUS;
  point[1].y = point[0].y;
  point[1].x = point[0].x + AZ_COMPASS_POINT_HEIGHT;
  point[2].x = azCompassCenterX;
  point[2].y = point[0].y + (3*AZ_COMPASS_POINT_HEIGHT)/2;
  gdk_draw_polygon(pixmap, creamGC, TRUE, point, 3);
  gdk_draw_string(pixmap, smallFont, greenGC,
		  azCompassCenterX-gdk_string_width(smallFont, "S")/2,
		  point[2].y + 15, "S");
  point[0].x = azCompassCenterX - AZ_COMPASS_RADIUS;
  point[0].y = azCompassCenterY - AZ_COMPASS_POINT_HEIGHT/2;
  point[1].x = point[0].x;
  point[1].y = point[0].y + AZ_COMPASS_POINT_HEIGHT;
  point[2].y = azCompassCenterY;
  point[2].x = point[0].x - (3*AZ_COMPASS_POINT_HEIGHT)/2;
  gdk_draw_polygon(pixmap, creamGC, TRUE, point, 3);
  gdk_draw_string(pixmap, smallFont, greenGC,
		  point[2].x - 12, azCompassCenterY + 5, "W");
  point[0].x = azCompassCenterX + AZ_COMPASS_RADIUS;
  point[0].y = azCompassCenterY - AZ_COMPASS_POINT_HEIGHT/2;
  point[1].x = point[0].x;
  point[1].y = point[0].y + AZ_COMPASS_POINT_HEIGHT;
  point[2].y = azCompassCenterY;
  point[2].x = point[0].x + (3*AZ_COMPASS_POINT_HEIGHT)/2;
  gdk_draw_polygon(pixmap, creamGC, TRUE, point, 3);
  gdk_draw_string(pixmap, smallFont, greenGC,
		  point[2].x + 10, azCompassCenterY + 5, "E");
  smallArrow[0].x = -AZ_COMPASS_POINT_HEIGHT/4;
  smallArrow[0].y = 0;
  smallArrow[1].x = smallArrow[0].x + AZ_COMPASS_POINT_HEIGHT/2;
  smallArrow[1].y = smallArrow[0].y;
  smallArrow[2].x = 0;
  smallArrow[2].y = (3*AZ_COMPASS_POINT_HEIGHT)/4;
  sqr2Over2 = sqrtf(2.0)*0.5;
  for (i = 0; i < 3; i++) {
    point[i].x = azCompassCenterX +
      roundf(sqr2Over2*((float)( AZ_COMPASS_RADIUS + smallArrow[i].x + smallArrow[i].y)));
    point[i].y = azCompassCenterY +
      roundf(sqr2Over2*((float)( AZ_COMPASS_RADIUS + smallArrow[i].y - smallArrow[i].x)));
  }
  gdk_draw_polygon(pixmap, creamGC, TRUE, point, 3);
  for (i = 0; i < 3; i++) {
    point[i].x = azCompassCenterX +
      roundf(sqr2Over2*((float)(-AZ_COMPASS_RADIUS - smallArrow[i].x - smallArrow[i].y)));
    point[i].y = azCompassCenterY +
      roundf(sqr2Over2*((float)( AZ_COMPASS_RADIUS + smallArrow[i].y - smallArrow[i].x)));
  }
  gdk_draw_polygon(pixmap, creamGC, TRUE, point, 3);
  for (i = 0; i < 3; i++) {
    point[i].x = azCompassCenterX +
      roundf(sqr2Over2*((float)(-AZ_COMPASS_RADIUS + smallArrow[i].x - smallArrow[i].y)));
    point[i].y = azCompassCenterY +
      roundf(sqr2Over2*((float)(-AZ_COMPASS_RADIUS - smallArrow[i].y - smallArrow[i].x)));
  }
  gdk_draw_polygon(pixmap, creamGC, TRUE, point, 3);
  for (i = 0; i < 3; i++) {
    point[i].x = azCompassCenterX +
      roundf(sqr2Over2*((float)( AZ_COMPASS_RADIUS + smallArrow[i].x + smallArrow[i].y)));
    point[i].y = azCompassCenterY +
      roundf(sqr2Over2*((float)(-AZ_COMPASS_RADIUS - smallArrow[i].y + smallArrow[i].x)));
  }
  gdk_draw_polygon(pixmap, creamGC, TRUE, point, 3);
  fScreenCenterX = (float)azCompassCenterX;
  fScreenCenterY = (float)azCompassCenterY;
  fCompassRadius = (float)AZ_COMPASS_RADIUS;
  fInnerRadius   = (float)(AZ_COMPASS_RADIUS - AZ_COMPASS_TICK_LENGTH);
  fInnerRadius2  = (float)(AZ_COMPASS_RADIUS - AZ_COMPASS_TICK_LENGTH/2);
  i = 0;
  for (az = 0.0; az < 360.0*DEGREES_TO_RADIANS; az += 5.0*DEGREES_TO_RADIANS) {
    sinAz = sinf(az); cosAz = cosf(az);
    xo = fScreenCenterX + fCompassRadius*cosAz;
    yo = fScreenCenterY + fCompassRadius*sinAz;
    if (i & 1) {
      xi = fScreenCenterX + fInnerRadius2*cosAz;
      yi = fScreenCenterY + fInnerRadius2*sinAz;
    } else {
      xi = fScreenCenterX + fInnerRadius*cosAz;
      yi = fScreenCenterY + fInnerRadius*sinAz;
    }
    gdk_draw_line(pixmap, blueGC, (int)xi, (int)yi, (int)xo, (int)yo);
    i++;
  }
  drawAzArrow(FALSE, centerAz);
}

static void fullRedraw(int drawAzSlewCompass)
{
  /*
  double startTime, endTime;
  struct timespec timeSpecNow;

  clock_gettime(CLOCK_REALTIME, &timeSpecNow);
  startTime = (double)timeSpecNow.tv_sec + ((double)timeSpecNow.tv_nsec)*1.0e-9;
  */
  if (drawAzSlewCompass) {
    drawAzCompass();
  } else {
    if ((!inFlashlightMode) && (!(displayingAnOptsPage && (aboutScreen == ABOUT_SCREEN)))) {
      if (shouldSayRedrawing)
	sayRedrawing();
      removeAllSensitiveAreas();
      redrawScreen();
    }
  }
  /*
  clock_gettime(CLOCK_REALTIME, &timeSpecNow);
  endTime = (double)timeSpecNow.tv_sec + ((double)timeSpecNow.tv_nsec)*1.0e-9;
  printf("The update took %f seconds\n", endTime-startTime);
  */
  gdk_draw_drawable(drawingArea->window,
		    drawingArea->style->fg_gc[GTK_WIDGET_STATE (drawingArea)],
		    pixmap,
		    0,0,0,0,
		    displayWidth, displayHeight);
}

static int periodicUpdate(gpointer data)
{
  dayInc = 0;
  if (skyIsVisible && !(inFlashlightMode || inAzCompassMode)
      && (!(displayingAnOptsPage && 
	    ((aboutScreen == ABOUT_SCREEN) || 
	     (aboutScreen == METEOR_SHOWERS_SCREEN) ||
	     (aboutScreen == SOLAR_SYSTEM_SCHEMATIC_SCREEN) ||
	     (aboutScreen == SOLAR_SYSTEM_SCALE_SCREEN)))))
    fullRedraw(FALSE);
  return(TRUE);
}

void switchScreens(void)
{
  if (labelMode) {
    limitingMag = limitingMagnitude1;
    magScale = magScale1;
    displayPlanetsAsSymbols = FALSE;
    displayConstellations = FALSE;
    showStars = showStars1;
    showPlanets = showPlanets1;
    showNames = showNames1;
    showMeteors = showMeteors1;
    showGreatCircles = showGreatCircles1;
    showDeepSky = showDeepSky1;
    labelMode = FALSE;
  } else {
    limitingMag = limitingMagnitude2;
    magScale = magScale2;
    displayPlanetsAsSymbols = TRUE;
    displayConstellations = TRUE;
    showStars = showStars2;
    showNames = showNames2;
    showMeteors = showMeteors2;
    showPlanets = showPlanets2;
    showGreatCircles = showGreatCircles2;
    showDeepSky = showDeepSky2;
    labelMode = TRUE;
  }
  shouldSwitchScreens = FALSE;
}

static gboolean buttonPressEvent(GtkWidget *widget, GdkEventButton *event)
{
  clock_gettime(CLOCK_REALTIME, &timeSpecNow);
  buttonPressTime = (double)timeSpecNow.tv_sec + ((double)timeSpecNow.tv_nsec)*1.0e-9;
  /* Get rid of the old user drawn outline, if it exists */
  if (nUserPoints > 0) {
    free(userPoly);
    nUserPoints = 0;
  }
  buttonPressed = TRUE;
  notAPan = FALSE;
  return(TRUE);
}

static gboolean motionNotifyEvent(GtkWidget *widget, GdkEventButton *event)
{
  int x, y;
  
  if (buttonPressed) {
    x = event->x; y = event->y;
    if (inAzCompassMode) {
      int iTheta;
      float dx, dy;
      
      dx = (float)(azCompassCenterX - x); dy = (float)(azCompassCenterY - y);
      chosenAz = atan2(dy, dx);
      iTheta = 5*roundf(chosenAz / (5.0*DEGREES_TO_RADIANS));
      chosenAz = ((float)iTheta)*DEGREES_TO_RADIANS;
      drawAzArrow(TRUE, chosenAz);
      gdk_flush();
    } else if (!displayingAnOptsPage && !inFlashlightMode) {
      if (nUserPoints == 0) {
	userPoly = malloc(sizeof(GdkPoint));
	if (userPoly == NULL) {
	  perror("malloc of userPoly");
	  exit(ERROR_EXIT);
	}
      } else {
	userPoly = realloc(userPoly, (nUserPoints+1)*sizeof(GdkPoint));
	if (userPoly == NULL) {
	  perror("realloc of userPoly");
	  exit(ERROR_EXIT);
	}
      }
      userPoly[nUserPoints  ].x = x;
      userPoly[nUserPoints++].y = y;
      /*
	Check to see if the pointer has moved a lot (> 100 pixels) while
	the screen has been pressed.  If so, set notAPan flag to disable
	presenting the azimuth compass.   This is required to allow zoom gestures
	which begin in the panning area.
      */
      if (!notAPan)
	if ((x-userPoly[0].x)*(x-userPoly[0].x) +
	    (y-userPoly[0].y)*(y-userPoly[0].y) > 10000)
	  notAPan = TRUE;
      gdk_draw_point(drawingArea->window, greenGC, x, y);
      gdk_flush();
    }
  }
  return(TRUE);
}

static gboolean processTap(int x, int y)
{
  int found = FALSE;
  sensitiveArea *sA = sensitiveAreaRoot;

  dprintf("In processTap, (%d, %d)\n", x, y);
  shouldSayRedrawing = TRUE;
  if (inFlashlightMode) {
    fullscreenStateChanging = TRUE;
    shouldSayRedrawing = FALSE;
    gtk_widget_show(controlButtonBox);
    gtk_window_unfullscreen(GTK_WINDOW(window));
    inFlashlightMode = FALSE;
    fullRedraw(FALSE);
  } else if (displayingAnOptsPage && (aboutScreen != SMALL_MOONCAL_SCREEN) &&
	     (aboutScreen != SOLAR_SYSTEM_SCHEMATIC_SCREEN) &&
	     (aboutScreen != SOLAR_SYSTEM_SCALE_SCREEN)) {
    shouldSayRedrawing = FALSE;
    if (aboutScreen == BIG_MOONCAL_SCREEN)
      gtk_window_unfullscreen(GTK_WINDOW(window));
    gtk_widget_show(controlButtonBox);
    displayingAnOptsPage = FALSE;
    fullRedraw(FALSE);
  } else {
    while ((sA != NULL) && (!found)) {
      dprintf("Checking item of type %d  BLC: %d, %d  TRC: %d, %d\n",
		sA->type, sA->bLCX, sA->bLCY, sA->tRCX, sA->tRCY);
      if ((x >= sA->bLCX) && (x <= sA->tRCX)
	    && (y >= sA->bLCY) && (y <= sA->tRCY))
	found = TRUE;
      else
	sA = sA->forwardPointer;
    }
    if (found) {
      dprintf("Found sensitive area - type = %d value = %f\n",
	      sA->type, sA->value);
      switch(sA->type) {
      case SA_TOP_AREA:
	if (inFullscreenMode) {
	  fullscreenStateChanging = TRUE;
	  shouldSayRedrawing = FALSE;
	  gtk_widget_show(controlButtonBox);
	  gtk_window_unfullscreen(GTK_WINDOW(window));
	  inFullscreenMode = FALSE;
	}
	break;
      case SA_FINGER_PAN_AREA:
	{
	  int centerAzInc, slewCompass;
	  
	  centerAzInc = (int)(0.2*(azSpan/DEGREES_TO_RADIANS) *
			      (double)(x - displayWidth/2)/(double)(displayWidth));
	  dprintf("In finger pan area  x = %d, y = %d, inc by %d degrees\n",
		  x, y, centerAzInc*5);
	  if ((centerAzInc == 0) && !zoomed)
	    slewCompass = TRUE;
	  else
	    slewCompass = FALSE;
	  centerAz += ((float)centerAzInc*5)*DEGREES_TO_RADIANS;
	  if (centerAz < 0.0)
	    centerAz += M_2PI;
	  if (centerAz > M_2PI)
	    centerAz -= M_2PI;
	  centerAzD = (double)centerAz;
	  if ((centerAzInc != 0) || !zoomed)
	    fullRedraw(slewCompass);
	}
	break;
      case SA_MONTH_LEFT_ARROW:
      case SA_MONTH_RIGHT_ARROW:
	smallMooncalDateOffset += sA->value;
	shouldSayRedrawing = FALSE;
	fullRedraw(FALSE);
	break;
      case SA_SOLAR_SYSTEM_BUTTON:
	dayInc = (int)sA->value;
	if (dayInc != 0) {
	  if (dayInc == ZOOM_SOLAR_SYSTEM_IN) {
	    outermostPlanet -= 1;
	    if (outermostPlanet == MOON)
	      outermostPlanet = EARTH;
	    dayInc = 0;
	  } else if (dayInc == ZOOM_SOLAR_SYSTEM_OUT) {
	    outermostPlanet += 1;
	    if (outermostPlanet == MOON)
	      outermostPlanet = MARS;
	    dayInc = 0;
	  }
	  shouldSayRedrawing = FALSE;
	  fullRedraw(FALSE);
	}
	break;
      default:
	fprintf(stderr, "Unknown SA type (%d)  BLC: %d, %d  TRC: %d, %d\n",
		sA->type, sA->bLCX, sA->bLCY, sA->tRCX, sA->tRCY);
      }
    } else {
      if (displayingAnOptsPage) {
	smallMooncalDateOffset = 0.0;
	if ((aboutScreen == SOLAR_SYSTEM_SCHEMATIC_SCREEN) ||
	    (aboutScreen == SOLAR_SYSTEM_SCALE_SCREEN)) {
	  dayInc = 0;
	  useCurrentTime = useCalendarTime = useJulianTime = FALSE;
	  switch (savedTimeMode) {
	  case 1:
	    useCalendarTime = TRUE;
	    break;
	  case 2:
	    useJulianTime = TRUE;
	    break;
	  default:
	    useCurrentTime = TRUE;
	  }
	  tJD = savedTJD;
	}
	shouldSayRedrawing = FALSE;
	gtk_widget_show(controlButtonBox);
	displayingAnOptsPage = FALSE;
      } else {
	shouldSwitchScreens = TRUE;
	if (constellationsInitialized)
	  switchScreens();
	else {
	  sayRedrawing();
	  return(TRUE);
	}
      }
      fullRedraw(FALSE);
    }
    dprintf("Tap handler, centerAz = %f\n",
	    centerAz/DEGREES_TO_RADIANS);
  }
  return(TRUE);
}

static void unzoom(void)
{
  float centerAzDeg;

  zoomed = FALSE;
  if (isnan(centerAz) || isnan(centerAzD)) {
    printf("Doing NaN centerAz reset\n");
    centerAz = 270.0*DEGREES_TO_RADIANS;
    centerAzD = (double)centerAz;
  }
  maximumZA  =  95.0*DEGREES_TO_RADIANS;
  azSpan = 90.0*DEGREES_TO_RADIANS;
  minimumZA  = thetaOffset = 0.0;
  piMinusThetaOffset = M_PI - thetaOffset;
  twoPiMinusThetaOffset = M_2PI - thetaOffset;
  centerAzDeg = centerAz/DEGREES_TO_RADIANS;
  centerAzDeg = ((float)((int)(centerAzDeg*0.2 + 0.5)))*5.0;
  centerAz = centerAzDeg*DEGREES_TO_RADIANS;
  centerAzD = (double)centerAz;
  fullRedraw(FALSE);
}

static gboolean buttonReleaseEvent(GtkWidget *widget, GdkEventButton *event)
{
  int x, y, tap;
  double touchDuration;

  if (inAzCompassMode) {
    centerAz = chosenAz;
    centerAzD = (double)centerAz;
    buttonPressed = FALSE;
    fullRedraw(TRUE);
    inAzCompassMode = FALSE;
    fullRedraw(FALSE);
  } else {
    clock_gettime(CLOCK_REALTIME, &timeSpecNow);
    buttonReleaseTime = (double)timeSpecNow.tv_sec + ((double)timeSpecNow.tv_nsec)*1.0e-9;
    touchDuration = buttonReleaseTime - buttonPressTime;
    x = event->x; y = event->y;
    if (touchDuration < 0.25)
      tap = TRUE;
    else
      tap = FALSE;
    dprintf("In buttonReleaseEvent, event->button %d (%d,%d) centerAz = %f\n",
	    event->button, x, y, centerAz/DEGREES_TO_RADIANS);
    dprintf("touchDuration = %f, tap = %d\n", touchDuration, tap);
    if (tap || displayingAnOptsPage)
      return(processTap(x, y));
    else if (!zoomed && !notAPan && (y > 4*displayHeight/5))
      /*
	This will cause the azimuth compass to appear if the user presses anywhere
	in the finger-pan area for a prolonged time.
      */
      return(processTap(displayWidth/2, y));
    else {
      /* Check for a gesture */
      if (nUserPoints > 2) {
	int xMin, xMax, yMin, yMax;
	int i, iTemp, pan;
	int x1, x2, y1, y2;
	float area = 0.0;
	float perimeter = 0.0;
	float xCen = 0.0;
	float yCen = 0.0;
	float maxArea, mX, mY, eta, azMin, zA1, azMax, zA2, xSide, ySide;
	double tThetaOffset, dummy;
	
	
	for (i = 0; i < nUserPoints; i++) {
	  x1 = userPoly[i].x; y1 = userPoly[i].y;
	  x2 = userPoly[(i+1) % nUserPoints].x; y2 = userPoly[(i+1) % nUserPoints].y;
	  iTemp = x1*y2 - x2*y1;
	  area += (float)iTemp;
	  perimeter += sqrt((float)((x1-x2)*(x1-x2) + (y1-y2)*(y1-y2)));
	  xCen += (float)((x1+x2) * iTemp);
	  yCen += (float)((y1+y2) * iTemp);
	}
	area = fabs(area)*0.5;
	maxArea = perimeter*perimeter/(4.0*M_PI);
	if (area/maxArea < 0.2)
	  pan = TRUE;
	else
	  pan = FALSE;
	dprintf("Area = %f, ratio = %f, pan = %d\n", area, area/maxArea, pan);
	if ((area > 300.0) && (!pan)) {
	  GdkPoint equivRec[4];
	  float ratio;
	  char *zooming = "ZOOMING";
	  
	  gdk_draw_string(drawingArea->window, smallFont, greenGC,
			  displayWidth-gdk_string_width(smallFont, zooming)-10,
			  yTopLabelSkip+gdk_string_height(smallFont, zooming)+10,
			  zooming);
	  xCen = fabs(xCen) / (6.0*area);
	  yCen = fabs(yCen) / (6.0*area);
	  dprintf("xCen = %f, yCen = %f\n", xCen, yCen);
	  gdk_draw_polygon(drawingArea->window, greenGC, FALSE, userPoly, nUserPoints);
	  ratio = (float)(displayHeight)/(float)(displayWidth);
	  xSide = 0.5 * sqrt(area/ratio); ySide = xSide * ratio;
	  xMin = (int)(xCen - xSide); yMin = (int)(yCen - ySide);
	  xMax = (int)(xCen + xSide); yMax = (int)(yCen + ySide);
	  equivRec[0].x = xMin; equivRec[0].y = yMin; 
	  equivRec[1].x = xMax; equivRec[1].y = yMin; 
	  equivRec[2].x = xMax; equivRec[2].y = yMax; 
	  equivRec[3].x = xMin; equivRec[3].y = yMax; 
	  dprintf("Bottom (%d, %d) -> (%d, %d)\n",
		  xMin, yMax, xMax, yMax);
	  pixelsToMercator(xCen, yMin, plotWidth, plotHeight, &mX, &mY);
	  invMercator(mX, &eta);
	  invThetaEta(mY, eta, &azMin, &zA1);
	  pixelsToMercator(xCen, yMax, plotWidth, plotHeight, &mX, &mY);
	  invMercator(mX, &eta);
	  invThetaEta(mY, eta, &azMax, &zA2);
	  if (zA2 < zA1) {
	    float tempZA;
	    
	    tempZA = zA2;
	    zA2 = zA1;
	    zA1 = tempZA;
	  }
	  if (azMax < azMin) {
	    float azTemp;
	    
	    azTemp = azMax;
	    azMax = azMin;
	    azMin = azTemp;
	  }
	  if (zA2 > ABSOLUTE_MAX_ZA)
	    zA2 = ABSOLUTE_MAX_ZA;
	  maximumZA = zA2;
	  minimumZA = zA1;
	  dprintf("Max ZA = %f, Min ZA = %f\n", maximumZA/DEGREES_TO_RADIANS,
		  minimumZA/DEGREES_TO_RADIANS);
	  thetaEta((double)((azMax+azMin)*0.5), minimumZA, &tThetaOffset, &dummy);
	  thetaOffset += tThetaOffset;
	  centerAz = (azMin+azMax+M_PI)*0.5;
	  centerAzD = (double)centerAz;
	  pixelsToMercator(xMin, yMax, plotWidth, plotHeight, &mX, &mY);
	  invMercator(mX, &eta);
	  invThetaEta(mY, eta, &azMin, &zA1);
	  pixelsToMercator(xMax, yMax, plotWidth, plotHeight, &mX, &mY);
	  invMercator(mX, &eta);
	  invThetaEta(mY, eta, &azMax, &zA2);
	  azSpan = fabs(azMax - azMin);
	  dprintf("azSpan = %f\n", azSpan/DEGREES_TO_RADIANS);
	  dprintf("thetaOffset = %f (%f)\n", thetaOffset, thetaOffset/DEGREES_TO_RADIANS);
	  gdk_draw_polygon(drawingArea->window, redGC, FALSE, equivRec, 4);
	  zoomed = TRUE;
	  gdk_flush();
	  fullRedraw(FALSE);
	} else if ((maxArea > 750.0) && pan && zoomed) {
	  int deltaX, deltaY, signDX, signDY;
	  float zACen, azCen, deltaZA, theta;
	  char *panning = "PANNING";
	  GdkPoint points[5];
	  
	  gdk_draw_string(drawingArea->window, smallFont, greenGC,
			  displayWidth-gdk_string_width(smallFont, panning)-10,
			  yTopLabelSkip+gdk_string_height(smallFont, panning)+10,
			  panning);
	  deltaX = userPoly[0].x - userPoly[nUserPoints-1].x;
	  deltaY = userPoly[0].y - userPoly[nUserPoints-1].y;
	  theta = atan2f(-deltaY, -deltaX);
	  dprintf("Theta = %f\n", theta/DEGREES_TO_RADIANS);
	  if (deltaX >= 0)
	    signDX = 1;
	  else
	    signDX = -1;
	  if (deltaY >= 0)
	    signDY = 1;
	  else
	    signDY = -1;
	  dprintf("deltaX = %d, deltaY = %d\n", deltaX, deltaY);
	  xCen = (float)(displayWidth/2 + deltaX);
	  yCen = (float)(displayHeight/2 + deltaY);
	  pixelsToMercator(xCen, yCen, plotWidth, plotHeight, &mX, &mY);
	  invMercator(mX, &eta);
	  invThetaEta(mY, eta, &azCen, &zACen);
	  deltaZA = 0.5*(maximumZA-minimumZA);
	  maximumZA = zACen + deltaZA;
	  minimumZA = zACen - deltaZA;
	  thetaEta((double)azCen, minimumZA, &tThetaOffset, &dummy);
	  thetaOffset += tThetaOffset;
	  centerAz = azCen + M_PI*0.5;
	  centerAzD = (double)centerAz;
	  points[0].x = userPoly[0].x;        points[0].y = userPoly[0].y;
	  points[1].x = points[0].x - deltaX; points[1].y = points[0].y - deltaY;
	  gdk_draw_lines(drawingArea->window, redGC, points, 2);
	  points[2].x = points[1].x + ((int)(20.0*cosf(theta + 0.75*M_PI)));
	  points[2].y = points[1].y + ((int)(20.0*sinf(theta + 0.75*M_PI)));
	  points[3].x = points[1].x + ((int)(20.0*cosf(theta + 1.25*M_PI)));
	  points[3].y = points[1].y + ((int)(20.0*sinf(theta + 1.25*M_PI)));
	  points[4].x = points[1].x;
	  points[4].y = points[1].y;
	  gdk_draw_polygon(drawingArea->window, redGC, TRUE, &points[2], 3);
	  gdk_flush();
	  fullRedraw(FALSE);
	} else {
	  if (zoomed)
	    unzoom();
	  else
	    return(processTap(x, y));
	}
      } else {
	if (zoomed)
	  unzoom();
	else
	  return(processTap(x, y));
      }
    }
  }
  buttonPressed = FALSE;
  return(TRUE);
}

int tokenCheck(char *line, char *token, int type, void *value)

#define INT_TOKEN     0
#define DOUBLE_TOKEN  1
#define FLOAT_TOKEN   2
#define STRING_TOKEN  3

/*
  Scan "line" for "token".   If found, read the value into
  "value" as an integer, double or float, depending on "type".

  Return TRUE IFF the token is seen.
*/
{
  if (strstr(line, token)) {
    int nRead;

    switch (type) {
    case INT_TOKEN:
      nRead = sscanf(&((char *)strstr(line, token))[strlen(token)+1], "%d", (int *)value);
      if (nRead != 1) {
	fprintf(stderr, "Unable to parse config file line \"%s\"\n", line);
	return(FALSE);
      }
      dprintf("%s = %d\n", token, *((int *)value));
      break;
    case DOUBLE_TOKEN:
      nRead = sscanf(&((char *)strstr(line, token))[strlen(token)+1], "%lf", (double *)value);
      if (nRead != 1) {
	fprintf(stderr, "Unable to parse config file line \"%s\"\n", line);
	return(FALSE);
      }
      dprintf("%s = %f\n", token, *((double *)value));
      break;
    case FLOAT_TOKEN:
      nRead = sscanf(&((char *)strstr(line, token))[strlen(token)+1], "%f", (float *)value);
      if (nRead != 1) {
	fprintf(stderr, "Unable to parse config file line \"%s\"\n", line);
	return(FALSE);
      }
      dprintf("%s = %f\n", token, *((float *)value));
      break;
    case STRING_TOKEN:
      nRead = sscanf(&((char *)strstr(line, token))[strlen(token)+1], "%s", (char *)value);
      if (nRead != 1) {
	fprintf(stderr, "Unable to parse config file line \"%s\"\n", line);
	return(FALSE);
      }
      dprintf("%s = \"%s\"\n", token, (char *)value);
      break;
    default:
      fprintf(stderr, "Unrecognized type (%d) passed to tokenCheck\n", type);
    }
    return(TRUE);
  } else
    return(FALSE);
}

/*
  Read in the values that the user can set.   All the parameters which can be set
  via the config file have default hard-coded values, so if a particular
  parameter is not present in the config file, or if the entire
*/
void parseConfigFile(void)
{
  int eOF = FALSE;
  int lineNumber = 0;
  int configFD;
  char inLine[MAX_FILE_NAME_SIZE];

  sprintf(inLine, "%s/config", dataDir);
  configFD = open(inLine, O_RDONLY);
  if (configFD < 0) {
    perror("config");
    fprintf(stderr, "Full file name was \"%s\"\n", inLine);
    return;
  }
  while (!eOF) {
    lineNumber++;
    if (getLine(configFD, &inLine[0], &eOF))
      if (strlen(inLine) > 0) {
	if (tokenCheck(inLine, "LIMITING_MAGNITUDE1", FLOAT_TOKEN, &limitingMagnitude1)
	    && (!labelMode)) {
	  limitingMag = limitingMagnitude1;
	  magScale = magScale1 = VISUAL_LIMIT/limitingMagnitude1;
	}
	if (tokenCheck(inLine, "LIMITING_MAGNITUDE2", FLOAT_TOKEN,
		       &limitingMagnitude2) && labelMode) {
	  limitingMag = limitingMagnitude2;
	  magScale = magScale2 = VISUAL_LIMIT/limitingMagnitude2;
	}
	if (tokenCheck(inLine, "SHOW_GREAT_CIRCLES1", INT_TOKEN, &showGreatCircles1))
	  if (!labelMode)
	    showGreatCircles = showGreatCircles1;
	if (tokenCheck(inLine, "SHOW_GREAT_CIRCLES2", INT_TOKEN, &showGreatCircles2))
	  if (labelMode)
	    showGreatCircles = showGreatCircles2;
	tokenCheck(inLine, "USE_ASTERISMS", INT_TOKEN, &useAsterisms);
	if (tokenCheck(inLine, "SHOW_DEEP_SKY1", INT_TOKEN, &showDeepSky1))
	  if (!labelMode)
	    showDeepSky = showDeepSky1;
	if (tokenCheck(inLine, "SHOW_DEEP_SKY2", INT_TOKEN, &showDeepSky2))
	  if (labelMode)
	    showDeepSky = showDeepSky2;
	if (tokenCheck(inLine, "SHOW_STARS1", INT_TOKEN, &showStars1) && (!labelMode))
	  showStars = showStars1;
	if (tokenCheck(inLine, "SHOW_STARS2", INT_TOKEN, &showStars2) && (labelMode))
	  showStars = showStars2;
	if (tokenCheck(inLine, "SHOW_PLANETS1", INT_TOKEN, &showPlanets1) && (!labelMode))
	  showPlanets = showPlanets1;
	if (tokenCheck(inLine, "SHOW_PLANETS2", INT_TOKEN, &showPlanets2) && (labelMode))
	  showPlanets = showPlanets2;
	if (tokenCheck(inLine, "SHOW_STAR_NAMES1", INT_TOKEN, &showNames1) && (!labelMode))
	  showNames = showNames1;
	if (tokenCheck(inLine, "SHOW_STAR_NAMES2", INT_TOKEN, &showNames2) && (labelMode))
	  showNames = showNames2;
	if (tokenCheck(inLine, "SHOW_METEORS1", INT_TOKEN, &showMeteors1) && (!labelMode))
	  showMeteors = showMeteors1;
	if (tokenCheck(inLine, "SHOW_METEORS2", INT_TOKEN, &showMeteors2) && (labelMode))
	  showMeteors = showMeteors2;
	tokenCheck(inLine, "DEBUG_MESSAGES_ON", INT_TOKEN, &debugMessagesOn);
	tokenCheck(inLine, "USE_GPSD", INT_TOKEN, &useGPSD);
	if (tokenCheck(inLine, "INITIAL_AZIMUTH", FLOAT_TOKEN, &initialAzimuth)) {
	  centerAzD = (initialAzimuth + 90.0)*DEGREES_TO_RADIANS;
	  centerAz  = (float)centerAzD;
	}
	if (tokenCheck(inLine, "DEFAULT_LOCATION_NAME", STRING_TOKEN, &locationName[0])) {
	  int i;

	  for (i = 0; i < strlen(locationName); i++)
	    if (locationName[i] == '_')
	      locationName[i] = ' ';
	}
	if (tokenCheck(inLine, "DEFAULT_LOCATION_LATITUDE", DOUBLE_TOKEN, &latitude)) {
	  latitude *= DEGREES_TO_RADIANS;
	  locationChanged = TRUE;
	}
	if (tokenCheck(inLine, "DEFAULT_LOCATION_LONGITUDE", DOUBLE_TOKEN, &longitude)) {
	  longitude *= DEGREES_TO_RADIANS;
	  locationChanged = TRUE;
	}
      }
  }
  close(configFD);
}

/*
  Write out the values that the user can set.
*/
void writeConfigFile(void)
{
  int i;
  char tempName[MAX_FILE_NAME_SIZE], newName[MAX_FILE_NAME_SIZE], oldName[MAX_FILE_NAME_SIZE];
  FILE *newConfigFile;

  sprintf(newName, "%s/config.new", dataDir);
  newConfigFile = fopen(newName, "w");
  if (newConfigFile == NULL) {
    perror("config.new");
    return;
  }
  dprintf("Writing config file.\n");
  fprintf(newConfigFile, "LIMITING_MAGNITUDE1 %6.2f\n", limitingMagnitude1);
  fprintf(newConfigFile, "LIMITING_MAGNITUDE2 %6.2f\n", limitingMagnitude2);
  fprintf(newConfigFile, "SHOW_GREAT_CIRCLES1 %d\n", showGreatCircles1);
  fprintf(newConfigFile, "SHOW_GREAT_CIRCLES2 %d\n", showGreatCircles2);
  fprintf(newConfigFile, "USE_ASTERISMS %d\n", useAsterisms);
  fprintf(newConfigFile, "SHOW_DEEP_SKY1 %d\n", showDeepSky1);
  fprintf(newConfigFile, "SHOW_DEEP_SKY2 %d\n", showDeepSky2);
  fprintf(newConfigFile, "SHOW_STARS1 %d\n", showStars1);
  fprintf(newConfigFile, "SHOW_STARS2 %d\n", showStars2);
  fprintf(newConfigFile, "SHOW_PLANETS1 %d\n", showPlanets1);
  fprintf(newConfigFile, "SHOW_PLANETS2 %d\n", showPlanets2);
  fprintf(newConfigFile, "SHOW_STAR_NAMES1 %d\n", showNames1);
  fprintf(newConfigFile, "SHOW_STAR_NAMES2 %d\n", showNames2);
  fprintf(newConfigFile, "SHOW_METEORS1 %d\n", showMeteors1);
  fprintf(newConfigFile, "SHOW_METEORS2 %d\n", showMeteors2);
  fprintf(newConfigFile, "USE_GPSD %d\n", useGPSD);
  fprintf(newConfigFile, "INITIAL_AZIMUTH %7.2f\n", initialAzimuth);
  fprintf(newConfigFile, "DEBUG_MESSAGES_ON %d\n", debugMessagesOn);
  sprintf(tempName, "%s", locationName);
  for (i = 0; i < strlen(tempName); i++)
    if (tempName[i] == ' ')
      tempName[i] = '_';
  fprintf(newConfigFile, "DEFAULT_LOCATION_NAME %s\n", tempName);
  fprintf(newConfigFile, "DEFAULT_LOCATION_LATITUDE %f\n", latitude/DEGREES_TO_RADIANS);
  fprintf(newConfigFile, "DEFAULT_LOCATION_LONGITUDE %f\n", longitude/DEGREES_TO_RADIANS);
  fclose(newConfigFile);
  sprintf(oldName, "%s/config", dataDir);
  rename(newName, oldName);
}

/*
  This function is called when the fullscreen button is clicked, it toggles the display
  in and out of fullscreen mode, and hides the buttons when in fullscreen mode
*/
void fullscreenButtonClicked(GtkButton *button, gpointer userData)
{
  dprintf("In fullscreenButtonClicked\n");
  shouldSayRedrawing = FALSE;
  fullscreenStateChanging = TRUE;
  inFullscreenMode = TRUE;
  gtk_widget_hide(controlButtonBox);
  gtk_window_fullscreen(GTK_WINDOW(window));
}

int optionsModified = FALSE;

/*
  Remove the display options page, and replace it with the sky
  display.
*/
void removeItemsBox(void)
{
  gtk_widget_hide(itemsTable);
  gtk_container_remove(GTK_CONTAINER(window), itemsTable);
  gtk_container_add(GTK_CONTAINER(window), mainBox);
  gtk_widget_show(mainBox);
  if (optionsModified) {
    dprintf("Doing redraw because checkBoxes modified\n");
    fullRedraw(FALSE);
    optionsModified = FALSE;
  }
  skyIsVisible = TRUE;
}

void checkTimeSpinBoxes(void)
{
  needNewTime = TRUE;
  if (useCalendarTime) {
    cYear   = gtk_spin_button_get_value((GtkSpinButton *)yearSpin);
    if (cYear <= 0)
      cYear++; /* There is no year 0 */
    cMonth  = gtk_spin_button_get_value((GtkSpinButton *)monthSpin);
    cDay    = gtk_spin_button_get_value((GtkSpinButton *)daySpin);
    cHour   = gtk_spin_button_get_value((GtkSpinButton *)hourSpin);
    cMinute = gtk_spin_button_get_value((GtkSpinButton *)minuteSpin);
    cSecond = gtk_spin_button_get_value((GtkSpinButton *)secondSpin);
  } else if (useJulianTime)
    julianDate   = gtk_spin_button_get_value((GtkSpinButton *)julSpin);
}

void checkItemsSpinBoxes(void)
{
  double testValue;

  testValue = gtk_spin_button_get_value((GtkSpinButton *)magSpin1);
  if (fabs(testValue - limitingMagnitude1) > 0.001) {
    limitingMagnitude1 = testValue;
    if (limitingMagnitude1 > VISUAL_LIMIT)
      showFaintStars = TRUE;
    else if (limitingMagnitude1 <= VISUAL_LIMIT)
      showFaintStars = FALSE;
    magScale1 = VISUAL_LIMIT/limitingMagnitude1;
    optionsModified = TRUE;
    if (!labelMode) {
      limitingMag = limitingMagnitude1;
      magScale = magScale1;
    }
  }
  testValue = gtk_spin_button_get_value((GtkSpinButton *)magSpin2);
  if (fabs(testValue - limitingMagnitude2) > 0.001) {
    limitingMagnitude2 = testValue;
    magScale2 = VISUAL_LIMIT/limitingMagnitude1;
    optionsModified = TRUE;
    if (labelMode) {
      limitingMag = limitingMagnitude2;
      magScale = magScale2;
    }
  }
}

/*
  Save any new items options in the config file, and then
  return to the sky display.
*/
void itemsSave(GtkButton *button, gpointer userData)
{
  checkItemsSpinBoxes();
  writeConfigFile();
  removeItemsBox();
}

/*
  Call back function for the items page OK button, which
  redraws the sky without saving chosen options.
*/
void itemsOk(GtkButton *button, gpointer userData)
{
  checkItemsSpinBoxes();
  writeConfigFile();
  removeItemsBox();
}

GtkCheckButton *starCheckButton1, *starCheckButton2;
GtkCheckButton *starNameCheckButton1, *starNameCheckButton2;
GtkCheckButton *meteorCheckButton1, *meteorCheckButton2;
GtkCheckButton *planetCheckButton1, *planetCheckButton2;
GtkCheckButton *greatCirclesCheckButton1, *greatCirclesCheckButton2, *useAsterismsCheckButton;
GtkCheckButton *deepSkyCheckButton1, *deepSkyCheckButton2;
GtkCheckButton *gPSButton, *customLocationButton, *menuLocationButton;
GtkCheckButton *currentTimeButton, *calendarTimeButton, *julianTimeButton;
GtkWidget *locationNameLabel, *locationNameText;

GtkWidget *displayedLocationName;   /* Holds the name of a location selected from menu */
int locationNameIsDisplayed = FALSE;
int menuShowing = FALSE;
int privateMenuShowing = FALSE;
GtkWidget *activeMenu;

void gtk_table_remove (GtkContainer *container,
		       GtkWidget    *widget)
{
  GtkTable *table;
  GtkTableChild *child;
  GList *children;
  
  g_return_if_fail (GTK_IS_TABLE (container));
  g_return_if_fail (widget != NULL);
  
  table = GTK_TABLE (container);
  children = table->children;
  
  while (children)
    {
      child = children->data;
      children = children->next;
      
      if (child->widget == widget)
	{
	  gboolean was_visible = GTK_WIDGET_VISIBLE (widget);
	  
	  gtk_widget_unparent (widget);
	  
	  table->children = g_list_remove (table->children, child);
	  g_free (child);
	  
	  if (was_visible && GTK_WIDGET_VISIBLE (container))
	    gtk_widget_queue_resize (GTK_WIDGET (container));
	  break;
	}
    }
}

void setTimeSensitivities(void)
{
  gtk_toggle_button_set_active((GtkToggleButton *)currentTimeButton, useCurrentTime);
  gtk_toggle_button_set_active((GtkToggleButton *)calendarTimeButton, useCalendarTime);
  gtk_toggle_button_set_active((GtkToggleButton *)julianTimeButton, useJulianTime);

  gtk_widget_set_sensitive(cDateLabel, useCalendarTime);
  gtk_widget_set_sensitive(cTimeLabel, useCalendarTime);
  gtk_widget_set_sensitive(yearSpin, useCalendarTime);
  gtk_widget_set_sensitive(monthSpin, useCalendarTime);
  gtk_widget_set_sensitive(daySpin, useCalendarTime);
  gtk_widget_set_sensitive(hourSpin, useCalendarTime);
  gtk_widget_set_sensitive(minuteSpin, useCalendarTime);
  gtk_widget_set_sensitive(secondSpin, useCalendarTime);

  gtk_widget_set_sensitive(julSpin, useJulianTime);
  gtk_widget_set_sensitive(julSpinLabel, useJulianTime);
}

int savePrivateShowing = FALSE;
GtkWidget *saveAsPrivate;

void setLocationSensitivities(void)
{
  gtk_toggle_button_set_active((GtkToggleButton *)gPSButton, useGPSD);
  gtk_toggle_button_set_active((GtkToggleButton *)customLocationButton, useCustomLocation);
  gtk_toggle_button_set_active((GtkToggleButton *)menuLocationButton, useMenuLocation);
  if (!useMenuLocation) {
    if (locationNameIsDisplayed)
      gtk_widget_hide(displayedLocationName);
    if (menuShowing) 
      gtk_widget_hide(activeMenu);
  } else {
    if (locationNameIsDisplayed)
      gtk_widget_show(displayedLocationName);
    if (menuShowing) 
      gtk_widget_show(activeMenu);
  }

  gtk_widget_set_sensitive(locationNameLabel, useCustomLocation);
  gtk_widget_set_sensitive(locationNameText, useCustomLocation);
  gtk_widget_set_sensitive(locationLatName, useCustomLocation);
  gtk_widget_set_sensitive(latSpinDD, useCustomLocation);
  gtk_widget_set_sensitive(latSpinMM, useCustomLocation);
  gtk_widget_set_sensitive(latSpinSS, useCustomLocation);
  gtk_widget_set_sensitive(locationLongName, useCustomLocation);
  gtk_widget_set_sensitive(longSpinDD, useCustomLocation);
  gtk_widget_set_sensitive(longSpinMM, useCustomLocation);
  gtk_widget_set_sensitive(longSpinSS, useCustomLocation);
  gtk_widget_set_sensitive(regionMenu, useMenuLocation);
  if (useGPSD && savePrivateShowing) {
    gtk_widget_hide(saveAsPrivate);
    savePrivateShowing = FALSE;
  } else if (!useGPSD && !savePrivateShowing) {
    gtk_widget_show(saveAsPrivate);
    savePrivateShowing = TRUE;
  }
}

GtkWidget *oSave; 
int oSaveShowing = FALSE;

void checkButtonCallback(GtkButton *button, gpointer userData)
{
  if (strstr("showStars1", userData) != NULL) {
    if (gtk_toggle_button_get_active((GtkToggleButton *)starCheckButton1)) {
      optionsModified = TRUE;
      showStars1 = TRUE;
    } else {
      optionsModified = TRUE;
      showStars1 = FALSE;
    }
    if (!labelMode)
      showStars = showStars1;
  } else if (strstr("showStars2", userData) != NULL) {
    if (gtk_toggle_button_get_active((GtkToggleButton *)starCheckButton2)) {
      optionsModified = TRUE;
      showStars2 = TRUE;
    } else {
      optionsModified = TRUE;
      showStars2 = FALSE;
    }
    if (labelMode)
      showStars = showStars2;
  } else if (strstr("showNames1", userData) != NULL) {
    if (gtk_toggle_button_get_active((GtkToggleButton *)starNameCheckButton1)) {
      optionsModified = TRUE;
      showNames1 = TRUE;
    } else {
      optionsModified = TRUE;
      showNames1 = FALSE;
    }
    if (!labelMode)
      showNames = showNames1;
  } else if (strstr("showNames2", userData) != NULL) {
    if (gtk_toggle_button_get_active((GtkToggleButton *)starNameCheckButton2)) {
      optionsModified = TRUE;
      showNames2 = TRUE;
    } else {
      optionsModified = TRUE;
      showNames2 = FALSE;
    }
    if (labelMode)
      showNames = showNames2;
  } else if (strstr("showMeteors1", userData) != NULL) {
    if (gtk_toggle_button_get_active((GtkToggleButton *)meteorCheckButton1)) {
      optionsModified = TRUE;
      showMeteors1 = TRUE;
    } else {
      optionsModified = TRUE;
      showMeteors1 = FALSE;
    }
    if (!labelMode)
      showMeteors = showMeteors1;
  } else if (strstr("showMeteors2", userData) != NULL) {
    if (gtk_toggle_button_get_active((GtkToggleButton *)meteorCheckButton2)) {
      optionsModified = TRUE;
      showMeteors2 = TRUE;
    } else {
      optionsModified = TRUE;
      showMeteors2 = FALSE;
    }
    if (labelMode)
      showMeteors = showMeteors2;
  } else if (strstr("showPlanets1", userData) != NULL) {
    if (gtk_toggle_button_get_active((GtkToggleButton *)planetCheckButton1)) {
      optionsModified = TRUE;
      showPlanets1 = TRUE;
    } else {
      optionsModified = TRUE;
      showPlanets1 = FALSE;
    }
    if (!labelMode)
      showPlanets = showPlanets1;
  } else if (strstr("showPlanets2", userData) != NULL) {
    if (gtk_toggle_button_get_active((GtkToggleButton *)planetCheckButton2)) {
      optionsModified = TRUE;
      showPlanets2 = TRUE;
    } else {
      optionsModified = TRUE;
      showPlanets2 = FALSE;
    }
    if (labelMode)
      showPlanets = showPlanets2;
  } else if (strstr("showGreatCircles1", userData) != NULL) {
    if (gtk_toggle_button_get_active((GtkToggleButton *)greatCirclesCheckButton1)) {
      optionsModified = TRUE;
      showGreatCircles1 = TRUE;
    } else {
      optionsModified = TRUE;
      showGreatCircles1 = FALSE;
    }
    if (!labelMode)
      showGreatCircles = showGreatCircles1;
  } else if (strstr("showGreatCircles2", userData) != NULL) {
    if (gtk_toggle_button_get_active((GtkToggleButton *)greatCirclesCheckButton2)) {
      optionsModified = TRUE;
      showGreatCircles2 = TRUE;
    } else {
      optionsModified = TRUE;
      showGreatCircles2 = FALSE;
    }
    if (labelMode)
      showGreatCircles = showGreatCircles2;
  } else if (strstr("useAsterisms", userData) != NULL) {
    if (gtk_toggle_button_get_active((GtkToggleButton *)useAsterismsCheckButton)) {
      optionsModified = TRUE;
      useAsterisms = TRUE;
    } else {
      optionsModified = TRUE;
      useAsterisms = FALSE;
    }
  } else if (strstr("showDeepSky1", userData) != NULL) {
    if (gtk_toggle_button_get_active((GtkToggleButton *)deepSkyCheckButton1)) {
      optionsModified = TRUE;
      showDeepSky1 = TRUE;
    } else {
      optionsModified = TRUE;
      showDeepSky1 = FALSE;
    }
    if (!labelMode)
      showDeepSky = showDeepSky1;
  } else if (strstr("showDeepSky2", userData) != NULL) {
    if (gtk_toggle_button_get_active((GtkToggleButton *)deepSkyCheckButton2)) {
      optionsModified = TRUE;
      showDeepSky2 = TRUE;
    } else {
      optionsModified = TRUE;
      showDeepSky2 = FALSE;
    }
    if (labelMode)
      showDeepSky = showDeepSky2;
  } else if (strstr("gPSButton", userData) != NULL) {
    if (gtk_toggle_button_get_active((GtkToggleButton *)gPSButton)) {
      optionsModified = TRUE;
      useGPSD = TRUE;
      useCustomLocation = FALSE;
      useMenuLocation = FALSE;
    } else {
      optionsModified = TRUE;
      useGPSD = FALSE;
    }
    setLocationSensitivities();
  } else if (strstr("customLocationButton", userData) != NULL) {
    if (gtk_toggle_button_get_active((GtkToggleButton *)customLocationButton)) {
      optionsModified = TRUE;
      useCustomLocation = TRUE;
      useMenuLocation = FALSE;
      useGPSD = FALSE;
    } else {
      optionsModified = TRUE;
      useCustomLocation = FALSE;
    }
    setLocationSensitivities();
  } else if (strstr("currentTimeButton", userData) != NULL) {
    optionsModified = TRUE;
    if (gtk_toggle_button_get_active((GtkToggleButton *)currentTimeButton)) {
      useCurrentTime = TRUE;
      useCalendarTime = FALSE;
      useJulianTime = FALSE;
    } else {
      useCurrentTime = FALSE;
    }
    setTimeSensitivities();
  } else if (strstr("calendarTimeButton", userData) != NULL) {
    optionsModified = TRUE;
    if (gtk_toggle_button_get_active((GtkToggleButton *)calendarTimeButton)) {
      useCurrentTime = FALSE;
      useCalendarTime = TRUE;
      useJulianTime = FALSE;
    } else {
      useCalendarTime = FALSE;
    }
    setTimeSensitivities();
  } else if (strstr("julianTimeButton", userData) != NULL) {
    optionsModified = TRUE;
    if (gtk_toggle_button_get_active((GtkToggleButton *)julianTimeButton)) {
      useCurrentTime = FALSE;
      useCalendarTime = FALSE;
      useJulianTime = TRUE;
    } else {
      useJulianTime = FALSE;
    }
    setTimeSensitivities();
  } else if (strstr("menuLocationButton", userData) != NULL) {
    if (gtk_toggle_button_get_active((GtkToggleButton *)menuLocationButton)) {
      optionsModified = TRUE;
      useCustomLocation = FALSE;
      useMenuLocation = TRUE;
      useGPSD = FALSE;
    } else {
      optionsModified = TRUE;
      useMenuLocation = FALSE;
    }
    setLocationSensitivities();
  } else
    fprintf(stderr, "checkButtonCallback passed unknown string \"%s\"\n", (char *)userData);
}

void itemsButtonClicked(GtkButton *button, gpointer userData)
{
  static int firstCall = TRUE;
  static GtkWidget *column1Label, *column2Label, *ok,
    *magSpinLabel1, *magSpinLabel2;
  static GtkObject *magAdjustment1, *magAdjustment2;

  if (firstCall) {
    int row = 0;

    itemsTable = gtk_table_new(16, 2, FALSE);
    column1Label = gtk_label_new("Screen 1");
    gtk_table_attach(GTK_TABLE(itemsTable), column1Label, 0, 1, row, row+1,
		     GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 0);
    gtk_widget_show(column1Label);

    column2Label = gtk_label_new("Screen 2");
    gtk_table_attach(GTK_TABLE(itemsTable), column2Label, 1, 2, row, row+1,
		     GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 0);
    gtk_widget_show(column2Label); row++;

    magSpinLabel1 = gtk_label_new("Mag. Limit");
    gtk_table_attach(GTK_TABLE(itemsTable), magSpinLabel1, 0, 1, row, row+1,
		     GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 0);
    gtk_widget_show((GtkWidget *)magSpinLabel1);
    magSpinLabel2 = gtk_label_new("Mag. Limit");
    gtk_table_attach(GTK_TABLE(itemsTable), magSpinLabel2, 1, 2, row, row+1,
		     GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 0);
    gtk_widget_show((GtkWidget *)magSpinLabel2);row++;

    if (haveFaintStars)
      magAdjustment1 = gtk_adjustment_new(limitingMagnitude1, -1.5,
					  9.0, 0.01, 0.1, 0.0);
    else
      magAdjustment1 = gtk_adjustment_new(limitingMagnitude1, -1.5,
					  faintestStar, 0.01, 0.1, 0.0);
    magSpin1 = gtk_spin_button_new((GtkAdjustment *)magAdjustment1, 0.5, 2);
    gtk_spin_button_set_numeric((GtkSpinButton *)magSpin1, TRUE);
    gtk_table_attach(GTK_TABLE(itemsTable), (GtkWidget *)magSpin1, 0, 1, row, row+1,
		     GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 0);
    gtk_widget_show((GtkWidget *)magSpin1);

    magAdjustment2 = gtk_adjustment_new(limitingMagnitude2, -1.5, VISUAL_LIMIT, 0.01, 0.1, 0.0);
    magSpin2 = gtk_spin_button_new((GtkAdjustment *)magAdjustment2, 0.5, 2);
    gtk_spin_button_set_numeric((GtkSpinButton *)magSpin2, TRUE);
    gtk_table_attach(GTK_TABLE(itemsTable), (GtkWidget *)magSpin2, 1, 2, row, row+1,
		     GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 0);
    gtk_widget_show((GtkWidget *)magSpin2); row++;

    starCheckButton1 = (GtkCheckButton *)gtk_check_button_new_with_label ("Display Stars");
    gtk_table_attach(GTK_TABLE(itemsTable), (GtkWidget *)starCheckButton1, 0, 1, row, row+1,
		     GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 0);
    gtk_widget_show((GtkWidget *)starCheckButton1);
    gtk_toggle_button_set_active((GtkToggleButton *)starCheckButton1, showStars1);
    g_signal_connect (G_OBJECT(starCheckButton1), "clicked",
                      G_CALLBACK(checkButtonCallback), "showStars1");

    starCheckButton2 = (GtkCheckButton *)gtk_check_button_new_with_label ("Display Stars");
    gtk_table_attach(GTK_TABLE(itemsTable), (GtkWidget *)starCheckButton2, 1, 2, row, row+1,
		     GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 0);
    gtk_widget_show((GtkWidget *)starCheckButton2); row++;
    gtk_toggle_button_set_active((GtkToggleButton *)starCheckButton2, showStars2);
    g_signal_connect (G_OBJECT(starCheckButton2), "clicked",
                      G_CALLBACK(checkButtonCallback), "showStars2");

    starNameCheckButton1 = (GtkCheckButton *)gtk_check_button_new_with_label ("Display Names");
    gtk_table_attach(GTK_TABLE(itemsTable), (GtkWidget *)starNameCheckButton1, 0, 1, row, row+1,
		     GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 0);
    gtk_widget_show((GtkWidget *)starNameCheckButton1);
    gtk_toggle_button_set_active((GtkToggleButton *)starNameCheckButton1, showNames1);
    g_signal_connect (G_OBJECT(starNameCheckButton1), "clicked",
                      G_CALLBACK(checkButtonCallback), "showNames1");

    starNameCheckButton2 = (GtkCheckButton *)gtk_check_button_new_with_label ("Display Names");
    gtk_table_attach(GTK_TABLE(itemsTable), (GtkWidget *)starNameCheckButton2, 1, 2, row, row+1,
		     GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 0);
    gtk_widget_show((GtkWidget *)starNameCheckButton2); row++;
    gtk_toggle_button_set_active((GtkToggleButton *)starNameCheckButton2, showNames2);
    g_signal_connect (G_OBJECT(starNameCheckButton2), "clicked",
                      G_CALLBACK(checkButtonCallback), "showNames2");

    planetCheckButton1 = (GtkCheckButton *)gtk_check_button_new_with_label ("Display Planets");
    gtk_table_attach(GTK_TABLE(itemsTable), (GtkWidget *)planetCheckButton1, 0, 1, row, row+1,
		     GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 0);
    gtk_widget_show((GtkWidget *)planetCheckButton1);
    gtk_toggle_button_set_active((GtkToggleButton *)planetCheckButton1, showPlanets1);
    g_signal_connect (G_OBJECT(planetCheckButton1), "clicked",
                      G_CALLBACK(checkButtonCallback), "showPlanets1");

    planetCheckButton2 = (GtkCheckButton *)gtk_check_button_new_with_label ("Display Planets");
    gtk_table_attach(GTK_TABLE(itemsTable), (GtkWidget *)planetCheckButton2, 1, 2, row, row+1,
		     GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 0);
    gtk_widget_show((GtkWidget *)planetCheckButton2); row++;
    gtk_toggle_button_set_active((GtkToggleButton *)planetCheckButton2, showPlanets2);
    g_signal_connect (G_OBJECT(planetCheckButton2), "clicked",
                      G_CALLBACK(checkButtonCallback), "showPlanets2");

    meteorCheckButton1 = (GtkCheckButton *)gtk_check_button_new_with_label ("Meteor Radiants");
    gtk_table_attach(GTK_TABLE(itemsTable), (GtkWidget *)meteorCheckButton1, 0, 1, row, row+1,
		     GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 0);
    gtk_widget_show((GtkWidget *)meteorCheckButton1);
    gtk_toggle_button_set_active((GtkToggleButton *)meteorCheckButton1, showMeteors1);
    g_signal_connect (G_OBJECT(meteorCheckButton1), "clicked",
                      G_CALLBACK(checkButtonCallback), "showMeteors1");

    meteorCheckButton2 = (GtkCheckButton *)gtk_check_button_new_with_label ("Meteor Radiants");
    gtk_table_attach(GTK_TABLE(itemsTable), (GtkWidget *)meteorCheckButton2, 1, 2, row, row+1,
		     GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 0);
    gtk_widget_show((GtkWidget *)meteorCheckButton2); row++;
    gtk_toggle_button_set_active((GtkToggleButton *)meteorCheckButton2, showMeteors2);
    g_signal_connect (G_OBJECT(meteorCheckButton2), "clicked",
                      G_CALLBACK(checkButtonCallback), "showMeteors2");

    deepSkyCheckButton1 = (GtkCheckButton *)gtk_check_button_new_with_label ("Deep Sky Obs");
    gtk_table_attach(GTK_TABLE(itemsTable), (GtkWidget *)deepSkyCheckButton1, 0, 1, row, row+1,
		     GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 0);
    gtk_widget_show((GtkWidget *)deepSkyCheckButton1);
    gtk_toggle_button_set_active((GtkToggleButton *)deepSkyCheckButton1, showDeepSky1);
    g_signal_connect (G_OBJECT(deepSkyCheckButton1), "clicked",
                      G_CALLBACK(checkButtonCallback), "showDeepSky1");

    deepSkyCheckButton2 = (GtkCheckButton *)gtk_check_button_new_with_label ("Deep Sky Obs");
    gtk_table_attach(GTK_TABLE(itemsTable), (GtkWidget *)deepSkyCheckButton2, 1, 2, row, row+1,
		     GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 0);
    gtk_widget_show((GtkWidget *)deepSkyCheckButton2); row++;
    gtk_toggle_button_set_active((GtkToggleButton *)deepSkyCheckButton2, showDeepSky2);
    g_signal_connect (G_OBJECT(deepSkyCheckButton2), "clicked",
                      G_CALLBACK(checkButtonCallback), "showDeepSky2");

    greatCirclesCheckButton1 = (GtkCheckButton *)gtk_check_button_new_with_label ("Great Circles");
    gtk_table_attach(GTK_TABLE(itemsTable), (GtkWidget *)greatCirclesCheckButton1, 0, 1, row, row+1,
		     GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 0);
    gtk_widget_show((GtkWidget *)greatCirclesCheckButton1);
    gtk_toggle_button_set_active((GtkToggleButton *)greatCirclesCheckButton1, showGreatCircles1);
    g_signal_connect (G_OBJECT(greatCirclesCheckButton1), "clicked",
                      G_CALLBACK(checkButtonCallback), "showGreatCircles1");

    greatCirclesCheckButton2 = (GtkCheckButton *)gtk_check_button_new_with_label ("Great Circles");
    gtk_table_attach(GTK_TABLE(itemsTable), (GtkWidget *)greatCirclesCheckButton2, 1, 2, row, row+1,
		     GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 0);
    gtk_widget_show((GtkWidget *)greatCirclesCheckButton2); row++;
    gtk_toggle_button_set_active((GtkToggleButton *)greatCirclesCheckButton2, showGreatCircles2);
    g_signal_connect (G_OBJECT(greatCirclesCheckButton2), "clicked",
                      G_CALLBACK(checkButtonCallback), "showGreatCircles2");

    useAsterismsCheckButton = (GtkCheckButton *)gtk_check_button_new_with_label ("Use Asterisms");
    gtk_table_attach(GTK_TABLE(itemsTable), (GtkWidget *)useAsterismsCheckButton, 1, 2, row, row+1,
		     GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 0);
    gtk_widget_show((GtkWidget *)useAsterismsCheckButton); row++;
    gtk_toggle_button_set_active((GtkToggleButton *)useAsterismsCheckButton, useAsterisms);
    g_signal_connect (G_OBJECT(useAsterismsCheckButton), "clicked",
                      G_CALLBACK(checkButtonCallback), "useAsterisms");

    ok = gtk_button_new_with_label ("OK");
    gtk_table_attach(GTK_TABLE(itemsTable), ok, 0, 2, row, row+1,
		     GTK_FILL, GTK_FILL, 0, 0);
    g_signal_connect (G_OBJECT(ok), "clicked",
                      G_CALLBACK (itemsOk), NULL);
    gtk_widget_show(ok); row++;

    g_object_ref(itemsTable);
    firstCall = FALSE;
  }
  gtk_widget_hide(mainBox);
  gtk_container_remove(GTK_CONTAINER(window), mainBox);
  gtk_container_add(GTK_CONTAINER(window), itemsTable);
  gtk_widget_show(itemsTable);
  skyIsVisible = FALSE;
}

void removeLocationBox(void)
{
  gtk_widget_hide(locationTable);
  gtk_container_remove(GTK_CONTAINER(window), locationTable);
  gtk_container_add(GTK_CONTAINER(window), mainBox);
  gtk_widget_show(mainBox);
  if (optionsModified) {
    dprintf("Doing redraw because checkBoxes modified\n");
    fullRedraw(FALSE);
    optionsModified = FALSE;
  }
  skyIsVisible = TRUE;
}

void removeOptionsBox(void)
{
  gtk_widget_hide(optionsTable);
  gtk_container_remove(GTK_CONTAINER(window), optionsTable);
  gtk_container_add(GTK_CONTAINER(window), mainBox);
  gtk_widget_show(mainBox);
  if (optionsModified) {
    dprintf("Doing redraw because checkBoxes modified\n");
    fullRedraw(FALSE);
    optionsModified = FALSE;
  }
  skyIsVisible = TRUE;
}

GtkWidget *timeTable, *tok;

void removeTimeBox(void)
{
  gtk_widget_hide(timeTable);
  gtk_container_remove(GTK_CONTAINER(window), timeTable);
  gtk_container_add(GTK_CONTAINER(window), mainBox);
  gtk_widget_show(mainBox);
  if (optionsModified) {
    dprintf("Doing redraw because checkBoxes modified\n");
    fullRedraw(FALSE);
    optionsModified = FALSE;
  }
  skyIsVisible = TRUE;
}

void checkLocationSpinBoxes(void)
{
  double tempValue, DD, MM, SS;

  if (useCustomLocation) {
    DD = gtk_spin_button_get_value((GtkSpinButton *)latSpinDD);
    MM = gtk_spin_button_get_value((GtkSpinButton *)latSpinMM);
    SS = gtk_spin_button_get_value((GtkSpinButton *)latSpinSS);
    if (DD < 0.0) {
      MM *= -1.0;
      SS *= -1.0;
    }
    tempValue = (DD + MM/60.0 + SS/3600.0) * DEGREES_TO_RADIANS;
    if (fabs(latitude - tempValue) > 1.0/3600.0) {
      latitude = tempValue;
      locationChanged = TRUE;
      gPSLocation = FALSE;
    optionsModified = TRUE;
    }
    DD = gtk_spin_button_get_value((GtkSpinButton *)longSpinDD);
    MM = gtk_spin_button_get_value((GtkSpinButton *)longSpinMM);
    SS = gtk_spin_button_get_value((GtkSpinButton *)longSpinSS);
    if (DD < 0.0) {
      MM *= -1.0;
      SS *= -1.0;
    }
    tempValue = (DD + MM/60.0 + SS/3600.0) * DEGREES_TO_RADIANS;
    if (fabs(longitude - tempValue) > 1.0/3600.0) {
      longitude = tempValue;
      locationChanged = TRUE;
      gPSLocation = FALSE;
      optionsModified = TRUE;
    }
    if (locationChanged)
      dprintf("Location Changed, lat = %f, long = %f\n",
	     latitude/DEGREES_TO_RADIANS, longitude/DEGREES_TO_RADIANS);
  }
}

void timeSave(GtkButton *button, gpointer userData)
{
  checkTimeSpinBoxes();
  writeConfigFile();
  removeTimeBox();
}

/*
  This routine marks the Private menu as never having been created.
  This routine is called by the routine which sves private catalog entries,
  so that the next time the private catalog is requested, it will be
  recreated with the newly added entry.   If the Private menu
  has already been created, THIS IS A MEMORY LEAK,
  because I do not free up the resources allocated when the Private
  memu is created.   The problem is that it's not merely a matter of
  freeing the memory previously malloced.   There are gtk resources
  which must be freed if the menu is to be destroyed cleanly.   If
  I free the malloc'd memory, and do not clean up the gtk resources, the
  program crashes.   I do not know how to clean up the gtk resources,
  so I leave to old memu resources allocated, but inaccessible.
*/
void destroyPrivateMenu(void) {
  int found = FALSE;
  locationClass *ptr;

  if (privateMenuShowing) {
    gtk_widget_hide(activeMenu);
    privateMenuShowing = menuShowing = FALSE;
  }
  ptr = locationClassRoot;
  while ((!found) && (ptr != NULL)) {
    if ((strcmp(ptr->name, "Private") == 0) && ptr->menuCreated) {
      ptr->menuCreated = FALSE;
      found = TRUE;
    } else
      ptr = ptr->next;
  }
}

/*
  This is the callback routine for the button that saves a new
  location in the private locations menu.   I can't just append
  the new position to the existing file, because if I do so, we
  could end up with multiple entries with the same name, if the
  user either clicks the button more than once, or tries to
  enter a new position with the same name.   So I will copy
  the old file to a new one, and replace any duplicated entry.
*/
void savePrivateLocation(GtkButton *button, gpointer userData)
{
  int i, nRead;
  int haveWrittenNewEntry = FALSE;
  double tLat, tLong;
  char oldFileName[MAX_FILE_NAME_SIZE], newFileName[MAX_FILE_NAME_SIZE],
    nameCopy[100], tName[100];
  FILE *oldMenuFile, *newMenuFile;

  checkLocationSpinBoxes();
  for (i = 0; i < strlen(locationName); i++)
    if (locationName[i] == ' ')
      nameCopy[i] ='_';
    else
      nameCopy[i] = locationName[i];
  nameCopy[strlen(locationName)] = (char)0;
  sprintf(oldFileName, "%s/menus/Private", dataDir);
  oldMenuFile = fopen(oldFileName, "r");
  if (oldMenuFile == NULL) {
    perror("Error opening old Private menu");
    exit(ERROR_EXIT);
  }
  sprintf(newFileName, "%s/menus/newPrivate", dataDir);
  newMenuFile = fopen(newFileName, "w");
  if (newMenuFile == NULL) {
    perror("Error opening new Private menu");
    exit(ERROR_EXIT);
  }
  while (!feof(oldMenuFile)) {
    nRead = fscanf(oldMenuFile, "%s %lf %lf", &tName[0], &tLat, &tLong);
    if (nRead == 3) {
      if (strcmp(tName, nameCopy) == 0) {
	fprintf(newMenuFile, "%s %7.3f %7.3f\n", nameCopy,
		latitude/DEGREES_TO_RADIANS, longitude/DEGREES_TO_RADIANS);
	haveWrittenNewEntry = TRUE;
      } else
	fprintf(newMenuFile, "%s %7.3f %7.3f\n", tName, tLat, tLong);
    }
  }
  if (!haveWrittenNewEntry)
    fprintf(newMenuFile, "%s %7.3f %7.3f\n", nameCopy,
	    latitude/DEGREES_TO_RADIANS, longitude/DEGREES_TO_RADIANS);
  fclose(oldMenuFile);
  fclose(newMenuFile);
  rename(newFileName, oldFileName);
  destroyPrivateMenu();
}

/*
  Save any new location options in the config file, and then
  return to the sky display.
*/
void locationSave(GtkButton *button, gpointer userData)
{
  checkLocationSpinBoxes();
  writeConfigFile();
  removeLocationBox();
}

/*
  Save any new "options" options in the config file, and then
  return to the sky display.
*/
void optionsSave(GtkButton *button, gpointer userData)
{
  writeConfigFile();
  removeOptionsBox();
}

void timeOk(GtkButton *button, gpointer userData)
{
  checkTimeSpinBoxes();
  checkLocationSpinBoxes();
  writeConfigFile();
  removeTimeBox();
}

/*
  Call back function for the location page OK button, which
  redraws the sky without saving chosen options.
*/
void locationOk(GtkButton *button, gpointer userData)
{
  checkLocationSpinBoxes();
  writeConfigFile();
  removeLocationBox();
}

/*
  Call back function for the options page OK button, which
  redraws the sky without saving chosen options.
*/
void optionsOk(GtkButton *button, gpointer userData)
{
  writeConfigFile();
  removeOptionsBox();
}

/*
  Call back function for the options page about button, which
  puts up information about the program.
*/
void about(GtkButton *button, gpointer userData)
{
  removeOptionsBox();
  gtk_widget_hide(controlButtonBox);
  aboutScreen = ABOUT_SCREEN;
  displayingAnOptsPage = TRUE;
}

/*
  Call back function for the options page soluni button, which
  puts up sun and moon information.
*/
void soluni(GtkButton *button, gpointer userData)
{
  removeOptionsBox();
  gtk_widget_hide(controlButtonBox);
  aboutScreen = SOLUNI_SCREEN;
  displayingAnOptsPage = TRUE;
}

/*
  Call back function for the options page Big Moon Calendar button, which
  draws a moon calendar for 21 months
*/
void bigMooncal(GtkButton *button, gpointer userData)
{
  removeOptionsBox();
  gtk_widget_hide(controlButtonBox);
  aboutScreen = BIG_MOONCAL_SCREEN;
  displayingAnOptsPage = TRUE;
}

/*
  Call back function for the options page Small Moon Calendar button, which
  draws a moon calendar for 1 month
*/
void smallMooncal(GtkButton *button, gpointer userData)
{
  removeOptionsBox();
  gtk_widget_hide(controlButtonBox);
  aboutScreen = SMALL_MOONCAL_SCREEN;
  displayingAnOptsPage = TRUE;
}

/*
  Call back function for the options page Planet Compass button, which
  shows the location of all planets.
*/
void planetCompass(GtkButton *button, gpointer userData)
{
  removeOptionsBox();
  gtk_widget_hide(controlButtonBox);
  aboutScreen = PLANETCOMPASS_SCREEN;
  displayingAnOptsPage = TRUE;
}

/*
  Call back function for the options page Meteor Showers button, which
  shows a table of this year's meteor showers.
*/
void meteorShowersScreen(GtkButton *button, gpointer userData)
{
  removeOptionsBox();
  gtk_widget_hide(controlButtonBox);
  aboutScreen = METEOR_SHOWERS_SCREEN;
  displayingAnOptsPage = TRUE;
}

/*
  Call back function for the options page Solar System button, which
  shows the location of all planets.
*/
void solarSystem(GtkButton *button, gpointer userData)
{
  int arg;

  arg = *((int *)userData);
  removeOptionsBox();
  gtk_widget_hide(controlButtonBox);
  if (arg == 1)
    aboutScreen = SOLAR_SYSTEM_SCHEMATIC_SCREEN;
  else
    aboutScreen = SOLAR_SYSTEM_SCALE_SCREEN;
 if (useCurrentTime)
   savedTimeMode = 0;
 else if (useCalendarTime)
   savedTimeMode = 1;
 else
   savedTimeMode = 2;
 savedTJD = tJD;
 displayingAnOptsPage = TRUE;
}

/*
  Call back function for the options page flashlight button, which
  sets the entire screen to one color.
*/
void flashlight(int color)
{
  if (color == 0)
    flashlightGC = whiteGC;
  else
    flashlightGC = redGC;
  removeOptionsBox();
  fullscreenStateChanging = TRUE;
  gtk_widget_hide(controlButtonBox);
  gtk_window_fullscreen(GTK_WINDOW(window));
  inFlashlightMode = TRUE;
}

void redFlashlight(GtkButton *button, gpointer userData)
{
  flashlight(1);
}

void whiteFlashlight(GtkButton *button, gpointer userData)
{
  flashlight(0);
}

void locationNameCallback(GtkButton *button, gpointer userData)
{
  char *newName;

  newName = (char *)gtk_entry_get_text((GtkEntry *)locationNameText);
  if (strcmp(locationName, newName) != 0) {
    optionsModified = TRUE;
    strcpy(locationName, newName);
  }
}

void locationCallback(gpointer callbackData, guint callbackAction, GtkWidget *widget)
{
  if (locationNameIsDisplayed) {
    gtk_widget_hide(displayedLocationName);
    gtk_table_remove((GtkContainer *)locationTable, displayedLocationName);
    /*
    gtk_widget_destroy(displayedLocationName);
    */
  }
  strcpy(locationName, ((locationEntry *)callbackData)->name);
  latitude = ((locationEntry *)callbackData)->latitude;
  longitude = ((locationEntry *)callbackData)->longitude;
  displayedLocationName = gtk_label_new(locationName);
  gtk_table_attach(GTK_TABLE(locationTable), displayedLocationName, 2, 3, 4, 5,
		   GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 0);
  gtk_widget_show(displayedLocationName);
  dprintf("New location \"%s\": %f %f\n", locationName, latitude/DEGREES_TO_RADIANS,
	 longitude/DEGREES_TO_RADIANS);
  locationNameIsDisplayed = TRUE;
  locationChanged = TRUE;
  newPosition();
}

void optionsCallback(gpointer callbackData, guint callbackAction, GtkWidget *widget)
{
  printf("In optionsCallback\n");
}

void categoryCallback(gpointer callbackData, guint callbackAction, GtkWidget *widget)
{
  int found = FALSE;
  locationClass *ptr;

  if (menuShowing) {
    gtk_widget_hide(activeMenu);
    menuShowing = privateMenuShowing = FALSE;
  }
  ptr = locationClassRoot;
  while ((!found) && (ptr != NULL)) {
    if (ptr->key == callbackAction)
      found = TRUE;
    else
      ptr = ptr->next;
  }
  if (found) {
    int dataFD, i;
    int nLocations = 0;
    int eOF = FALSE;
    char *fileName;
    char inLine[100];
    locationEntry *lPtr;
    locationEntry *listEnd = NULL;

    dprintf("Must display menu for \"%s\"\n", ptr->name);
    if (!(ptr->menuCreated)) {
      dprintf("The menu must be created\n");
      ptr->entry = NULL;
      fileName = malloc(strlen(dataDir)+strlen("/menus/")+strlen(ptr->name)+10);
      if (fileName == NULL) {
	perror("file name for menu data");
	exit(ERROR_EXIT);
      }
      sprintf(fileName, "%s/menus/%s", dataDir, ptr->name);
      dataFD = open(fileName, O_RDONLY);
      free(fileName);
      if (dataFD < 0) {
	perror("open of menu a file");
	exit(ERROR_EXIT);
      }
      while (!eOF) {
	getLine(dataFD, &inLine[0], &eOF);
	if (!eOF) {
	  int nRead, i;
	  float latitude, longitude;
	  char tempName[100];
	  
	  nRead = sscanf(inLine, "%s %f %f", &tempName[0], &latitude, &longitude);
	  if (nRead == 3) {
	    locationEntry *newLoc;
	    
	    for (i = 0; i < strlen(tempName); i++)
	      if (tempName[i] == '_')
		tempName[i] = ' ';
	    newLoc = (locationEntry *)malloc(sizeof(locationEntry));
	    if (newLoc == NULL) {
	      perror("newLoc");
	      exit(ERROR_EXIT);
	    }
	    newLoc->next = NULL;
	    newLoc->name = malloc(strlen(tempName)+1);
	    if (newLoc->name == NULL) {
	      perror("newLoc->name");
	      exit(ERROR_EXIT);
	    }
	    strcpy(newLoc->name, tempName);
	    newLoc->latitude = latitude*DEGREES_TO_RADIANS;
	    newLoc->longitude = longitude*DEGREES_TO_RADIANS;
	    if (ptr->entry == NULL)
	      ptr->entry = newLoc;
	    else
	      listEnd->next = newLoc;
	    listEnd = newLoc;
	    nLocations++;
	  }
	}
      }
      close(dataFD);
      ptr->nLocations = nLocations;
      ptr->itemFactoryEntry = (GtkItemFactoryEntry *)malloc((1+nLocations)*
							    sizeof(GtkItemFactoryEntry));
      if (ptr->itemFactoryEntry == NULL) {
	perror("ptr->itemFactoryEntry");
	exit(ERROR_EXIT);
      }
      ptr->itemFactoryEntry[0].path = (gchar *)malloc(strlen(ptr->name)+2);
      if (ptr->itemFactoryEntry[0].path == NULL) {
	perror("ptr->itemFactoryEntry[0].path");
	exit(ERROR_EXIT);
      }
      sprintf(ptr->itemFactoryEntry[0].path, "/%s", ptr->name);
      ptr->itemFactoryEntry[0].accelerator = NULL;
      ptr->itemFactoryEntry[0].callback = NULL;
      ptr->itemFactoryEntry[0].callback_action = 0;
      ptr->itemFactoryEntry[0].item_type = (gchar *)malloc(strlen("<Branch>")+1);
      if (ptr->itemFactoryEntry[0].item_type == NULL) {
	perror("ptr->itemFactoryEntry[0].item_type");
	exit(ERROR_EXIT);
      }
      strcpy(ptr->itemFactoryEntry[0].item_type, "<Branch>");
      lPtr = ptr->entry;
      ptr->accelGroup = gtk_accel_group_new();
      sprintf(inLine, "<Orrery%s>", ptr->name);
      ptr->itemFactory = gtk_item_factory_new(GTK_TYPE_MENU_BAR, inLine,
					      ptr->accelGroup);

      gtk_item_factory_create_item (ptr->itemFactory,
                                    &ptr->itemFactoryEntry[0],
                                    lPtr, 1);
      for (i = 1; i <= nLocations; i++) {
	char path[100];
	
	sprintf(path, "/%s/%s", ptr->name, lPtr->name);
	ptr->itemFactoryEntry[i].path = (gchar *)malloc(strlen(path)+1);
	if (ptr->itemFactoryEntry[i].path == NULL) {
	  perror("ptr->itemFactoryEntry[i].path");
	  exit(ERROR_EXIT);
	}
	strcpy(ptr->itemFactoryEntry[i].path, path);
	ptr->itemFactoryEntry[i].accelerator = NULL;
	ptr->itemFactoryEntry[i].callback = locationCallback;
	ptr->itemFactoryEntry[i].item_type = NULL;
        gtk_item_factory_create_item (ptr->itemFactory,
                                      &ptr->itemFactoryEntry[i],
                                      lPtr, 1);
	lPtr = lPtr->next;
      }
      ptr->menu = gtk_item_factory_get_widget(ptr->itemFactory, inLine);
      ptr->menuCreated = TRUE;
      gtk_table_attach(GTK_TABLE(locationTable), (GtkWidget *)ptr->menu, 1, 2, 4, 5,
		       GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 0);
      gtk_widget_set_sensitive(ptr->menu, TRUE);
    }
    gtk_widget_show((GtkWidget *)ptr->menu);
    activeMenu = ptr->menu;
    menuShowing = TRUE;
    if (strcmp(ptr->name, "Private") == 0)
      privateMenuShowing = TRUE;
  }
}

/*
  Sets up the page allowing the observation time to be specified.
*/
void timeButtonClicked(GtkButton *button, gpointer userData)
{
  static int firstCall = TRUE;
  static GtkObject *yearAdj, *monthAdj, *dayAdj, *hourAdj, *minuteAdj, *secondAdj;
  static GtkObject *julAdjustment;

  if (firstCall) {
    timeTable = gtk_table_new(11, 4, FALSE);
    
    tok = gtk_button_new_with_label ("OK");
    gtk_table_attach(GTK_TABLE(timeTable), tok, 0, 4, 10, 11,
		     GTK_FILL, GTK_FILL, 0, 0);
    g_signal_connect (G_OBJECT(tok), "clicked",
                      G_CALLBACK (timeOk), NULL);
    gtk_widget_show(tok);

    currentTimeButton = (GtkCheckButton *)gtk_check_button_new_with_label ("Now");
    gtk_table_attach(GTK_TABLE(timeTable), (GtkWidget *)currentTimeButton, 0, 1, 0, 1,
		     GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 0);
    gtk_widget_show((GtkWidget *)currentTimeButton);
    gtk_toggle_button_set_active((GtkToggleButton *)currentTimeButton, useCurrentTime);
    g_signal_connect (G_OBJECT(currentTimeButton), "clicked",
                      G_CALLBACK(checkButtonCallback), "currentTimeButton");

    calendarTimeButton = (GtkCheckButton *)gtk_check_button_new_with_label ("Calendar");
    gtk_table_attach(GTK_TABLE(timeTable), (GtkWidget *)calendarTimeButton, 1, 2, 0, 1,
		     GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 0);
    gtk_widget_show((GtkWidget *)calendarTimeButton);
    gtk_toggle_button_set_active((GtkToggleButton *)calendarTimeButton, useCalendarTime);
    g_signal_connect (G_OBJECT(calendarTimeButton), "clicked",
                      G_CALLBACK(checkButtonCallback), "calendarTimeButton");

    julianTimeButton = (GtkCheckButton *)gtk_check_button_new_with_label ("Jul. Date");
    gtk_table_attach(GTK_TABLE(timeTable), (GtkWidget *)julianTimeButton, 0, 1, 1, 2,
		     GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 0);
    gtk_widget_show((GtkWidget *)julianTimeButton);
    gtk_toggle_button_set_active((GtkToggleButton *)julianTimeButton, useJulianTime);
    g_signal_connect (G_OBJECT(julianTimeButton), "clicked",
                      G_CALLBACK(checkButtonCallback), "julianTimeButton");

    cDateLabel = gtk_label_new("YYYYMMDD");
    gtk_table_attach(GTK_TABLE(timeTable), cDateLabel, 0, 1, 2, 3,
		     GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 0);
    gtk_widget_set_sensitive(cDateLabel, useCalendarTime);
    gtk_widget_show(cDateLabel);

    cTimeLabel = gtk_label_new("HHMMSS");
    gtk_table_attach(GTK_TABLE(timeTable), cTimeLabel, 0, 1, 3, 4,
		     GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 0);
    gtk_widget_set_sensitive(cTimeLabel, useCalendarTime);
    gtk_widget_show(cTimeLabel);

    yearAdj = gtk_adjustment_new(cYear, -3000.0, 3000.0, 1.0, 1.0, 0.0);
    yearSpin = gtk_spin_button_new((GtkAdjustment *)yearAdj, 0.5, 0);
    gtk_spin_button_set_numeric((GtkSpinButton *)yearSpin, TRUE);
    gtk_table_attach(GTK_TABLE(timeTable), (GtkWidget *)yearSpin, 1, 2, 2, 3,
		     GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 0);
    gtk_widget_set_sensitive(yearSpin, useCalendarTime);
    gtk_widget_show((GtkWidget *)yearSpin);

    monthAdj = gtk_adjustment_new(cMonth, 1.0, 12.0, 1.0, 1.0, 0.0);
    monthSpin = gtk_spin_button_new((GtkAdjustment *)monthAdj, 0.5, 0);
    gtk_spin_button_set_numeric((GtkSpinButton *)monthSpin, TRUE);
    gtk_table_attach(GTK_TABLE(timeTable), (GtkWidget *)monthSpin, 2, 3, 2, 3,
		     GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 0);
    gtk_widget_set_sensitive(monthSpin, useCalendarTime);
    gtk_widget_show((GtkWidget *)monthSpin);

    dayAdj = gtk_adjustment_new(cDay, 1.0, 31.0, 1.0, 1.0, 0.0);
    daySpin = gtk_spin_button_new((GtkAdjustment *)dayAdj, 0.5, 0);
    gtk_spin_button_set_numeric((GtkSpinButton *)daySpin, TRUE);
    gtk_table_attach(GTK_TABLE(timeTable), (GtkWidget *)daySpin, 3, 4, 2, 3,
		     GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 0);
    gtk_widget_set_sensitive(daySpin, useCalendarTime);
    gtk_widget_show((GtkWidget *)daySpin);

    hourAdj = gtk_adjustment_new(cHour, 0.0, 23.0, 1.0, 1.0, 0.0);
    hourSpin = gtk_spin_button_new((GtkAdjustment *)hourAdj, 0.5, 0);
    gtk_spin_button_set_numeric((GtkSpinButton *)hourSpin, TRUE);
    gtk_table_attach(GTK_TABLE(timeTable), (GtkWidget *)hourSpin, 1, 2, 3, 4,
		     GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 0);
    gtk_widget_set_sensitive(hourSpin, useCalendarTime);
    gtk_widget_show((GtkWidget *)hourSpin);

    minuteAdj = gtk_adjustment_new(cMinute, 0.0, 59.0, 1.0, 1.0, 0.0);
    minuteSpin = gtk_spin_button_new((GtkAdjustment *)minuteAdj, 0.5, 0);
    gtk_spin_button_set_numeric((GtkSpinButton *)minuteSpin, TRUE);
    gtk_table_attach(GTK_TABLE(timeTable), (GtkWidget *)minuteSpin, 2, 3, 3, 4,
		     GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 0);
    gtk_widget_set_sensitive(minuteSpin, useCalendarTime);
    gtk_widget_show((GtkWidget *)minuteSpin);

    secondAdj = gtk_adjustment_new(cSecond, 0.0, 59.0, 1.0, 1.0, 0.0);
    secondSpin = gtk_spin_button_new((GtkAdjustment *)secondAdj, 0.5, 0);
    gtk_spin_button_set_numeric((GtkSpinButton *)secondSpin, TRUE);
    gtk_table_attach(GTK_TABLE(timeTable), (GtkWidget *)secondSpin, 3, 4, 3, 4,
		     GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 0);
    gtk_widget_set_sensitive(secondSpin, useCalendarTime);
    gtk_widget_show((GtkWidget *)secondSpin);

    julSpinLabel = gtk_label_new("Julian Date");
    gtk_widget_set_sensitive(julSpinLabel, useJulianTime);
    gtk_table_attach(GTK_TABLE(timeTable), julSpinLabel, 0, 1, 4, 5,
		     GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 0);
    gtk_widget_show((GtkWidget *)julSpinLabel);
    julAdjustment = gtk_adjustment_new(tJD, 628459.0, 2817037.0, 0.001, 1.0, 0.0);
    julSpin = gtk_spin_button_new((GtkAdjustment *)julAdjustment, 0.5, 3);
    gtk_spin_button_set_numeric((GtkSpinButton *)julSpin, TRUE);
    gtk_table_attach(GTK_TABLE(timeTable), (GtkWidget *)julSpin, 1, 4, 4, 5,
		     GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 0);
    gtk_widget_set_sensitive(julSpin, useJulianTime);
    gtk_widget_show((GtkWidget *)julSpin);

    g_object_ref(timeTable);
    firstCall = FALSE;
  }
  gtk_widget_hide(mainBox);
  gtk_container_remove(GTK_CONTAINER(window), mainBox);
  gtk_container_add(GTK_CONTAINER(window), timeTable);
  gtk_widget_show(timeTable);
  skyIsVisible = FALSE;
}

/*
  Sets up the page allowing vaious global options to be specified
*/
void optionsButtonClicked(GtkButton *button, gpointer userData)
{
  static int firstCall = TRUE;
  static int one = 1;
  static int two = 2;
  static GtkWidget *ok, *whiteFlashlightButton, *redFlashlightButton, *soluniButton, *aboutButton,
    *bigMooncalButton, *smallMooncalButton, *planetCompassButton, *solarSystemSchematicButton,
    *solarSystemScaleButton, *meteorShowersButton;

  if (firstCall) {
    int row = 0;

    optionsTable = gtk_table_new(11, 2, FALSE);

    whiteFlashlightButton = gtk_button_new_with_label ("White Flashlight");
    gtk_table_attach_defaults(GTK_TABLE(optionsTable), whiteFlashlightButton,
		     0, 1, row, row+1);
    g_signal_connect (G_OBJECT(whiteFlashlightButton), "clicked",
                      G_CALLBACK (whiteFlashlight), NULL);
    gtk_widget_show(whiteFlashlightButton);
    redFlashlightButton = gtk_button_new_with_label ("Red Flashlight");
    gtk_table_attach_defaults(GTK_TABLE(optionsTable), redFlashlightButton,
		     1, 2, row, row+1);
    g_signal_connect (G_OBJECT(redFlashlightButton), "clicked",
                      G_CALLBACK (redFlashlight), NULL);
    gtk_widget_show(redFlashlightButton);
    row++;

    soluniButton = gtk_button_new_with_label ("Sun and Moon");
    gtk_table_attach_defaults(GTK_TABLE(optionsTable), soluniButton,
		     0, 1, row, row+1);
    g_signal_connect (G_OBJECT(soluniButton), "clicked",
                      G_CALLBACK (soluni), NULL);
    gtk_widget_show(soluniButton);
    aboutButton = gtk_button_new_with_label ("Symbol Key");
    gtk_table_attach_defaults(GTK_TABLE(optionsTable), aboutButton,
		     1, 2, row, row+1);
    g_signal_connect (G_OBJECT(aboutButton), "clicked",
                      G_CALLBACK (about), NULL);
    gtk_widget_show(aboutButton);

    row++;

    bigMooncalButton = gtk_button_new_with_label ("Big Moon Calendar");
    gtk_table_attach_defaults(GTK_TABLE(optionsTable), bigMooncalButton,
		     0, 1, row, row+1);
    g_signal_connect (G_OBJECT(bigMooncalButton), "clicked",
                      G_CALLBACK (bigMooncal), NULL);
    gtk_widget_show(bigMooncalButton);
    smallMooncalButton = gtk_button_new_with_label ("This Month's Moons");
    gtk_table_attach_defaults(GTK_TABLE(optionsTable), smallMooncalButton,
		     1, 2, row, row+1);
    g_signal_connect (G_OBJECT(smallMooncalButton), "clicked",
                      G_CALLBACK (smallMooncal), NULL);
    gtk_widget_show(smallMooncalButton);
    row++;

    meteorShowersButton = gtk_button_new_with_label ("Meteor Showers");
    gtk_table_attach_defaults(GTK_TABLE(optionsTable), meteorShowersButton,
		     0, 1, row, row+1);
    g_signal_connect (G_OBJECT(meteorShowersButton), "clicked",
                      G_CALLBACK (meteorShowersScreen), NULL);
    gtk_widget_show(meteorShowersButton);

    planetCompassButton = gtk_button_new_with_label ("Planet Compass");
    gtk_table_attach_defaults(GTK_TABLE(optionsTable), planetCompassButton,
		     1, 2, row, row+1);
    g_signal_connect (G_OBJECT(planetCompassButton), "clicked",
                      G_CALLBACK (planetCompass), NULL);
    gtk_widget_show(planetCompassButton);
    row++;

    solarSystemSchematicButton = gtk_button_new_with_label ("Schematic Solar Sys.");
    gtk_table_attach_defaults(GTK_TABLE(optionsTable), solarSystemSchematicButton,
		     0, 1, row, row+1);
    g_signal_connect (G_OBJECT(solarSystemSchematicButton), "clicked",
                      G_CALLBACK (solarSystem), &one);
    gtk_widget_show(solarSystemSchematicButton);

    solarSystemScaleButton = gtk_button_new_with_label ("To-Scale Solar Sys.");
    gtk_table_attach_defaults(GTK_TABLE(optionsTable), solarSystemScaleButton,
		     1, 2, row, row+1);
    g_signal_connect (G_OBJECT(solarSystemScaleButton), "clicked",
                      G_CALLBACK (solarSystem), &two);
    gtk_widget_show(solarSystemScaleButton);
    row++;

    row++;
    ok = gtk_button_new_with_label ("OK");
    gtk_table_attach_defaults(GTK_TABLE(optionsTable), ok, 0, 2, row, row+1);
    g_signal_connect (G_OBJECT(ok), "clicked",
                      G_CALLBACK (optionsOk), NULL);
    gtk_widget_show(ok);

    g_object_ref(optionsTable);
    firstCall = FALSE;
  }
  gtk_widget_hide(mainBox);
  gtk_container_remove(GTK_CONTAINER(window), mainBox);
  gtk_container_add(GTK_CONTAINER(window), optionsTable);
  gtk_widget_show(optionsTable);
  skyIsVisible = FALSE;
}

/*
  Sets up the page allowing the observer's location to be specified
*/
void locationButtonClicked(GtkButton *button, gpointer userData)
{
  static int firstCall = TRUE;
  static GtkWidget *ok;
  static GtkItemFactory *itemFactory;
  static GtkItemFactoryEntry *itemFactoryEntry;
  static GtkObject *latAdjDD, *latAdjMM, *latAdjSS, *longAdjDD, *longAdjMM, *longAdjSS;
  static GtkAccelGroup *accelGroup;

  if (firstCall) {
    double latTemp, longTemp;
    char *fileName, inLine[100];
    int menuFileFD, i;
    int nMenuClasses = 0;
    int eOF = FALSE;
    locationClass *lastClass = NULL;

    locationTable = gtk_table_new(11, 4, FALSE);

    saveAsPrivate = gtk_button_new_with_label ("Save in Private Menu");
    gtk_table_attach(GTK_TABLE(locationTable), saveAsPrivate, 0, 4, 8, 9,
		     GTK_FILL, GTK_FILL, 0, 0);
    g_signal_connect (G_OBJECT(saveAsPrivate), "clicked",
                      G_CALLBACK (savePrivateLocation), NULL);
    if (!useGPSD) {
      gtk_widget_show(saveAsPrivate);
      savePrivateShowing = TRUE;
    } else
      savePrivateShowing = FALSE;

    ok = gtk_button_new_with_label ("OK");
    gtk_table_attach(GTK_TABLE(locationTable), ok, 0, 4, 10, 11,
		     GTK_FILL, GTK_FILL, 0, 0);
    g_signal_connect (G_OBJECT(ok), "clicked",
                      G_CALLBACK (locationOk), NULL);
    gtk_widget_show(ok);

    gPSButton = (GtkCheckButton *)gtk_check_button_new_with_label ("GPS");
    gtk_table_attach(GTK_TABLE(locationTable), (GtkWidget *)gPSButton, 0, 1, 0, 1,
		     GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 0);
    gtk_widget_show((GtkWidget *)gPSButton);
    gtk_toggle_button_set_active((GtkToggleButton *)gPSButton, useGPSD);
    g_signal_connect (G_OBJECT(gPSButton), "clicked",
                      G_CALLBACK(checkButtonCallback), "gPSButton");

    customLocationButton = (GtkCheckButton *)gtk_check_button_new_with_label ("Custom");
    gtk_table_attach(GTK_TABLE(locationTable), (GtkWidget *)customLocationButton, 1, 2, 0, 1,
		     GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 0);
    gtk_widget_show((GtkWidget *)customLocationButton);
    gtk_toggle_button_set_active((GtkToggleButton *)customLocationButton, useCustomLocation);
    g_signal_connect (G_OBJECT(customLocationButton), "clicked",
                      G_CALLBACK(checkButtonCallback), "customLocationButton");

    locationNameLabel = gtk_label_new("Name");
    gtk_table_attach(GTK_TABLE(locationTable), locationNameLabel, 0, 1, 1, 2,
		     GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 0);
    gtk_widget_set_sensitive(locationNameLabel, useCustomLocation);
    gtk_widget_show(locationNameLabel);
    locationNameText = gtk_entry_new();
    gtk_entry_set_text((GtkEntry *)locationNameText, locationName);
    gtk_table_attach(GTK_TABLE(locationTable), locationNameText, 1, 4, 1, 2,
		     GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 0);
    g_signal_connect (G_OBJECT(locationNameText), "changed",
                      G_CALLBACK(locationNameCallback), "Uh, Clem");
    gtk_widget_set_sensitive(locationNameText, useCustomLocation);
    gtk_widget_show(locationNameText);
    locationLatName = gtk_label_new("Latitude");
    gtk_table_attach(GTK_TABLE(locationTable), locationLatName, 0, 1, 2, 3,
		     GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 0);
    gtk_widget_set_sensitive(locationLatName, useCustomLocation);
    gtk_widget_show(locationLatName);
    latTemp = latitude / DEGREES_TO_RADIANS;
    latDD = (double)((int)latTemp);
    latMM = (double)((int)((latTemp - latDD)*60.0));
    latSS = (latTemp - latDD - 60.0*latMM)*3600.0;
    if (latDD < 0.0) {
      latMM *= -1.0;
      latSS *= -1.0;
    }
    latAdjDD = gtk_adjustment_new(latDD, -90.0, 90.0, 1.0, 1.0, 0.0);
    latSpinDD = gtk_spin_button_new((GtkAdjustment *)latAdjDD, 0.5, 0);
    gtk_spin_button_set_numeric((GtkSpinButton *)latSpinDD, TRUE);
    gtk_table_attach(GTK_TABLE(locationTable), (GtkWidget *)latSpinDD, 1, 2, 2, 3,
		     GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 0);
    gtk_widget_set_sensitive(latSpinDD, useCustomLocation);
    gtk_widget_show((GtkWidget *)latSpinDD);
    latAdjMM = gtk_adjustment_new(latMM, 0.0, 60.0, 1.0, 1.0, 0.0);
    latSpinMM = gtk_spin_button_new((GtkAdjustment *)latAdjMM, 0.5, 0);
    gtk_spin_button_set_numeric((GtkSpinButton *)latSpinMM, TRUE);
    gtk_table_attach(GTK_TABLE(locationTable), (GtkWidget *)latSpinMM, 2, 3, 2, 3,
		     GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 0);
    gtk_widget_set_sensitive(latSpinMM, useCustomLocation);
    gtk_widget_show((GtkWidget *)latSpinMM);
    latAdjSS = gtk_adjustment_new(latSS, 0.0, 60.0, 1.0, 1.0, 0.0);
    latSpinSS = gtk_spin_button_new((GtkAdjustment *)latAdjSS, 0.5, 0);
    gtk_spin_button_set_numeric((GtkSpinButton *)latSpinSS, TRUE);
    gtk_table_attach(GTK_TABLE(locationTable), (GtkWidget *)latSpinSS, 3, 4, 2, 3,
		     GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 0);
    gtk_widget_set_sensitive(latSpinSS, useCustomLocation);
    gtk_widget_show((GtkWidget *)latSpinSS);
    locationLongName = gtk_label_new("Longitude");
    gtk_table_attach(GTK_TABLE(locationTable), locationLongName, 0, 1, 3, 4,
		     GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 0);
    gtk_widget_set_sensitive(locationLongName, useCustomLocation);
    gtk_widget_show(locationLongName);
    longTemp = longitude / DEGREES_TO_RADIANS;
    longDD = (double)((int)longTemp);
    longMM = (double)((int)((longTemp - longDD)*60.0));
    longSS = (longTemp - longDD - 60.0*longMM)*3600.0;
    if (longDD < 0.0) {
      longMM *= -1.0;
      longSS *= -1.0;
    }
    longAdjDD = gtk_adjustment_new(longDD, -180.0, 180.0, 1.0, 1.0, 0.0);
    longSpinDD = gtk_spin_button_new((GtkAdjustment *)longAdjDD, 0.5, 0);
    gtk_spin_button_set_numeric((GtkSpinButton *)longSpinDD, TRUE);
    gtk_table_attach(GTK_TABLE(locationTable), (GtkWidget *)longSpinDD, 1, 2, 3, 4,
		     GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 0);
    gtk_widget_set_sensitive(longSpinDD, useCustomLocation);
    gtk_widget_show((GtkWidget *)longSpinDD);
    longAdjMM = gtk_adjustment_new(longMM, 0.0, 60.0, 1.0, 1.0, 0.0);
    longSpinMM = gtk_spin_button_new((GtkAdjustment *)longAdjMM, 0.5, 0);
    gtk_spin_button_set_numeric((GtkSpinButton *)longSpinMM, TRUE);
    gtk_table_attach(GTK_TABLE(locationTable), (GtkWidget *)longSpinMM, 2, 3, 3, 4,
		     GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 0);
    gtk_widget_set_sensitive(longSpinMM, useCustomLocation);
    gtk_widget_show((GtkWidget *)longSpinMM);
    longAdjSS = gtk_adjustment_new(longSS, 0.0, 60.0, 1.0, 1.0, 0.0);
    longSpinSS = gtk_spin_button_new((GtkAdjustment *)longAdjSS, 0.5, 0);
    gtk_spin_button_set_numeric((GtkSpinButton *)longSpinSS, TRUE);
    gtk_table_attach(GTK_TABLE(locationTable), (GtkWidget *)longSpinSS, 3, 4, 3, 4,
		     GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 0);
    gtk_widget_set_sensitive(longSpinSS, useCustomLocation);
    gtk_widget_show((GtkWidget *)longSpinSS);

    menuLocationButton = (GtkCheckButton *)gtk_check_button_new_with_label ("Menu");
    gtk_table_attach(GTK_TABLE(locationTable), (GtkWidget *)menuLocationButton, 2, 3, 0, 1,
		     GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 0);
    gtk_widget_show((GtkWidget *)menuLocationButton);
    gtk_toggle_button_set_active((GtkToggleButton *)menuLocationButton, useMenuLocation);
    g_signal_connect (G_OBJECT(menuLocationButton), "clicked",
                      G_CALLBACK(checkButtonCallback), "menuLocationButton");

    /*
      Make the menu that holds the classes of locations for the main locations
      menu, such as Europe or Observatory.
    */
    fileName = malloc(strlen(dataDir)+strlen("/menus/regions")+10);
    if (fileName == NULL) {
      perror("regions file");
      exit(ERROR_EXIT);
    }
    sprintf(fileName, "%s/menus/regions", dataDir);
    menuFileFD = open(fileName, O_RDONLY);
    if (menuFileFD < 0) {
      perror(fileName);
      exit(ERROR_EXIT);
    }
    free(fileName);
    while (!eOF) {
      getLine(menuFileFD, &inLine[0], &eOF);
      if (!eOF) {
	int nRead, key;
	char name[50];

	nRead = sscanf(inLine, "%s %d", &name[0], &key);
	if (nRead == 2) {
	  locationClass *newClass;

	  newClass = malloc(sizeof(locationClass));
	  if (newClass == NULL) {
	    perror("newClass");
	    exit(ERROR_EXIT);
	  }
	  newClass->name = malloc(strlen(name)+1);
	  if (newClass->name == NULL) {
	    perror("newClass->name");
	    exit(ERROR_EXIT);
	  }
	  strcpy(newClass->name, name);
	  newClass->key = key;
	  newClass->menuCreated = FALSE;
	  newClass->next = NULL;
	  if (locationClassRoot == NULL)
	    locationClassRoot = newClass;
	  else
	    lastClass->next = newClass;
	  lastClass = newClass;
	  nMenuClasses++;
	}
      }
    }
    close(menuFileFD);
    itemFactoryEntry = (GtkItemFactoryEntry *)malloc((1+nMenuClasses)*
						     sizeof(GtkItemFactoryEntry));
    if (itemFactoryEntry == NULL) {
      perror("itemFactoryEntry");
      exit(ERROR_EXIT);
    }
    itemFactoryEntry[0].path = (gchar *)malloc(strlen("/Region")+1);
    if (itemFactoryEntry[0].path == NULL) {
      perror("(itemFactoryEntry[0].path");
      exit(ERROR_EXIT);
    }
    strcpy(itemFactoryEntry[0].path, "/Region");
    itemFactoryEntry[0].accelerator = NULL;
    itemFactoryEntry[0].callback = NULL;
    itemFactoryEntry[0].callback_action = 0;
    itemFactoryEntry[0].extra_data = NULL;
    itemFactoryEntry[0].item_type = (gchar *)malloc(strlen("<Branch>")+1);
    if (itemFactoryEntry[0].item_type == NULL) {
      perror("itemFactoryEntry[0].item_type");
      exit(ERROR_EXIT);
    }
    strcpy(itemFactoryEntry[0].item_type, "<Branch>");
    lastClass = locationClassRoot;
    for (i = 1; i <= nMenuClasses; i++) {
      int j;
      char path[100];

      dprintf("Class name \"%s\", key %d\n", lastClass->name, lastClass->key);
      sprintf(path, "/Region/%s", lastClass->name);
      for (j = 0; j < strlen(path); j++)
	if (path[j] == '_')
	  path[j] = ' ';
      itemFactoryEntry[i].path = (gchar *)malloc(strlen(path)+1);
      if (itemFactoryEntry[i].path == NULL) {
	perror("itemFactoryEntry[i].path");
	exit(ERROR_EXIT);
      }
      strcpy(itemFactoryEntry[i].path, path);
      itemFactoryEntry[i].accelerator = NULL;
      itemFactoryEntry[i].callback = categoryCallback;
      itemFactoryEntry[i].callback_action = lastClass->key;
      itemFactoryEntry[i].item_type = NULL;
      lastClass = lastClass->next;
    }
    accelGroup = gtk_accel_group_new();
    itemFactory = gtk_item_factory_new(GTK_TYPE_MENU_BAR, "<OrreryRegions>",
				       accelGroup);
    gtk_item_factory_create_items(itemFactory, nMenuClasses+1, itemFactoryEntry,
				  NULL);
    regionMenu = gtk_item_factory_get_widget(itemFactory, "<OrreryRegions>");
    gtk_table_attach(GTK_TABLE(locationTable), (GtkWidget *)regionMenu, 0, 1, 4, 5,
		     GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 0);
    gtk_widget_set_sensitive(regionMenu, useMenuLocation);
    gtk_widget_show((GtkWidget *)regionMenu);
    /* End of making location class menu */

    g_object_ref(locationTable);
    firstCall = FALSE;
  }
  gtk_widget_hide(mainBox);
  gtk_container_remove(GTK_CONTAINER(window), mainBox);
  gtk_container_add(GTK_CONTAINER(window), locationTable);
  gtk_widget_show(locationTable);
  skyIsVisible = FALSE;
}

/*
  The following function is called ad the idle job.   It reads in the constellation
  files, and determines which stars are in the constellations by comparing the
  constelation vertex coordinates to all star coordinates.   This is done so that
  the stars making up a constellation are always plotted, even if they are
  fainter than the magnitude cutoff for the constellation display.   This is a
  time consuming process (several seconds), so it is done in the backround as the
  idle process, in the hope that it will have completed before the user wishes to
  display the constellations.

  This function is executed many times, and each time it does a little more work.
  If there is still work to be done, it exits returning TRUE.   When it is
  finally done, it exits returning FALSE.   This multiple execution strategy
  is used to prevent this function from blocking execution of other routines
  for an extended period.   If it were allowed to execute a single time, from
  start to finish, it would prevent the display from being updated for several
  seconds, which is unacceptable.
*/
static int  initConstellations(gpointer ignored)
{
  static int filesRead = FALSE;
  int i;
  static int callCount = 0;
  int count = 0;
  char dataDirName[1000], inLine[100];
  DIR *dirPtr;
  struct dirent *nextEnt;
  constellation *thisConst, *lastConst = NULL;
  static objectListEntry *currentEntry;

  dprintf("Starting initConstellations (%d)\n", ++callCount);
  if (!filesRead) {
    dprintf("Reading in constellations\n");
    sprintf(dataDirName, "%s/constellations/", dataDir);
    dirPtr = opendir(dataDirName);
    while ((nextEnt = readdir(dirPtr)) != NULL)
      if (strstr(nextEnt->d_name, ".") == NULL) {
	int nPoints = 0;
	int eof = FALSE;
	short starNeededs[100];
	int rAHH, rAMM, decDD, decMM;
	double rASS, decSS, rA, dec;
	double rAs[100], decs[100], sinDecs[100], cosDecs[100];
	char fileName[MAX_FILE_NAME_SIZE];
	int dataFD, negZero;
	
	sprintf(fileName, "%s/constellations/%s", dataDir, nextEnt->d_name);
	dataFD = open(fileName, O_RDONLY);
	if (dataFD >= 0) {
	  int nRead, ii;
	  
	  thisConst = (constellation *)malloc(sizeof(constellation));
	  if (thisConst == NULL) {
	    perror("malloc (thisConst) returned NULL");
	    exit(ERROR_EXIT);
	  }
	  thisConst->next = NULL;
	  if (constellations == NULL)
	    constellations = thisConst;
	  else
	    lastConst->next = thisConst;
	  lastConst = thisConst;
	  thisConst->name = (char *)malloc(strlen(nextEnt->d_name)+1);
	  if (thisConst->name == NULL) {
	    perror("malloc (thisConst->name) returned NULL");
	    exit(ERROR_EXIT);
	  }
	  sprintf(thisConst->name, "%s", nextEnt->d_name);
	  for (ii = 0; ii < strlen(thisConst->name); ii++)
	    if (thisConst->name[ii] == '_')
	      thisConst->name[ii] = ' ';
	  getLine(dataFD, &inLine[0], &eof);
	  nRead = sscanf(inLine, "%d", &(thisConst->type));
	  getLine(dataFD, &inLine[0], &eof);
	  nRead = sscanf(inLine, "%d:%d:%lf %d:%d:%lf", &rAHH, &rAMM, &rASS,
			 &decDD, &decMM, &decSS);
	  rA = (double)rAHH + ((double)rAMM)/60.0 + rASS/3600.0;
	  rA *= HOURS_TO_RADIANS;
	  if (decDD < 0) {
	    decMM *= -1;
	    decSS *= -1.0;
	  }
	  dec = (double)decDD + ((double)decMM)/60.0 + decSS/3600.0;
	  dec *= DEGREES_TO_RADIANS;
	  thisConst->nameRA = rA;
	  thisConst->sinNameDec = sin(dec);
	  thisConst->cosNameDec = cos(dec);
	  while ((nRead == 6) && (!eof)) {
	    getLine(dataFD, &inLine[0], &eof);
	    nRead = sscanf(inLine, "%d:%d:%lf %d:%d:%lf", &rAHH, &rAMM, &rASS,
			 &decDD, &decMM, &decSS);
	    if (nRead == 6) {
	      rA = (double)rAHH + ((double)rAMM)/60.0 + rASS/3600.0;
	      rA *= HOURS_TO_RADIANS;
	      if (strstr(inLine, "-00:") == NULL)
		negZero = FALSE;
	      else
		negZero = TRUE;
	      if ((decDD < 0) || negZero) {
		decMM *= -1;
		decSS *= -1.0;
	      }
	      dec = (double)decDD + ((double)decMM)/60.0 + decSS/3600.0;
	      dec *= DEGREES_TO_RADIANS;
	      if (rA == 0.0)
		starNeededs[nPoints] = FALSE;
	      else
		starNeededs[nPoints] = TRUE;
	      rAs[nPoints] = rA;
	      decs[nPoints] = dec;
	      sinDecs[nPoints] = sin(dec);
	      cosDecs[nPoints++] = cos(dec);
	    }
	  }
	  thisConst->nPoints = nPoints;
	  thisConst->starNeeded = (short *)malloc(nPoints*sizeof(short));
          if (thisConst->starNeeded == NULL) {
            perror("malloc of thisConst->starNeeded returned NULL");
            exit(ERROR_EXIT);
          }
	  thisConst->rA = (double *)malloc(nPoints*sizeof(double));
	  if (thisConst->rA == NULL) {
	    perror("malloc of thisConst->rA returned NULL");
	    exit(ERROR_EXIT);
	  }
	  thisConst->dec = (double *)malloc(nPoints*sizeof(double));
	  if (thisConst->dec == NULL) {
	    perror("malloc of thisConst->dec returned NULL");
	    exit(ERROR_EXIT);
	  }
	  thisConst->sinDec = (double *)malloc(nPoints*sizeof(double));
	  if (thisConst->sinDec == NULL) {
	    perror("malloc of thisConst->sinDec returned NULL");
	    exit(ERROR_EXIT);
	  }
	  thisConst->cosDec = (double *)malloc(nPoints*sizeof(double));
	  if (thisConst->cosDec == NULL) {
	    perror("malloc of thisConst->cosDec returned NULL");
	    exit(ERROR_EXIT);
	  }
	  for (i = 0; i < nPoints; i++) {
	    thisConst->starNeeded[i] = starNeededs[i];
	    thisConst->rA[i] = rAs[i];
	    thisConst->dec[i] = decs[i];
	    thisConst->sinDec[i] = sinDecs[i];
	    thisConst->cosDec[i] = cosDecs[i];
	  }
	  close(dataFD);
	} else
	  perror(nextEnt->d_name);
      }
    currentEntry = objectListRoot;
    filesRead = TRUE;
    return(TRUE);
  }
  if (currentEntry != NULL) {
    dprintf("Starting while loop, hip #%d\n", currentEntry->hipNumber);
    while (currentEntry != NULL) {
      if (currentEntry->mag <= CONSTELLATION_LIMIT)
	inConstellation(currentEntry);
      currentEntry = currentEntry->forwardPointer;
      if (count++ > 100) {
	dprintf("Exiting after processing 100\n");
	return(TRUE);
      }
    }
  }
  constellationsInitialized = TRUE;
  if (shouldSwitchScreens) {
    switchScreens();
    fullRedraw(FALSE);
  }
  dprintf("Exiting initConstellations\n");
  return(FALSE);
}

/*
  Signal handler to catch ^C and restore initial correlator state and
  antenna offsets before exiting.
*/
void signalHandler(int signum)
{
  if (signum == SIGPIPE) {
    close(gPSDSocket);
    gPSDSocketOpen = FALSE;
  } else
    fprintf(stderr,"point received unexpected signal #%d\n",signum);
}

struct sigaction action, oldAction;

void usage(void)
{
  fprintf(stderr, "usage: orrery [-d directory]\n");
  exit(-1);
}

int main(int argc, char **argv)
{
  /*
    OK, the following is a kludge.  I admit it.  I originally called the
    popt package to parse the command line arguments.   But some Openmoko
    distributions do not include the popt lib.   This became a hassle.
    So I decided not only to eliminate the popt dependancy, but to use no
    command line parsing at all, so that there are no unneeded dependancies.
    there is only one command line option available, after all, so it;s
    not too hare to just handle it in its entirety here.
   */

  if (((argc != 1) && (argc != 3)) ||
      ((argc == 3) && (strcmp(argv[1], "-d") != 0)))
    usage();
  if (argc == 3)
    sprintf(dirName, "%s/", argv[2]);
  action.sa_flags = 0;
  sigemptyset(&action.sa_mask);
  action.sa_handler = signalHandler;
  sigaction(SIGPIPE, &action, &oldAction);
   if (argc != 3)
    strcpy(dataDir, DATA_DIR);
  else
    strcpy(dataDir, argv[2]);

  dprintf("Calling parseConfigFile\n");
  parseConfigFile();
  dprintf("Calling readStarNamesFile");
  readStarNamesFile();
  dprintf("Calling readHipparcoseCatalog\n");
  readHipparcosCatalog();
  dprintf("Calling newPosition\n");
  newPosition();
  dprintf("Calling gtk_init\n");
  gtk_init(&argc, &argv);
  
  /* Initialize main window */
  window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_widget_set_size_request (GTK_WIDGET (window), 480, 640);
  gtk_window_set_title (GTK_WINDOW (window), "orrery");
  g_signal_connect (G_OBJECT (window), "delete_event",
		    G_CALLBACK (gtk_main_quit), NULL);

  mainBox = gtk_vbox_new(FALSE, 0);
  gtk_container_add(GTK_CONTAINER(window), mainBox);
  g_object_ref(mainBox); /* This keeps mainBox from being destroyed when not displayed */
  gtk_widget_show(mainBox);

  controlButtonBox = gtk_hbox_new(FALSE, 0);
  gtk_widget_show(controlButtonBox);
  /* Set up buttons which appear at the bottom when not fullscreened */

  /* Now, the items button, which controls what things are plotted on the display */
  itemsButton = gtk_button_new_with_mnemonic("items");
  gtk_widget_set_size_request(itemsButton, -1, BUTTON_HEIGHT);
  g_signal_connect ((gpointer) itemsButton, "clicked",
                    G_CALLBACK (itemsButtonClicked), NULL);
  gtk_box_pack_end(GTK_BOX(controlButtonBox), itemsButton, TRUE, TRUE, 0);

  /* Now, the location button, which controls where the observer is */
  locationButton = gtk_button_new_with_mnemonic("place");
  gtk_widget_set_size_request(locationButton, -1, BUTTON_HEIGHT);
  g_signal_connect ((gpointer) locationButton, "clicked",
                    G_CALLBACK (locationButtonClicked), NULL);
  gtk_box_pack_end(GTK_BOX(controlButtonBox), locationButton, TRUE, TRUE, 0);

  /* Now, the time button, which controls what time the program thinks it is. */
  timeButton = gtk_button_new_with_mnemonic("time");
  gtk_widget_set_size_request(timeButton, -1, BUTTON_HEIGHT);
  g_signal_connect ((gpointer) timeButton, "clicked",
                    G_CALLBACK (timeButtonClicked), NULL);
  gtk_box_pack_end(GTK_BOX(controlButtonBox), timeButton, TRUE, TRUE, 0);

  /* Now, the options button, which controls where the observer is */
  optionsButton = gtk_button_new_with_mnemonic("opts");
  gtk_widget_set_size_request(optionsButton, -1, BUTTON_HEIGHT);
  g_signal_connect ((gpointer) optionsButton, "clicked",
                    G_CALLBACK (optionsButtonClicked), NULL);
  gtk_box_pack_end(GTK_BOX(controlButtonBox), optionsButton, TRUE, TRUE, 0);

  /* Last, the fullscreen button */
  fullscreenButton = gtk_button_new_with_mnemonic("FS");
  gtk_widget_set_size_request(fullscreenButton, -1, BUTTON_HEIGHT);
  gtk_box_pack_end(GTK_BOX(controlButtonBox), fullscreenButton, TRUE, TRUE, 0);
  g_signal_connect ((gpointer) fullscreenButton, "clicked",
                    G_CALLBACK (fullscreenButtonClicked), NULL);

  gtk_box_pack_end(GTK_BOX(mainBox), controlButtonBox, TRUE, TRUE, 0);
  gtk_widget_show(controlButtonBox);

  /* Set up window in which to plot the graphics ...*/
  drawingArea = gtk_drawing_area_new();
  gtk_widget_set_size_request (GTK_WIDGET(drawingArea), VGA_WIDTH, VGA_HEIGHT-BUTTON_HEIGHT);
  gtk_box_pack_end(GTK_BOX(mainBox), drawingArea, TRUE, TRUE, 0);
  g_signal_connect(G_OBJECT(drawingArea), "expose_event",
		   G_CALLBACK(exposeEvent), NULL);
  g_signal_connect(G_OBJECT(drawingArea), "configure_event",
		   G_CALLBACK(configureEvent), NULL);
  g_signal_connect(G_OBJECT(drawingArea), "button_release_event",
		   G_CALLBACK(buttonReleaseEvent), NULL);
  g_signal_connect(G_OBJECT(drawingArea), "button_press_event",
		   G_CALLBACK(buttonPressEvent), NULL);
  g_signal_connect(G_OBJECT(drawingArea), "motion_notify_event",
		   G_CALLBACK(motionNotifyEvent), NULL);
  g_timeout_add(60000, periodicUpdate, NULL);
  gtk_widget_show(window);
  makeGraphicContexts(window);
  makeFonts(window);
  gtk_widget_set_events(drawingArea,
			GDK_EXPOSURE_MASK       | GDK_BUTTON_PRESS_MASK  |
			GDK_BUTTON_RELEASE_MASK | GDK_POINTER_MOTION_MASK );
  gtk_widget_show(drawingArea);
  gtk_widget_show(fullscreenButton);
  gtk_widget_show(itemsButton);
  gtk_widget_show(locationButton);
  gtk_widget_show(timeButton);
  gtk_widget_show(optionsButton);
  centerAzD = (double)centerAz;
  gtk_idle_add(initConstellations, NULL);
  dprintf("Calling gtk_main\n");
  gtk_main();
  return 0;
}
